Package: EpiTxDb
Type: Package
Title: Storing and accessing epitranscriptomic information using the
        AnnotationDbi interface
Version: 1.23.0
Date: 2025-09-23
Authors@R: person("Felix G.M.", 
                  "Ernst", 
                  email = "felix.gm.ernst@outlook.com", 
                  role = c("aut", "cre"),
                  comment = c(ORCID = "0000-0001-5064-0928"))
Description: 
    EpiTxDb facilitates the storage of epitranscriptomic information. More 
    specifically, it can keep track of modification identity, position, the
    enzyme for introducing it on the RNA, a specifier which determines the
    position on the RNA to be modified and the literature references each
    modification is associated with.
License: Artistic-2.0
Encoding: UTF-8
LazyData: false
biocViews: Software, Epitranscriptomics
Depends: R (>= 4.0), AnnotationDbi, Modstrings
Imports: methods, utils, httr, xml2, curl, rex, GenomicFeatures,
        txdbmaker, GenomicRanges, Seqinfo, BiocGenerics, BiocFileCache,
        S4Vectors, IRanges, RSQLite, DBI, Biostrings, tRNAdbImport
Suggests: BiocStyle, knitr, rmarkdown, testthat, httptest,
        AnnotationHub, ensembldb, ggplot2, EpiTxDb.Hs.hg38,
        BSgenome.Hsapiens.UCSC.hg38, BSgenome.Scerevisiae.UCSC.sacCer3,
        TxDb.Hsapiens.UCSC.hg38.knownGene
Collate: 'AllGenerics.R' 'EpiTxDb-SELECT-helpers.R' 'EpiTxDb-schema.R'
        'EpiTxDb.R' 'EpiTxDb-class.R' 'makeEpiTxDb.R'
        'makeEpiTxDbFromGRanges.R' 'shiftGenomicToTranscript.R'
        'makeEpiTxDbFromRMBase.R' 'makeEpiTxDbFromtRNAdb.R'
        'modifications.R' 'modificationsBy.R' 'ranges-helpers.R'
        'select-methods.R'
RoxygenNote: 7.3.3
BugReports: https://github.com/FelixErnst/EpiTxDb/issues
URL: https://github.com/FelixErnst/EpiTxDb
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/EpiTxDb
git_branch: devel
git_last_commit: 23585ca
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-11-04
NeedsCompilation: no
Packaged: 2025-11-04 22:30:03 UTC; biocbuild
Author: Felix G.M. Ernst [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5064-0928>)
Maintainer: Felix G.M. Ernst <felix.gm.ernst@outlook.com>

# Generated by roxygen2: do not edit by hand

export(EPITXDB_RMBASE_URL)
export(downloadRMBaseFiles)
export(getRMBaseDataAsGRanges)
export(gettRNAdbDataAsGRanges)
export(listAvailableGenomesFromRMBase)
export(listAvailableModFromRMBase)
export(listAvailableOrganismsFromRMBase)
export(listAvailableOrganismsFromtRNAdb)
export(makeEpiTxDb)
export(makeEpiTxDbFromGRanges)
export(makeEpiTxDbFromRMBase)
export(makeEpiTxDbFromRMBaseFiles)
export(makeEpiTxDbFromtRNAdb)
exportClasses(EpiTxDb)
exportMethods(as.integer)
exportMethods(as.list)
exportMethods(columns)
exportMethods(keys)
exportMethods(keytypes)
exportMethods(modifications)
exportMethods(modificationsBy)
exportMethods(modifiedSeqsByTranscript)
exportMethods(organism)
exportMethods(positionSequence)
exportMethods(rescale)
exportMethods(select)
exportMethods(seqinfo)
exportMethods(seqlevels)
exportMethods(shiftGenomicToTranscript)
exportMethods(shiftTranscriptToGenomic)
import(AnnotationDbi)
import(BiocGenerics)
import(IRanges)
import(Modstrings)
import(RSQLite)
import(S4Vectors)
import(Seqinfo)
import(methods)
import(rex)
import(txdbmaker)
importClassesFrom(IRanges,PartitioningByEnd)
importClassesFrom(IRanges,PartitioningByWidth)
importFrom(BiocFileCache,BiocFileCache)
importFrom(BiocFileCache,bfcadd)
importFrom(BiocFileCache,bfcquery)
importFrom(BiocGenerics,organism)
importFrom(Biostrings,DNAStringSet)
importFrom(Biostrings,subseq)
importFrom(Biostrings,vwhichPDict)
importFrom(DBI,dbListFields)
importFrom(DBI,dbListTables)
importFrom(GenomicFeatures,transcripts)
importFrom(GenomicRanges,findOverlaps)
importFrom(IRanges,PartitioningByEnd)
importFrom(IRanges,PartitioningByWidth)
importFrom(S4Vectors,findMatches)
importFrom(Seqinfo,seqinfo)
importFrom(Seqinfo,seqlevels)
importFrom(curl,curl)
importFrom(httr,POST)
importFrom(httr,content)
importFrom(httr,modify_url)
importFrom(tRNAdbImport,import.tRNAdb)
importFrom(utils,packageDescription)
importFrom(utils,read.delim)
importFrom(xml2,xml_find_all)

# EpiTxDb <img src="https://raw.githubusercontent.com/Bioconductor/BiocStickers/devel/EpiTxDb/EpiTxDb.png" height="200" align="right">

<!-- badges: start -->
[![R-CMD-check](https://github.com/FelixErnst/EpiTxDb/workflows/R-CMD-check-bioc-devel/badge.svg)](https://github.com/FelixErnst/EpiTxDb/actions/)
[![BioC Build](https://bioconductor.org/shields/build/devel/bioc/EpiTxDb.svg)](http://bioconductor.org/checkResults/devel/bioc-LATEST/EpiTxDb/)
[![codecov](https://codecov.io/gh/FelixErnst/EpiTxDb/branch/devel/graph/badge.svg)](https://codecov.io/gh/FelixErnst/EpiTxDb)
[![BioC Years](https://bioconductor.org/shields/years-in-bioc/EpiTxDb.svg)](https://doi.org/doi:10.18129/B9.bioc.EpiTxDb)
<!-- badges: end -->


The epitranscriptome includes all post-transcriptional modifications of the RNA
and describes and additional layer of information encoded on RNA. Like the term
epigenome it is not about a change in nucleotide sequences, but the addition of
functional elements through modifications.

With the development of high throughput detection strategies for specific RNA
modifications, such as miCLIP and Pseudo-Seq amongst other, a large number of
modified positions have been identified and were summarized via the RMBase 
project ([Xuan et al. 2017, Sun et al. 2015](#Literature)) project.

To make these information avaialble within the Bioconductor universe `EpiTxDb`
was developed, which facilitates the storage of epitranscriptomic information.
More specifically, it can keep track of modification identity, position, the
enzyme for introducing it on the RNA, a specifier which determines the position
on the RNA to be modified and the literature references each modification is
associated with.

# Installation

The current version of the `EpiTxDb` package is available from Bioconductor.

## Github

```
remotes::install_github("FelixErnst/EpiTxDb")
#
library(EpiTxDb)
```

## Bioconductor

```
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

# The following initializes usage of Bioc devel
BiocManager::install(version='devel')
BiocManager::install("EpiTxDb")
library(EpiTxDb)
```

# Literature

- Jia-Jia Xuan, Wen-Ju Sun, Ke-Ren Zhou, Shun Liu, Peng-Hui Lin, Ling-Ling
Zheng, Liang-Hu Qu, Jian-Hua Yang (2017): "RMBase v2.0: Deciphering the Map of
RNA Modifications from Epitranscriptome Sequencing Data." Nucleic Acids
Research, Volume 46, Issue D1, 4 January 2018, Pages D327–D334.
doi:[10.1093/nar/gkx934](https://doi.org/10.1093/nar/gkx934)

- Wen-Ju Sun, Jun-Hao Li, Shun Liu, Jie Wu, Hui Zhou, Liang-Hu Qu, Jian-Hua Yang
(2018): "RMBase: a resource for decoding the landscape of RNA modifications from
high-throughput sequencing data", Nucleic Acids Research, Volume 44, Issue D1, 4
January 2016, Pages D259–D265.
doi:[10.1093/nar/gkv1036](https://doi.org/10.1093/nar/gkv1036).

#' @include EpiTxDb-class.R
NULL

#' @name modifications
#' @aliases modifications modificationsBy
#'
#' @title Getting modification data from a \code{EpiTxDb-object}
#'
#' @description \code{modifications} and \code{modificationsBy} are functions to
#' extract modification annotation from a \code{\link[=EpiTxDb-class]{EpiTxDb}}
#' object.
#'
#' \code{modifiedSeqsByTranscript} returns a
#' \code{\link[Modstrings:ModStringSet]{ModRNAStringSet}} from a \code{EpiTxDb}
#' object and compatible \code{RNAStringSet} object. This used the
#' \code{\link[Modstrings:separate]{combineIntoModstrings()}} function from the
#' \code{Modstrings} package.
#'
#' @param x a \code{\link[=EpiTxDb-class]{EpiTxDb}}
#' @param columns Columns to include in the result. If the vector is named,
#'   those names are used for the corresponding column in the element metadata
#'   of the returned object. (default: \code{columns =
#'   c("mod_id","mod_type","mod_name")})
#' @param by By which information type should the result be split into? A
#'   \code{character} value from one of the following values:
#'   \itemize{
#'     \item seqnames
#'     \item mod_type
#'     \item reaction
#'     \item specifier
#'     \item specifier_type
#'   }
#' @param filter Either NULL or a named list of vectors to be used to restrict
#'   the output. Valid names for this list are: "mod_id", "mod_type",
#'   "mod_name", "sn_id", "sn_name", "rx_genename", "rx_ensembl",
#'   "rx_ensembltrans", "rx_entrezid", "spec_genename", "spec_type",
#'   "spec_ensembl", "spec_ensembltrans", "spec_entrezid" , "ref_type" and
#'   "ref". (default: \code{filter = NULL})
#' @param use.names \code{TRUE} or \code{FALSE}. If \code{TRUE}, the
#'   modification names are set as the names of the returned object. (default:
#'   \code{use.names = FALSE})
#' @param sequences A \code{RNAStringSet}, which can be used as input for
#'   \code{\link[Modstrings:separate]{combineIntoModstrings()}}. See
#'   \code{\link[Modstrings:separate]{?combineIntoModstrings}} for additional
#'   details.
#' @param ... Not used.
#'
#' @return a \code{\link[GenomicRanges:GRanges-class]{GRanges}} object for
#'   \code{modifications} and a
#'   \code{\link[GenomicRanges:GRangesList-class]{GRangesList}} for
#'   \code{modificationsBy}.
#'
#' @examples
#' etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
#'                         package="EpiTxDb")
#' etdb <- loadDb(etdb_file)
#' etdb
NULL

# helper functions for extracting feature data ---------------------------------

.extract_modifications <- function(epitxdb, table, mcolumns = character(0),
                                   filter = list(), core_columns){
    schema_version <- EpiTxDb_schema_version(epitxdb)
    names(mcolumns) <- EPITXDB_column2table(mcolumns, from_table = table,
                                            schema_version = schema_version)
    proxy_column <- orderby <- c(modification = "_mod_id")

    ## 1st SELECT: extract stuff from the proxy table.
    columns1 <- union(core_columns, mcolumns[names(mcolumns) == table])
    df1 <- EpiTxDb_SELECT_from_LEFT_JOIN(epitxdb, table, columns1,
                                         filter = filter, orderby = orderby)

    ## Additional SELECTs: 1 additional SELECT per satellite table
    foreign_columns <- mcolumns[names(mcolumns) != table]
    foreign_columns <- split(foreign_columns, names(foreign_columns))
    satellite_tables <- names(foreign_columns)
    names(satellite_tables) <- satellite_tables
    df_list <- lapply(satellite_tables, function(satellite_table) {
        columns2 <- foreign_columns[[satellite_table]]
        if (length(filter) == 0L) {
            filter2 <- list()
        } else {
            filter2 <- list(df1[[proxy_column]])
            names(filter2) <- proxy_column
        }
        if (satellite_table %in% c("seqnames","reaction","specifier",
                                   "reference")) {
            columns2 <- c(proxy_column, columns2)
            EpiTxDb_SELECT_from_LEFT_JOIN(epitxdb, satellite_table, columns2,
                                          filter = filter2, orderby = orderby)
        } else {
            stop(satellite_table, ": unsupported satellite table")
        }
    })
    df1 <- list(df1)
    names(df1) <- table
    ans <- c(df1, df_list)
    ans
}


# .extract_features_as_GRanges() -----------------------------------------------

.make_DataFrame_from_df_list <- GenomicFeatures:::.make_DataFrame_from_df_list
.assignMetadataList <- GenomicFeatures:::.assignMetadataList

.as_db_columns <- function(columns)
    sub("^(mod_id|sn_id|rx_id|spec_id|ref_id)$", "_\\1", columns)

.extract_modifications_as_GRanges <- function(epitxdb, mcolumns = character(0),
                                              filter = list(),
                                              use.names = FALSE)
{
    if (!isTRUEorFALSE(use.names)) {
        stop("'use.names' must be TRUE or FALSE")
    }
    table <- c("modification")
    # save the selected metadata columns for subsetting later on
    mcolumns0 <- mcolumns
    # add the seqnames columns, since the always need to be extracted
    mcolumns <- unique(c(mcolumns,"sn_id","sn_name"))
    db_mcolumns <- db_mcolumns0 <- .as_db_columns(mcolumns)
    core_columns <- EPITXDB_table_columns(table)
    names(filter) <- .as_db_columns(names(filter))
    df_list <- .extract_modifications(epitxdb, table, db_mcolumns, filter,
                                      core_columns)
    DF <- .make_DataFrame_from_df_list(df_list)
    DF$seqnames <- .get_seqnames(DF)
    ans <- GenomicRanges::makeGRangesFromDataFrame(
        DF,
        seqinfo = .get_EpiTxDb_seqinfo(epitxdb),
        seqnames.field = "seqnames",
        start.field = "mod_start",
        end.field = "mod_end",
        strand.field = "mod_strand")
    if (use.names) {
        ans_names <- DF[ ,"mod_name"]
        names(ans) <- ans_names
    }
    mcols(ans) <- DF[db_mcolumns0]
    colnames(mcols(ans)) <- mcolumns
    # subset to the selected columns
    mcols(ans) <- mcols(ans)[,colnames(mcols(ans)) %in% mcolumns0,drop = FALSE]
    ans
}

# translate external to internal column names
translateCols <- function(columns, epitxdb){
    ## preserve any names
    oriColNames <- names(columns)
    ## and save the original column strings
    oriCols <- columns

    oriLen <- length(columns) ## not always the same as length(oriCols)
    ## get the available abbreviations as a translation vector (exp)
    names <- .makeColAbbreviations(epitxdb)
    exp <- sub("^_","", names(names))
    names(exp) <- names

    ## Then replace only those IDs that match the UC names
    m <- match(oriCols, names(exp))
    idx = which(!is.na(m))
    columns[idx] <- exp[m[idx]]

    if(length(columns) == oriLen && is.null(oriColNames)) {
        names(columns) <- oriCols
    } else if(length(columns) == oriLen && !is.null(oriColNames)) {
        names(columns) <- oriColNames
    } else if(length(columns) != oriLen) {
        stop("names were lost in translateCols() helper")
    }
    columns
}

.extractFromEpiTxDb <- function(epitxdb, mcolumns = NULL,
                                filter = NULL, use.names = FALSE){
    user_mcolumns <- mcolumns
    mcolumns <- translateCols(mcolumns, epitxdb)
    if (is.null(filter))
        filter <- list()
    names(filter) <- translateCols(names(filter), epitxdb)
    ans <- .extract_modifications_as_GRanges(epitxdb, mcolumns, filter,
                                             use.names)
    names(mcols(ans)) <- if (is.null(names(user_mcolumns))) user_mcolumns
    else names(user_mcolumns)
    .assignMetadataList(ans, epitxdb)
}

#' @rdname modifications
#' @export
setMethod("modifications", "EpiTxDb",
    function(x, columns = c("mod_id","mod_type","mod_name"),
             filter = NULL, use.names = FALSE){
        .extractFromEpiTxDb(x, mcolumns = columns, filter = filter,
                            use.names = use.names)
    }
)

#' @include EpiTxDb-class.R
#' @include makeEpiTxDb.R
#' @include shiftGenomicToTranscript.R
NULL

#' @name makeEpiTxDbFromRMBase
#'
#' @title Create a \code{EpiTxDb} object from RMBase v2.0 online resources
#'
#' @description
#' \code{makeEpiTxDbFromRMBase} will make use of the RMBase v2.0 online
#' resources.
#'
#' @param organism A \code{character} value, which must match an organism
#'   descriptor on the RMBase download website.
#' @param genome A \code{character} value, which must match a genome
#'   descriptor on the RMBase download website.
#' @param modtype A \code{character} value, which must match one or more
#'   modification descriptors on the RMBase download website.
#' @param files From \code{organism}, \code{genome} and \code{modtype} the
#'   available files will be downloaded using the
#'   \code{\link[BiocFileCache:BiocFileCache-class]{BiocFileCache}} interface
#'   and passed on to \code{makeEpiTxDbFromRMBaseFiles}. However, individual
#'   files can be provided as well.
#' @param tx A \code{\link[GenomicRanges:GRangesList-class]{GRangesList}} object
#'   which will be used to shift the genomic coordinates to transcript
#'   coordinates. This is optional, but highly recommended. (default:
#'   \code{tx = NULL}).
#' @param sequences A named \code{DNAStringSet} or \code{RNAStringSet}, which
#'   will be used to check whether the defined modifications are compatible with
#'   the original base. This uses
#'   \code{\link[Modstrings:separate]{removeIncompatibleModifications()}}
#'   function from the \code{Modstrings} package.
#' @param verbose \code{TRUE} or \code{FALSE}: Should verbose message be prined?
#' @param metadata,reassign.ids See \code{\link[=makeEpiTxDb]{makeEpiTxDb}}
#'
#' @return a \code{EpiTxDb} object.
#'
#' @export
NULL

#' @rdname makeEpiTxDbFromRMBase
#' @export
EPITXDB_RMBASE_URL <- "https://rna.sysu.edu.cn/rmbase/download/"
EPITXDB_RMBASE_URL_JSON <- paste0(EPITXDB_RMBASE_URL,"ajax/download.json")

# makeEpiTxDbFromRMBase --------------------------------------------------------

.get_RMBase_rnames <- function(organism, genome, modtype){
    rmbase_data <- NULL
    utils::data("rmbase_data", envir = environment(), package = "EpiTxDb")
    rmbase_data[rmbase_data$organism == organism & rmbase_data$genome == genome & rmbase_data$mod == modtype,]$dataSet
}

.check_RMBase_files_available <- function(bfc, organism, genome, modtype){
    # get BiocFileCache information
    rnames <- .get_RMBase_rnames(organism, genome, modtype)
    bfci <- BiocFileCache::bfcinfo(bfc)
    m <- match(rnames,bfci$rname)
    m <- m[!is.na(m)]
    res <- bfci[m,]
    #
    nrow(res) == length(rnames)
}

.get_RMBase_files_available <- function(bfc, organism, genome, modtype){
    # get BiocFileCache information
    rnames <- .get_RMBase_rnames(organism, genome, modtype)
    bfci <- BiocFileCache::bfcinfo(bfc)
    m <- match(rnames,bfci$rname)
    m <- m[!is.na(m)]
    res <- bfci[m,]
    #
    files <- as.list(res$rpath)
    if(length(rnames) != length(files)){
        stop(".")
    }
    #
    m <- match(bfci$rname, rnames)
    m <- m[!is.na(m)]
    # update resource if expired
    check_update_required <- BiocFileCache::bfcneedsupdate(bfc, res$rid)
    check_update_required <- check_update_required &
        !is.na(check_update_required)
    files[check_update_required] <-
        Map(function(rid){
            BiocFileCache::bfcdownload(bfc,rid)
        },
        res$rid[check_update_required])
    unname(unlist(files))
}

#' @importFrom BiocFileCache bfcquery bfcadd BiocFileCache
.download_RMBase_files <- function(bfc, organism, genome, modtype){
    # get file names
    files <- .get_RMBase_files(organism)
    f_genome <- vapply(strsplit(files,"_"),"[",character(1),2L) == genome
    f_mod <- vapply(strsplit(files[f_genome],"_"),"[",character(1),4L) %in% modtype
    files <- files[f_genome][f_mod]
    urls <- paste0(EPITXDB_RMBASE_URL,organism,"/zip/",files)
    #
    rnames <- .get_RMBase_rnames(organism, genome, modtype)
    if(length(rnames) != length(urls)){
        stop(".")
    }
    # get BiocFileCache info
    bfci <- BiocFileCache::bfcinfo(bfc)
    # check which rnames are available
    rnames_available <- rnames %in% bfci$rname
    # get file name if available
    files <- as.list(rep("",length(rnames)))
    files[rnames_available] <-
        Map(function(rname){
            BiocFileCache::bfcquery(bfc,rname)[,"rpath"]
        },
        rnames[rnames_available])
    # download if not available and return file name
    files[!rnames_available] <-
        Map(function(url, rname){
            BiocFileCache::bfcadd(bfc,rname,url)
        },
        urls[!rnames_available],
        rnames[!rnames_available])
    unname(unlist(files))
}

#' @rdname makeEpiTxDbFromRMBase
#' @export
downloadRMBaseFiles <- function(organism, genome, modtype){
    bfc <- BiocFileCache::BiocFileCache()
    if(!.check_RMBase_files_available(bfc, organism, genome, modtype)){
        modtypes <- listAvailableModFromRMBase(organism, genome)
        if(!all(modtype %in% modtypes)){
            stop("'modtype' must be a valid modification type from ",
                 "listAvailableModFromRMBase() for the given 'organism' and ",
                 "'genome'.")
        }
        message("Downloading RMBase v2.0 files ...")
        files <- .download_RMBase_files(bfc, organism, genome, modtype)
    } else {
        files <- .get_RMBase_files_available(bfc, organism, genome, modtype)
    }
    files
}

#' @rdname makeEpiTxDbFromRMBase
#' @export
makeEpiTxDbFromRMBase <- function(organism, genome, modtype, tx = NULL,
                                  sequences = NULL, metadata = NULL,
                                  reassign.ids = FALSE, verbose = FALSE){
    # Input check
    if(!(is.logical(reassign.ids) && length(reassign.ids) == 1L)){
        stop("'reassign.ids' must be TRUE or FALSE.", call. = FALSE)
    }
    if(!(is.logical(verbose) && length(verbose) == 1L)){
        stop("'verbose' must be TRUE or FALSE.", call. = FALSE)
    }
    #
    message("Loading RMBase files ...")
    files <- downloadRMBaseFiles(organism, genome, modtype)
    makeEpiTxDbFromRMBaseFiles(files, tx = tx, sequences = sequences,
                               metadata = metadata, reassign.ids = reassign.ids,
                               verbose = verbose)
}

# makeEpiTxDbFromRMBaseFiles ---------------------------------------------------

.get_RMBase_header <- function(file){
    header <- readLines(file,7L)
    header
}

.get_RMBase_colnames <- function(file){
    colnames <- .get_RMBase_header(file)[7L]
    strsplit(trimws(gsub("#","",colnames)), "\\p{Zs}+|\\t", perl=TRUE)[[1L]]
}

EPITXDB_RMBASE_REQ_COLUMS <- c("chromosome", "modStart", "modEnd", "modId",
                               "score", "strand", "modName", "modType",
                               "supportList", "pubmedIds", "geneName",
                               "sequence")

#' @importFrom utils read.delim
.read_RMBase_file <- function(file){
    rmb <- try(read.delim(file, skip = 7L), silent = TRUE)
    if(is(rmb,"try-error")){
        stop("Malformated input file: ", file,".\nError message: ",
             as.character(rmb),
             call. = FALSE)
    }
    if(ncol(rmb) < 15L){
        stop(".")
    }
    colnames <- .get_RMBase_colnames(file)
    if(length(colnames) < ncol(rmb)){
        stop("Header of input could not read correctly. Mismatching number of ",
             "data columns and column labels.",
             call. = FALSE)
    }
    colnames(rmb) <- colnames[seq_len(ncol(rmb))]
    f <- EPITXDB_RMBASE_REQ_COLUMS %in% colnames(rmb)
    if(!all(f)){
        stop("Missing required column from RMBase data. (File: ",
             file, " / Missing columns: ",
             EPITXDB_RMBASE_REQ_COLUMS[!(f)],")",
             call. = FALSE)
    }
    rmb <- rmb[,colnames(rmb) %in% EPITXDB_RMBASE_REQ_COLUMS]
    rmb
}

# if a prefix in the chromosome identifier is present, try to remove it
# only if it simplifies the result
.simplify_chromosome_identifiers <- function(chromosome, seqlevels = NA){
    chromosome <- as.character(chromosome)
    # fix some weird mito chromosome annotation
    f <- chromosome %in% "Mito"
    if(any(f)){
        chromosome[f] <- gsub("Mito","M",chromosome[f])
    }
    # mismatching chromosomes
    if(all(!is.na(seqlevels))){
        mm_chr <- !(chromosome %in% seqlevels)
    } else {
        return(factor(chromosome,unique(chromosome)))
    }
    f_chr_remove <- grepl("Chr|chr",chromosome[mm_chr])
    if(any(f_chr_remove)){
        if(all(levels(factor(gsub("Chr|chr","",chromosome[mm_chr][f_chr_remove]))) %in%
               levels(chromosome[mm_chr][f_chr_remove])) ||
           all(levels(chromosome[mm_chr][f_chr_remove]) %in%
               levels(factor(gsub("Chr|chr","",chromosome[mm_chr][f_chr_remove]))))){
            tmp <- gsub("Chr|chr","",chromosome[mm_chr])
            chromosome[mm_chr] <- tmp
        }
    } else {
        chr_add1 <- paste0("chr",chromosome[mm_chr])
        if(all(chr_add1 %in% seqlevels)){
            chromosome[mm_chr] <- chr_add1
        } else {
            chr_add2 <- paste0("Chr",chromosome[mm_chr])
            if(all(chr_add2 %in% seqlevels)){
                chromosome[mm_chr] <- chr_add2
            }
        }
    }
    factor(chromosome,unique(chromosome))
}

.fix_non_standard_mod_types <- function(mod_type){
    mod_type <- gsub("m22","m2,2",mod_type)
    mod_type <- gsub("m42","m4,4",mod_type)
    mod_type <- gsub("m62","m6,6",mod_type)
    mod_type <- gsub("Tm","Um",mod_type)
    if("m5C" %in% unique(mod_type)){
        mod_type <- gsub("N","m5C",mod_type)
    }
    mod_type
}

#' @importFrom Biostrings DNAStringSet subseq
.extract_GRanges_from_RMBase <- function(rmb, seqtype = "RNA", verbose = FALSE){
    ############################################################################
    ### check modification information on correct base
    seq <- Biostrings::DNAStringSet(rmb$sequence)
    if(unique(width(seq)) != 41L){
        stop(".")
    }
    seqtype(seq) <- seqtype
    seqtype <- paste0("Mod",seqtype(seq))
    # get the type of modification annotation
    mod_type <- .fix_non_standard_mod_types(as.character(rmb$modType))
    nc_type <- Modstrings:::.get_nc_type(mod_type, seqtype)
    if(nc_type != "short"){
        stop(".")
    }
    # get original base
    codec <- Modstrings:::modscodec(seqtype)
    modValues <- Modstrings:::.norm_seqtype_modtype(mod_type, seqtype,
                                                    nc_type)
    f <- Modstrings:::values(codec)[match(modValues,
                                          Modstrings:::values(codec))]
    # check if reported base matches original base, 21 is the middle position of
    # 41 nucleotides always returned from RMBase
    subseq <- Biostrings::subseq(seq,21L,21L)
    base_mm <- Modstrings:::originatingBase(codec)[f] != subseq
    # if not delete the modifications with a warning
    if(any(base_mm)){
        warning("Detected ",sum(base_mm),
                " mismatch(es) of modification and originating base in ",
                "RMBase data for '", paste(unique(mod_type), collapse = "', '"),
                "'. Removing them ... ",
                call. = FALSE)
        if(verbose){
            print(rmb[base_mm,])
        }
        rmb <- rmb[!base_mm,]
    }
    ############################################################################
    # extract position data
    strand <- rmb$strand
    # pos <- rmb$modStart
    # pos[strand == "-"] <- rmb$modEnd[strand == "-"]
    pos <- rmb$modEnd
    # extract modification information
    mod_id <- rmb$modId
    mod_type <- .fix_non_standard_mod_types(as.character(rmb$modType))
    mod_name <- rmb$modName
    # extract intermediate transcript data
    chromosome <- rmb$chromosome
    gene_name <- rmb$geneName
    if(any(is.na(gene_name)) || any(is.null(gene_name))){
        stop(".")
    }
    # extract references
    f_pmid <- rmb$pubmedIds != "" & rmb$pubmedIds != "-"
    f_ref <- rmb$supportList != "" & rmb$supportList != "-"
    ref_type <- list(IRanges::CharacterList(as.list(rep("RMBASE_MOD_ID",
                                                        nrow(rmb)))),
                     IRanges::CharacterList(as.list(rep("PMID",
                                                        nrow(rmb)))),
                     IRanges::CharacterList(as.list(rep("RMBase_REF",
                                                        nrow(rmb)))))
    ref_type[[2]][!f_pmid] <- ""
    ref_type[[3]][!f_ref] <- ""
    ref_type <- do.call(pc,ref_type)
    ref_type <- ref_type[ref_type != ""]
    reference <- list(relist(as.character(rmb$modId),
                             IRanges::PartitioningByWidth(lengths(rmb$modId))),
                      relist(as.character(rmb$pubmedIds),
                             IRanges::PartitioningByWidth(lengths(rmb$pubmedIds))),
                      relist(as.character(rmb$supportList),
                             IRanges::PartitioningByWidth(lengths(rmb$supportList))))
    reference[[2]][!f_pmid] <- ""
    reference[[3]][!f_ref] <- ""
    reference <- do.call(pc,reference)
    reference <- reference[reference != ""]
    # assemble metadata columns
    gene_name <-
        IRanges::CharacterList(strsplit(as.character(gene_name),","))
    mcols <- DataFrame(mod_id = mod_id,
                       mod_type = mod_type,
                       mod_name = mod_name,
                       chromosome = chromosome,
                       gene_name = gene_name,
                       score = rmb$score,
                       ref_type = ref_type,
                       ref = reference)
    # create GRanges result
    ans <- GenomicRanges::GRanges(seqnames = chromosome,
                                  ranges = IRanges::IRanges(pos, width = 1L),
                                  strand = strand,
                                  mcols)
    #
    ans
}

.get_RMBase_data <- function(files, verbose = FALSE){
    grl <- lapply(files,
                  function(file){
                      rmb <- .read_RMBase_file(file)
                      .extract_GRanges_from_RMBase(rmb, verbose = verbose)
                  })
    grl
}

.gene_names_to_chromosome <- function(gr){
    chromosome <- mcols(gr)$gene_name
    GenomicRanges::GRanges(seqnames = chromosome,
                           ranges = IRanges::ranges(gr),
                           strand = strand(gr),
                           mcols(gr))
}



.check_tx_sequences <- function(tx, sequences){
    if(is.null(sequences)){
        return()
    }
    if(!is(sequences,"DNAStringSet") & !is(sequences,"RNAStringSet")){
        stop("'sequences' must be a DNA/RNAStringSet.", call. = FALSE)
    }
    if(is.null(names(sequences)) || !all(names(sequences) == names(tx) )){
        stop("'sequences' must be named and names have to match 'tx'",
             call. = FALSE)
    }
    if(!all(sum(width(tx)) == width(sequences))){
        stop("'sequences' and 'tx' must have the same dimensions, e.g. the ",
             "length of sequences and annotation must match.",
             call. = FALSE)
    }
}

#' @rdname makeEpiTxDbFromRMBase
#' @export
getRMBaseDataAsGRanges <- function(files, verbose = FALSE){
    # Input check
    if(!(is.logical(verbose) && length(verbose) == 1L)){
        stop("'verbose' must be TRUE or FALSE.", call. = FALSE)
    }
    #
    message("Assembling RMBase data ...")
    # getting raw data from RMBase files
    grl <- .get_RMBase_data(files, verbose = verbose)
    gr <- unlist(GenomicRanges::GRangesList(grl))
    gr
}

.remove_incompatible_mod_by_seq <- function(gr, seq){
    seq <- as(seq,"RNAStringSet")
    colnames(mcols(gr)) <- gsub("mod_type","mod",colnames(mcols(gr)))
    # do plus and minus strand separatly, since removeIncompatibleModifications
    # only accepts plus strand
    gr_plus <-
        Modstrings::removeIncompatibleModifications(gr[strand(gr) == "+"], seq)
    gr_minus <- gr[strand(gr) == "-"]
    strand(gr_minus) <- "+"
    gr_minus <-
        Modstrings::removeIncompatibleModifications(gr_minus,
                                                    Biostrings::complement(seq))
    strand(gr_minus) <- "-"
    gr <- c(gr_plus,gr_minus)
    gr <- gr[order(seqnames(gr),start(gr),strand(gr))]
    colnames(mcols(gr)) <- gsub("^mod$","mod_type",colnames(mcols(gr)))
    gr
}

.add_sequence_check_to_metadata <- function(metadata){
    metadata <- rbind(metadata,
                      data.frame(name = "Checked against sequence",
                                 value = "Yes",
                                 check.names = FALSE,
                                 stringsAsFactors = FALSE))
    metadata
}

#' @rdname makeEpiTxDbFromRMBase
#' @export
makeEpiTxDbFromRMBaseFiles <- function(files, tx = NULL, sequences = NULL,
                                       metadata = NULL, reassign.ids = FALSE,
                                       verbose = FALSE){
    # Input check
    if(!(is.logical(reassign.ids) && length(reassign.ids) == 1L)){
        stop("'reassign.ids' must be TRUE or FALSE.", call. = FALSE)
    }
    if(!(is.logical(verbose) && length(verbose) == 1L)){
        stop("'verbose' must be TRUE or FALSE.", call. = FALSE)
    }
    #
    gr <- getRMBaseDataAsGRanges(files, verbose = verbose)
    if(!is.null(tx)){
        sl <- Seqinfo::seqlevels(tx)
    } else if(!is.null(sequences)) {
        sl <- names(sequences)
    } else {
        sl <- Seqinfo::seqlevels(gr)
    }
    chromosome <- .simplify_chromosome_identifiers(seqnames(gr), sl)
    gr <- GenomicRanges::GRanges(seqnames = chromosome,
                                 ranges = IRanges::ranges(gr),
                                 strand = BiocGenerics::strand(gr),
                                 S4Vectors::mcols(gr))
    # shift position to transcript coordinates
    if(!is.null(tx)){
        tx <- .norm_tx(tx)
        .check_tx_sequences(tx, sequences)
        message("Shifting RMBase result's coordinates based on transcript ",
                "data ...")
        gr <- shiftGenomicToTranscript(gr, tx)
        f <- !duplicated(paste0(as.character(gr),"-",mcols(gr)$mod_type))
        gr <- gr[f]
    }
    # check if all the referenced modification match their originating base
    if(!is.null(sequences)){
        # remove any positions which do not match the originating base
        gr <- .remove_incompatible_mod_by_seq(gr, sequences)
        metadata <- .add_sequence_check_to_metadata(metadata)
    }
    #
    mcols(gr)$mod_id <- seq_along(gr)
    makeEpiTxDbFromGRanges(gr, metadata = metadata, reassign.ids = reassign.ids)
}


#' @rdname makeEpiTxDbFromRMBase
#' @importFrom curl curl
#' @export
listAvailableOrganismsFromRMBase <- function(){
    # con <- curl::curl(EPITXDB_RMBASE_URL_JSON)
    # downloadData <- jsonlite::fromJSON(con)
    # rmbase_data <- downloadData$data[,1:3]
    # colnames(rmbase_data) <- c("species","mod","dataSet")
    # re <- rex("download/",
    #           capture(alphas, name="organism"),
    #           "/RMBase_",
    #           capture(zero_or_more(any,type="lazy"), name="genome"),
    #           "_",anything,"/zip/",
    #           capture(zero_or_more(any,type="lazy"), name="file"),"'")
    # rmbase_data <- cbind(rmbase_data,re_matches(downloadData$data$Download,re))
    rmbase_data <- NULL
    utils::data("rmbase_data", envir = environment(), package = "EpiTxDb")
    as.character(unique(rmbase_data$organism))
}

#' @importFrom curl curl
#' @import rex
.get_RMBase_files <- function(organism){
    rmbase_data <- NULL
    utils::data("rmbase_data", envir = environment(), package = "EpiTxDb")
    as.character(rmbase_data[rmbase_data$organism == organism,]$file)
}

.get_RMBase_genomes <- function(files){
    unique(vapply(strsplit(files,"_"),"[",character(1),2L))
}

.listAvailableGenomesFromRMBase <- function(organism){
    # files <- .get_RMBase_files(organism)
    # .get_RMBase_genomes(files)
    rmbase_data <- NULL
    utils::data("rmbase_data", envir = environment(), package = "EpiTxDb")
    as.character(unique(rmbase_data[rmbase_data$organism == organism,]$genome))
}

#' @rdname makeEpiTxDbFromRMBase
#' @export
listAvailableGenomesFromRMBase <- function(organism){
    if(!(organism %in% listAvailableOrganismsFromRMBase())){
        stop("'organism' must be a valid organism from ",
             "listAvailableOrganismsfromRMBase()")
    }
    .listAvailableGenomesFromRMBase(organism)
}

.get_RMBase_mod <- function(organism, genome){
    # f_genome <- vapply(strsplit(files,"_"),"[",character(1),2L) == genome
    # ans <- unique(vapply(strsplit(files[f_genome],"_"),"[",character(1),4L))
    # ans[!grepl("mod",ans)]
    rmbase_data <- NULL
    utils::data("rmbase_data", envir = environment(), package = "EpiTxDb")
    as.character(rmbase_data[rmbase_data$organism == organism &
                                 rmbase_data$genome == genome,]$mod)
}

#' @rdname makeEpiTxDbFromRMBase
#' @export
listAvailableModFromRMBase <- function(organism, genome){
    if(!(organism %in% listAvailableOrganismsFromRMBase())){
        stop("'organism' must be a valid organism from ",
             "listAvailableOrganismsfromRMBase()")
    }
    if(!(genome %in% listAvailableGenomesFromRMBase(organism))){
        stop("'genome' must be a valid genome for the given 'organism' from ",
             "listAvailableGenomesFromRMBase()")
    }
    .get_RMBase_mod(organism, genome)
}

context("makeEpiTxDbFromRMBase")
test_that("makeEpiTxDbFromRMBase:",{
    #
    expect_type(listAvailableOrganismsFromRMBase(),"character")
    expect_error(listAvailableGenomesFromRMBase("human1"),
                 "'organism' must be a valid organism")
    expect_type(listAvailableGenomesFromRMBase("human"),"character")
    expect_error(listAvailableModFromRMBase("human1"),
                 "'organism' must be a valid organism")
    expect_error(listAvailableModFromRMBase("human"),
                 'argument "genome" is missing')
    expect_error(listAvailableModFromRMBase("human","hg191"),
                 "'genome' must be a valid genome for the")
    expect_type(listAvailableModFromRMBase("human","hg19"),"character")
    #
    # files <- system.file("extdata", "RMBase_testdata.txt.gz",
    #                      package = "EpiTxDb")
    # expect_warning(etdb <- makeEpiTxDbFromRMBaseFiles(files))
    # expect_s4_class(etdb,"EpiTxDb")
    # expect_s3_class(metadata(etdb),"data.frame")
    # #
    # actual <- modifications(etdb)
    # expect_s4_class(actual,"GRanges")
    # expect_length(actual,512L)
    # dbDisconnect(etdb$conn)
})

#' @include AllGenerics.R
NULL

#' @name positionSequence
#'
#' @title Generate integer sequences from position information of \code{Ranges}
#'
#' @description \code{positionSequence} generates sequences of integer values
#' along the range information of \code{x}. This can be used for navigating
#' specific positions on a range information.
#'
#' @param x a \code{Ranges} object, like a
#'   \code{\link[GenomicRanges:GRanges-class]{GRanges}} or
#'   \code{\link[IRanges:IRanges-class]{IRanges}}, or a \code{RangesList}
#'   object, like a \code{\link[GenomicRanges:GRangesList-class]{GRangesList}}
#'   or \code{\link[IRanges:IRangesList-class]{IRangesList}}
#' @param order \code{TRUE} or \code{FALSE}: Should the position be ordered?
#'   (default: \code{order = FALSE})
#' @param decreasing \code{TRUE} or \code{FALSE}: If \code{order = TRUE} Should
#'   the position be ordered in a decreasing order? (default: \code{order =
#'   FALSE})
#'
#' @return a \code{integer} vector if x is a
#'   \code{\link[GenomicRanges:GRanges-class]{GRanges}} object and a
#'   \code{IntegerList} if x is a
#'   \code{\link[GenomicRanges:GRangesList-class]{GRangesList}}
#'
#' @examples
#' library(GenomicRanges)
#' # Returns an integer vector
#' gr <- GRanges("chr1:1-5:+")
#' positionSequence(gr)
#' gr2 <- GRanges("chr1:1-5:-")
#' positionSequence(gr)
#' # returns an IntegerList
#' grl <- GRangesList("1" = gr,"2" = gr,"3" = gr2) # must be named
#' positionSequence(grl)
NULL

# per element of GRangesList unique
.get_strand_u_GRangesList <- function(grl){
    strand_u <- unique(IRanges::CharacterList(strand(grl)))
    ans <- unlist(strand_u)
    if(length(strand_u) != length(ans)){
        stop("Non unqiue strands per GRangesList element.")
    }
    ans
}

# Vectorize version of seq specific for start/ends from a RangesList
.seqs_rl_strand <- function(rl, order = FALSE, decreasing = FALSE){
    strand_u <- .get_strand_u_GRangesList(rl)
    strand_minus <- strand_u == "-"
    ansP <- .seqs_rl_by(rl[!strand_minus])
    ansM <- .seqs_rl_by(rl[strand_minus], by = -1L)
    if(order){
        ansM <- ansM[IRanges::IntegerList(lapply(ansM, order,
                                                 decreasing = decreasing))]
    }
    ans <- c(ansP, ansM)
    ans[match(names(rl),names(ans))]
}

# Vectorize version of seq specific for start/ends from a RangesList with a by
# option
.seqs_rl_by <- function(rl, by = 1L){
    starts <- unlist(start(rl))
    ends <- unlist(end(rl))
    .seqs_l_by(starts, ends, by)
}

#' @importFrom IRanges PartitioningByWidth PartitioningByEnd
#' @importClassesFrom IRanges PartitioningByWidth PartitioningByEnd
# Vectorize version of seq using to input lists
.seqs_l_by <- function(from, to, by = 1L){
    if(is.null(names(from)) || is.null(names(to))){
        stop("Inputs must be named.")
    }
    if(length(from) != length(to)){
        stop("Inputs must have the same length.")
    }
    if(by == 0L){
        stop("by cannot be zero.")
    }
    if(any(names(from) != names(to))){
        stop("Unmatched names.")
    }
    if(by < 0L){ # switch from/to around if negative
        tmp <- to
        to <- from
        from <- tmp
        rm(tmp)
    }
    ans <- Map(
        function(f,t){
            seq.int(f,t,by)
        },
        from,
        to)
    ans <- IRanges::IntegerList(ans)
    partitioning <- IRanges::PartitioningByEnd(ans)
    width_x <- IRanges::IntegerList(split(width(partitioning),
                                          names(partitioning)))
    m <- match(unique(names(from)),names(width_x))
    width_x <- width_x[m]
    width_ans <- sum(width_x)
    ans <- relist(unname(unlist(ans)),
                  IRanges::PartitioningByWidth(width_ans,
                                               names = names(width_ans)))
    unique(ans)
}

#' @rdname positionSequence
#' @export
setMethod("positionSequence","Ranges",
    function(x, order = FALSE, decreasing = FALSE){
        class <- paste0(class(x),"List")
        x <- do.call(class,list(x))
        names(x) <- "tmp"
        ans <- .seqs_rl_strand(x, order = order, decreasing = decreasing)
        ans[["tmp"]]
    }
)

#' @rdname positionSequence
#' @export
setMethod("positionSequence","RangesList",
    function(x, order = FALSE, decreasing = FALSE){
        .seqs_rl_strand(x, order = order, decreasing = decreasing)
    })

#' @rdname positionSequence
#' @export
setMethod("as.integer","Ranges",
    function(x){
        positionSequence(x)
    })

setAs("RangesList","IntegerList",
    function(from){
        positionSequence(from)
    })

# rescale ----------------------------------------------------------------------

#' @name rescale
#'
#' @title Rescaling \code{Ranges} object
#'
#' @description
#' \code{rescale()} rescales \code{IRanges}, \code{GRanges}, \code{IRangesList}
#' and \code{GRangesList} by using minima and maxima derived from \code{to} and
#' \code{from}.
#'
#' @param x a \code{IRanges}, \code{GRanges}, \code{IRangesList} and
#'   \code{GRangesList} object
#' @param to,from an \code{IRanges} object, a \code{character} vector coercible
#'   to \code{IRanges} or a \code{integer} vector parallel to \code{x} or with
#'   \code{length = 1L}.
#'
#' @return an object of the same type and dimensions as \code{x}
#'
#' @author H. Pagès, F. Ernst
#'
#' @seealso
#' \code{\link[IRanges:IRanges-constructor]{IRanges}} for details on
#' \code{character} vectors coercible to \code{IRanges}.
#'
#' @examples
#' x <- IRanges("5-10")
#' # widen the ranges
#' rescale(x, 100, 10)
#' # widen and shift
#' rescale(x, "31-60", "5-14")
NULL

.zoom0 <- function(x, z = 1)
{
    stopifnot(is(x, "Ranges"), is.numeric(z))
    if (length(z) > length(x) && length(z) != 1L)
        stop("'z' is longer than 'x'")
    if (anyNA(z) || min(z) <= -1L)
        stop("'z' contains NAs and/or negative values")
    new_start <- as.integer(start(x) * z)
    new_width <- as.integer(width(x) * z)
    BiocGenerics:::replaceSlots(x, start=new_start, width=new_width)
}

.normarg_scale <- function(scale)
{
    if (is(scale, "IRanges"))
        return(scale)
    if (is.numeric(scale))
        return(IRanges(1L, scale))
    as(scale, "IRanges")
}

#' @rdname rescale
#' @export
setMethod("rescale","IRanges",
    function(x, to = 1L, from = 1L){
        to <- .normarg_scale(to)
        from <- .normarg_scale(from)
        ans <- shift(x, -start(from))
        ans <- .zoom0(ans, width(to) / width(from))
        shift(ans, start(to))
    }
)

#' @rdname rescale
#' @export
setMethod("rescale","IRangesList",
    function(x, to = 1L, from = 1L){
        to <- .normarg_scale(to)
        from <- .normarg_scale(from)
        ans <- shift(x, -start(from))
        ans <- Map(.zoom0,ans,width(to) / width(from))
        ans <- IRanges::IRangesList(ans)
        shift(ans, start(to))
    }
)

#' @rdname rescale
#' @export
setMethod("rescale","GRanges",
    function(x, to = 1L, from = 1L){
        ranges(x) <- rescale(ranges(x), to = to, from = from)
        x
    }
)

#' @rdname rescale
#' @export
setMethod("rescale","GRangesList",
    function(x, to = 1L, from = 1L){
        ranges(x) <- rescale(ranges(x), to = to, from = from)
        x
    }
)

#' @include EpiTxDb-class.R
NULL

#' @name shiftGenomicToTranscript
#'
#' @title Shift \code{GRanges} coordinates based on another \code{GRanges}
#' object
#'
#' @description
#' \code{shiftGenomicToTranscript} shifts positions of a
#' \code{\link[GenomicRanges:GRanges-class]{GRanges}}
#' object based on coordinates of another \code{GRanges} object. The most common
#' application is to shift genomic coordinates to transcript coordinates, which
#' is reflected in the name. \code{shiftTranscriptToGenomic} implements the
#' reverse operation.
#'
#' Matches are determined by
#' \code{\link[GenomicRanges:findOverlaps-methods]{findOverlaps}} for
#' \code{shiftGenomicToTranscript} and by
#' \code{\link[S4Vectors:Vector-comparison]{findMatches}} for
#' \code{shiftTranscriptToGenomic} using the \code{seqnames} of the
#' \code{subject} and the \code{names} of \code{tx}.
#'
#' @param subject a \code{\link[GenomicRanges:GRanges-class]{GRanges}} or
#'   \code{\link[GenomicRanges:GRangesList-class]{GRangesList}} object
#' @param tx a named \code{\link[GenomicRanges:GRangesList-class]{GRangesList}}
#'   object.
#'
#' @return a \code{\link[GenomicRanges:GRanges-class]{GRanges}} or
#'   \code{\link[GenomicRanges:GRangesList-class]{GRangesList}} object depending
#'   on the type of \code{subject}
#'
#' @examples
#' library(GenomicRanges)
#' # Construct some example data
#' subject1 <- GRanges("chr1", IRanges(3, 6),
#'                     strand = "+")
#' subject2 <- GRanges("chr1", IRanges(c(17,23), width=3),
#'                     strand = c("+","-"))
#' subject3 <- GRanges("chr2", IRanges(c(51, 54), c(53, 59)),
#'                     strand = "-")
#' subject <- GRangesList(a=subject1, b=subject2, c=subject3)
#' tx1 <- GRanges("chr1", IRanges(1, 40),
#'                strand="+")
#' tx2 <- GRanges("chr1", IRanges(10, 30),
#'                strand="+")
#' tx3 <- GRanges("chr2", IRanges(50, 60),
#'                strand="-")
#' tx <- GRangesList(a=tx1, b=tx2, c=tx3)
#'
#' # shift to transcript coordinates. Since the third subject does not have
#' # a match in tx it is dropped with a warning
#' shifted_grl <- shiftGenomicToTranscript(subject,tx)
#'
#' # ... and back
#' shifted_grl2 <- shiftTranscriptToGenomic(shifted_grl,tx)
#'
#' # comparison of ranges work. However the seqlevels differ
#' ranges(shifted_grl2) == ranges(subject[list(1,c(1,1),c(1,2))])
NULL

# helper functions -------------------------------------------------------------

.norm_tx <- function(tx){
    if(missing(tx) || !is(tx,"GRangesList")){
        stop("'tx' must be a named 'GRangesList'.", call. = FALSE)
    }
    if(is.null(names(tx))){
        stop("'tx' must be a named 'GRangesList'.", call. = FALSE)
    }
    if(anyDuplicated(names(tx))){
        stop("names of 'tx' must be unique.", call. = FALSE)
    }
    tx
}

# shiftTranscriptToGenomic ----------------------------------------------------

#' @importFrom S4Vectors findMatches
.hitsFUN_TX_TO_GENOME <- function(subject,tx){
    FUN <- S4Vectors::findMatches
    seqnames <- seqnames(subject)
    if(is(seqnames,"RleList")){
        seqnames <- as(seqnames,"CharacterList")
        hits <- suppressWarnings(lapply(seqnames,function(s){FUN(names(tx),s)}))
    } else {
        seqnames <- as.character(seqnames)
        hits <- suppressWarnings(FUN(names(tx), seqnames))
    }
    hits
}

#' @importFrom GenomicRanges findOverlaps
.hitsFUN_GENOME_TO_TX <- function(subject,tx){
    hits <- suppressWarnings(GenomicRanges::findOverlaps(tx,subject))
    hits
}

.compareFUN_TX_TO_GENOME <- function(y, z, seqs){
    seqs[y][IRanges::IntegerList(as.list(z))]
}

.compareFUN_GENOME_TO_TX <- function(y, z, seqs){
    which(seqs[y] == z)
}

.compareFUN_TX_TO_GENOME_List <- function(y, z, seqs){
    seqs[y][z]
}

.compareFUN_GENOME_TO_TX_List <- function(y, z, seqs){
    ans <- Map(
        function(a,b){
            match(b,seqs[[a]],nomatch = 0L)
        },
        y,
        z)
    IRanges::IntegerList(ans)
}

.resultFUN_TX_TO_GENOME <- function(tx_hits, subject_hits, starts, ends){
    # drop problematic subjects
    invalid_pos <- lengths(starts) < 1L | lengths(ends) < 1L
    if(any(invalid_pos)){
        warning("Dropping elements of 'subject' which could not be ",
                "repositioned.",
                call. = FALSE)
    }
    tx_hits <- tx_hits[!invalid_pos]
    subject_hits <- subject_hits[!invalid_pos]
    starts <- starts[!invalid_pos]
    ends <- ends[!invalid_pos]
    # transfer subject information
    mcols(subject_hits)$seq_start <- start(subject_hits)
    mcols(subject_hits)$seq_end <- end(subject_hits)
    mcols(subject_hits)$seq_strand <- strand(subject_hits)
    mcols(subject_hits)$seq_name <- seqnames(subject_hits)
    # prepare reformat of GRanges results
    seqnames <- unlist(unique(seqnames(tx_hits)),use.names = FALSE)
    strand <- unlist(unique(strand(tx_hits)),use.names = FALSE)
    mcols <- mcols(subject_hits)
    tmp <- starts
    starts[strand == "-"] <- ends[strand == "-"]
    ends[strand == "-"] <- tmp[strand == "-"]
    ranges <- IRanges::IRanges(start = unlist(starts), end = unlist(ends))
    # reformat GRanges results
    ans <- GenomicRanges::GRanges(seqnames = seqnames,
                                  ranges = ranges,
                                  strand = strand,
                                  mcols)
    non_dup <- !duplicated(paste0(as.character(ans),"_",mcols(ans)$mod_type,
                                  "_",mcols(ans)$seq_name))
    ans <- ans[non_dup]
    ans
}

.resultFUN_GENOME_TO_TX <- function(tx_hits, subject_hits, starts, ends){
    # drop problematic subjects
    invalid_pos <- lengths(starts) < 1L | lengths(ends) < 1L|
        strand(subject_hits) != unlist(unique(strand(tx_hits)))
    if(any(invalid_pos)){
        warning("Dropping elements of 'subject' which could not be ",
                "repositioned.",
                call. = FALSE)
    }
    tx_hits <- tx_hits[!invalid_pos]
    subject_hits <- subject_hits[!invalid_pos]
    starts <- starts[!invalid_pos]
    ends <- ends[!invalid_pos]
    # transfer subject information
    mcols(subject_hits)$seq_start <- start(subject_hits)
    mcols(subject_hits)$seq_end <- end(subject_hits)
    mcols(subject_hits)$seq_strand <- strand(subject_hits)
    mcols(subject_hits)$seq_name <- seqnames(subject_hits)
    # reformat GRanges results
    seqnames <- names(tx_hits)
    mcols <- mcols(subject_hits)
    tmp <- starts
    starts[strand(subject_hits) == "-"] <- ends[strand(subject_hits) == "-"]
    ends[strand(subject_hits) == "-"] <- tmp[strand(subject_hits) == "-"]
    ranges <- IRanges::IRanges(start = unlist(starts), end = unlist(ends))
    #
    ans <- GenomicRanges::GRanges(seqnames = seqnames,
                                  ranges = ranges,
                                  strand = "*",
                                  mcols)
    ans
}

.resultFUN_TX_TO_GENOME_List <- function(tx_hits, subject_hits, starts, ends){
    # drop problematic subjects
    invalid_pos <- starts == 0L | ends == 0L
    if(any(any(invalid_pos))){
        warning("Dropping elements of 'subject' which could not be ",
                "repositioned.",
                call. = FALSE)
    }
    tx_hits <- tx_hits[any(!invalid_pos)]
    subject_hits <- subject_hits[!invalid_pos]
    starts <- starts[!invalid_pos]
    ends <- ends[!invalid_pos]
    tx_hits <- tx_hits[lengths(tx_hits) > 0L]
    subject_hits <- subject_hits[lengths(subject_hits) > 0L]
    starts <- starts[lengths(starts) > 0L]
    ends <- ends[lengths(ends) > 0L]
    # transfer subject information
    mcols(subject_hits,level="within")[,"seq_start"] <- start(subject_hits)
    mcols(subject_hits,level="within")[,"seq_end"] <- end(subject_hits)
    mcols(subject_hits,level="within")[,"seq_strand"] <- strand(subject_hits)
    mcols(subject_hits,level="within")[,"seq_name"] <- seqnames(subject_hits)
    # prepare reformat of GRanges results
    seqnames <- rep(unlist(unique(seqnames(tx_hits)),use.names = FALSE),
                    lengths(subject_hits))
    strand <- rep(unlist(unique(strand(tx_hits)),use.names = FALSE),
                  lengths(subject_hits))
    mcols <- unlist(mcols(subject_hits,level="within"), use.names = FALSE)
    u_starts <- unlist(starts, use.names = FALSE)
    u_ends <- unlist(ends, use.names = FALSE)
    u_tmp <- u_starts
    u_starts[strand == "-"] <- u_ends[strand == "-"]
    u_ends[strand == "-"] <- u_tmp[strand == "-"]
    ranges <- IRanges::IRanges(start = u_starts, end = u_ends)
    # reformat GRanges results
    ans <- GenomicRanges::GRanges(seqnames = seqnames,
                                  ranges = ranges,
                                  strand = strand,
                                  mcols)
    ans <- relist(ans,subject_hits)
    names(ans) <- names(tx_hits)
    ans
}

.resultFUN_GENOME_TO_TX_List <- function(tx_hits, subject_hits, starts, ends){
    # drop problematic subjects
    invalid_pos <- starts == 0L | ends == 0L |
        unname(as(strand(subject_hits) != unique(strand(tx_hits)),"LogicalList"))
    if(any(any(invalid_pos))){
        warning("Dropping elements of 'subject' which could not be ",
                "repositioned.",
                call. = FALSE)
    }
    tx_hits <- tx_hits[any(!invalid_pos)]
    subject_hits <- subject_hits[!invalid_pos]
    starts <- starts[!invalid_pos]
    ends <- ends[!invalid_pos]
    tx_hits <- tx_hits[lengths(tx_hits) > 0L]
    subject_hits <- subject_hits[lengths(subject_hits) > 0L]
    starts <- starts[lengths(starts) > 0L]
    ends <- ends[lengths(ends) > 0L]
    # transfer subject information
    mcols(subject_hits,level="within")[,"seq_start"] <- start(subject_hits)
    mcols(subject_hits,level="within")[,"seq_end"] <- end(subject_hits)
    mcols(subject_hits,level="within")[,"seq_strand"] <- strand(subject_hits)
    mcols(subject_hits,level="within")[,"seq_name"] <- seqnames(subject_hits)
    # reformat GRanges results
    seqnames <- rep(names(tx_hits),lengths(subject_hits))
    mcols <- unlist(mcols(subject_hits,level="within"), use.names = FALSE)
    # switch starts ends depending on strand
    strand <- unlist(starts > ends, use.names = FALSE)
    u_starts <- unlist(starts, use.names = FALSE)
    u_ends <- unlist(ends, use.names = FALSE)
    u_tmp <- u_starts
    u_starts[strand] <- u_ends[strand]
    u_ends[strand] <- u_tmp[strand]
    ranges <- IRanges::IRanges(start = u_starts, end = u_ends)
    #
    ans <- GenomicRanges::GRanges(seqnames = seqnames,
                                  ranges = ranges,
                                  strand = "*",
                                  mcols)
    ans <- relist(ans,subject_hits)
    ans
}

.fixed_queryHits <- function(hits){
    if(is(hits,"list")){
        qh <- as(lapply(hits,queryHits),"IntegerList")
        qh <- unlist(unique(qh),use.names = FALSE)
    } else {
        qh <- queryHits(hits)
    }
    qh
}
.fixed_subjectHits <- function(hits){
    if(is(hits,"list")){
        sh <- as(lapply(hits,subjectHits),"IntegerList")

    } else {
        sh <- subjectHits(hits)
    }
    sh
}

.non_hits <- function(subject, sh){
    if((is(subject,"GRanges") | is(subject,"GRangesList")) & is.numeric(sh)){
        return(seq_along(subject) %in% sh)
    } else if(is(subject,"GRangesList") & is(sh,"IntegerList")) {
        return(names(subject) %in% names(sh))
    }
    stop(".")
}

.shift <- function(subject, tx, seqs, .hitsFUN, .compareFUN, .resultFUN){
    # match seqnames of subject and names of tx
    hits <- .hitsFUN(subject, tx)
    if(length(hits) == 0L){
        stop("No 'subject'/'tx' matches found.",
             call. = FALSE)
    }
    # assemble results vectors
    qh <- .fixed_queryHits(hits)
    sh <- .fixed_subjectHits(hits)
    f <- .non_hits(subject, sh)
    if(!all(f)){
        warning("Coordinates for ",sum(!f)," ranges of 'subject' not found: '",
                paste(unique(as.character(head(subject[!f], 10L))),
                      collapse = "','"),
                "'",ifelse(sum(!f) > 10L, " and more ..",""),".",
                call. = FALSE)
    }
    tx_hits <- tx[qh]
    subject_hits <- subject[sh]
    # reposition modification on transcript
    chunk_size <- 10^5
    chunks_n <- ceiling(length(qh) / chunk_size)
    f <- factor(unlist(Map(rep.int,seq_len(chunks_n),chunk_size)))
    f <- f[seq_along(qh)]
    # reposition starts and ends on sequence of positions
    starts <- mapply(.compareFUN,
                     split(qh, f),
                     split(start(subject_hits), f),
                     MoreArgs = list(seqs = seqs),
                     SIMPLIFY = FALSE)
    starts <- do.call(c,unname(starts))
    ends <- mapply(.compareFUN,
                   split(qh, f),
                   split(end(subject_hits), f),
                   MoreArgs = list(seqs = seqs),
                   SIMPLIFY = FALSE)
    ends <- do.call(c,unname(ends))
    ans <- .resultFUN(tx_hits, subject_hits, starts, ends)
    ans
}


#' @rdname shiftGenomicToTranscript
#' @export
setMethod("shiftTranscriptToGenomic", c("GRanges","GRangesList"),
    function(subject, tx){
        tx <- .norm_tx(tx)
        seqs <- positionSequence(tx)
        .shift(subject, tx, seqs, .hitsFUN_TX_TO_GENOME,
               .compareFUN_TX_TO_GENOME, .resultFUN_TX_TO_GENOME)
    }
)

#' @rdname shiftGenomicToTranscript
#' @export
setMethod("shiftTranscriptToGenomic", c("GRangesList","GRangesList"),
    function(subject, tx){
        tx <- .norm_tx(tx)
        seqs <- positionSequence(tx)
        ans <- .shift(subject, tx, seqs, .hitsFUN_TX_TO_GENOME,
                      .compareFUN_TX_TO_GENOME_List,
                      .resultFUN_TX_TO_GENOME_List)
        ans
    }
)


# shiftGenomicToTranscript ----------------------------------------------------

#' @rdname shiftGenomicToTranscript
#' @export
setMethod("shiftGenomicToTranscript", c("GRanges","GRangesList"),
    function(subject, tx){
        tx <- .norm_tx(tx)
        seqs <- positionSequence(tx)
        .shift(subject, tx, seqs, .hitsFUN_GENOME_TO_TX,
               .compareFUN_GENOME_TO_TX,
               .resultFUN_GENOME_TO_TX)
    }
)

#' @rdname shiftGenomicToTranscript
#' @export
setMethod("shiftGenomicToTranscript", c("GRangesList","GRangesList"),
    function(subject, tx){
        tx <- .norm_tx(tx)
        seqs <- positionSequence(tx)
        ans <- .shift(subject, tx, seqs, .hitsFUN_GENOME_TO_TX,
                      .compareFUN_GENOME_TO_TX_List,
                      .resultFUN_GENOME_TO_TX_List)
        ans
    }
)

#' @include EpiTxDb-class.R
#' @include makeEpiTxDb.R
NULL

#' @name makeEpiTxDbFromGRanges
#'
#' @title Create a \code{EpiTxDb} object from a \code{GRanges} object
#'
#' @description
#' \code{makeEpiTxDbFromGRanges} extracts informations from a
#' \code{\link[GenomicRanges:GRanges-class]{GRanges}} object. The following
#' metadata columns can be used:
#' \itemize{
#' \item \code{mod_id}, \code{mod_type}, \code{mod_name} and \code{tx_ensembl}.
#' The first three are mandatory, whereas \code{tx_ensembl} is optional.
#' \item \code{rx_genename}, \code{rx_rank}, \code{rx_ensembl},
#' \code{rx_ensembltrans} and \code{rx_entrezid}
#' \item \code{spec_type}, \code{spec_genename}, \code{spec_ensembl},
#' \code{spec_ensembltrans} and \code{spec_entrezid}
#' \item \code{ref_type} and \code{ref}
#' }
#' ... and passed on the \code{\link[=makeEpiTxDb]{makeEpiTxDb}}.
#'
#' @param gr A \code{\link[GenomicRanges:GRanges-class]{GRanges}} object, which
#' contains at least the mandatory columns.
#' @param metadata A 2-column \code{data.frame} containing meta information to
#' be included in the \code{EpiTxDb} object. This \code{data.frame} is just
#' passed to \code{\link[=makeEpiTxDb]{makeEpiTxDb}}. See
#' \code{\link[=makeEpiTxDb]{makeEpiTxDb}} for more information about the format
#' of metadata. (default: \code{metadata = NULL})
#' @param reassign.ids = FALSE
#'
#' @return a \code{EpiTxDb} object.
#'
#' @export
#'
#' @examples
#' library(GenomicRanges)
#' gr <- GRanges(seqnames = "test",
#'               ranges = IRanges::IRanges(1,1),
#'               strand = "+",
#'               DataFrame(mod_id = 1L,
#'                         mod_type = "Am",
#'                         mod_name = "Am_1"))
#' etdb <- makeEpiTxDbFromGRanges(gr)
NULL

# makeEpiTxDbfromGRanges -------------------------------------------------------

# extract modification table
.get_gr_modifications <- function(gr){
    modifications <- list(mod_id = mcols(gr)$mod_id,
                          mod_type = mcols(gr)$mod_type,
                          mod_name = mcols(gr)$mod_name,
                          mod_start = start(gr),
                          mod_end = end(gr),
                          mod_strand = strand(gr),
                          sn_id = as.integer(seqnames(gr)),
                          sn_name= seqnames(gr))
    modifications <- modifications[!vapply(modifications, is.null, logical(1))]
    if(length(modifications) > 1L){
        modifications <- data.frame(modifications, check.names = FALSE,
                                    stringsAsFactors = FALSE)
    } else {
        stop("Couldn't extranct modification information from 'GRanges' object.",
             call. = FALSE)
    }
    modifications
}

# extract reactions table
.get_gr_reactions <- function(gr){
    reactions <- list(mod_id = mcols(gr)$mod_id,
                      rx_genename = mcols(gr)$rx_genename,
                      rx_rank = mcols(gr)$rx_rank,
                      rx_ensembl = mcols(gr)$rx_ensembl,
                      rx_ensembltrans = mcols(gr)$rx_ensembltrans,
                      rx_entrezid = mcols(gr)$rx_entrezid)
    reactions <- reactions[!vapply(reactions, is.null, logical(1))]
    if(length(reactions) > 1L){
        reactions <- data.frame(reactions, check.names = FALSE,
                                stringsAsFactors = FALSE)
    } else {
        reactions <- NULL
    }
    reactions
}

# extract specifiers table
.get_gr_specifiers <- function(gr){
    specifiers <- list(mod_id = mcols(gr)$mod_id,
                       spec_type = mcols(gr)$spec_type,
                       spec_genename = mcols(gr)$spec_genename,
                       spec_ensembl = mcols(gr)$spec_ensembl,
                       spec_ensembltrans = mcols(gr)$spec_ensembltrans,
                       spec_entrezid = mcols(gr)$spec_entrezid)
    specifiers <- specifiers[!vapply(specifiers, is.null, logical(1))]
    if(length(specifiers) > 1L){
        specifiers <- data.frame(specifiers, check.names = FALSE,
                                 stringsAsFactors = FALSE)
    } else {
        specifiers <- NULL
    }
    specifiers
}

# extract references table
.get_gr_references <- function(gr){
    ref_lengths <- lengths(mcols(gr)$ref_type)
    if(length(ref_lengths) == 0L){
        return(NULL)
    }
    references <- list(mod_id = unlist(Map(rep,mcols(gr)$mod_id,ref_lengths)),
                       ref_type = unlist(mcols(gr)$ref_type),
                       ref = unlist(mcols(gr)$ref))
    references <- references[!vapply(references, is.null, logical(1))]
    if(length(references) == 0L){
        return(NULL)
    }
    data.frame(references, check.names = FALSE,
               stringsAsFactors = FALSE)
}

#' @rdname makeEpiTxDbFromGRanges
#' @export
makeEpiTxDbFromGRanges <- function(gr, metadata = NULL, reassign.ids = FALSE){
    modifications <- .get_gr_modifications(gr)
    reactions <- .get_gr_reactions(gr)
    specifiers <- .get_gr_specifiers(gr)
    references <- .get_gr_references(gr)
    makeEpiTxDb(modifications = modifications, reactions = reactions,
                specifiers = specifiers, references = references,
                metadata = metadata, reassign.ids = reassign.ids)
}

library(GenomicRanges)

context("makeEpiTxDbFromGRanges")
test_that("makeEpiTxDbFromGRanges:",{
    gr <- GRanges(seqnames = "test",
                  ranges = IRanges::IRanges(1,1),
                  strand = "+",
                  DataFrame(mod_id = 1L,
                            mod_type = "Am",
                            mod_name = "Am_1"))
    etdb <- makeEpiTxDbFromGRanges(gr)
    expect_s4_class(etdb,"EpiTxDb")
    expect_s3_class(metadata(etdb),"data.frame")
    dbDisconnect(etdb$conn)
})

#' @title \code{EpiTxDb} - Storing and accessing epitranscriptomic information
#' using the AnnotationDbi interface
#'
#' @author Felix G M Ernst [aut]
#'
#' @description
#' title
#'
#' @references
#' Jia-Jia Xuan, Wen-Ju Sun, Ke-Ren Zhou, Shun Liu, Peng-Hui Lin, Ling-Ling
#' Zheng, Liang-Hu Qu, Jian-Hua Yang (2017): "RMBase v2.0: Deciphering the Map of
#' RNA Modifications from Epitranscriptome Sequencing Data." Nucleic Acids
#' Research, Volume 46, Issue D1, 4 January 2018, Pages D327–D334.
#' doi: 10.1093/nar/gkx934
#'
#' Jühling, Frank; Mörl, Mario; Hartmann, Roland K.; Sprinzl, Mathias; Stadler,
#' Peter F.; Pütz, Joern (2009): "TRNAdb 2009: Compilation of tRNA Sequences and
#' tRNA Genes." Nucleic Acids Research 37 (suppl_1): D159–D162.
#' doi: 10.1093/nar/gkn772
#'
#' Sprinzl, Mathias; Vassilenko, Konstantin S. (2005): "Compilation of tRNA
#' Sequences and Sequences of tRNA Genes." Nucleic Acids Research 33 (suppl_1):
#' D139–D140. doi: 10.1093/nar/gki012
#'
#' @name EpiTxDb-package#'
NULL

#' @keywords internal
"_PACKAGE"


#' @import methods
#' @import txdbmaker
#' @import BiocGenerics
#' @import S4Vectors
#' @import IRanges
#' @import AnnotationDbi
#' @import RSQLite
#' @import Seqinfo
#' @import Modstrings
NULL

#' @name EpiTxDb-data
#'
#' @title EpiTxDb internal data
#'
#' @usage data(rmbase_data)
#' @format data.frame
"rmbase_data"

#' @include EpiTxDb-class.R
NULL

#' @name select
#' @aliases select columns keys keytypes
#'
#' @title Using the "select" interface on \code{EpiTxDb} objects
#'
#' @description
#'   As expected for a \code{AnnotationDb} object, the general accessors
#'   \code{select}, \code{keys}, \code{columns} and \code{keytypes} can be used
#'   to get information from a \code{\link[=EpiTxDb-class]{EpiTxDb}} object.
#'
#' @param x a \code{\link[=EpiTxDb-class]{EpiTxDb}} object
#' @param keys,columns,keytype,... See
#'   \code{\link[AnnotationDbi:AnnotationDb-class]{AnnotationDb}} for more
#'   comprehensive description of the methods \code{select}, \code{keys},
#'   \code{columns} and \code{keytypes} and their arguments.
#'
#' @return a \code{data.frame} object for \code{select()} and a \code{character}
#'   vecor for \code{keytypes()}, \code{keys()} and \code{columns()}.
#'
#' @examples
#' etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
#'                          package="EpiTxDb")
#' etdb <- loadDb(etdb_file)
#' etdb
NULL

# helper functions for column name selection/conversion ------------------------

# Helpers to access/process the table names and columns
#' @importFrom DBI dbListTables dbListFields
.getTableColMapping <- function(x){
    conn <- dbconn(x)
    tables <- DBI::dbListTables(conn)
    tCols <- lapply(tables, DBI::dbListFields, conn = conn)
    names(tCols) <- tables
    ## right up front we are getting rid of metadata tables...
    tCols[names(tCols) != "metadata"]
}

# used to match and generate abbreviated column names
.makeColAbbreviations <- function(x){
    tCols <- .getTableColMapping(x)
    longNames <- unique(unlist(tCols, use.names = FALSE))
    abbrev <- unique(toupper((gsub("_","",unlist(tCols, use.names = FALSE)))))
    abbrev <- gsub("^SEQNAMES", "SN", abbrev)
    abbrev <- gsub("^REACTION", "RX", abbrev)
    abbrev <- gsub("^SPECIFIER", "SPEC", abbrev)
    abbrev <- gsub("^REFERENCE", "REF", abbrev)
    names(abbrev) <- longNames
    abbrev
}

# For when you need to get the true table names back from the abbrev's
.reverseColAbbreviations <- function(x, cnames){
    abr <- .makeColAbbreviations(x)
    names(abr)[match(cnames,abr)]
}

# used to retrieve vector of table names that go with vector of col names
.getTableNames <- function(x, cnames){
    realColNames <- .reverseColAbbreviations(x, cnames)
    tCols <- .getTableColMapping(x)
    ## Translate all names to one unique vector.
    getTabNames <- function(name, tCols){names(tCols[grep(name, tCols)])}
    tabNames <- lapply(realColNames, getTabNames, tCols)
    names(tabNames) <- realColNames
    tabNames
}

.getSimpleTableNames <- function(x, cnames){
    unique(unlist(.getTableNames(x, cnames)))
}

# helper functions for generating SQL statements from selected columns ---------

# this just takes the 1 letter abrevs and makes them into a sorted string
# that can be used as a key below
.encodeSortedTableKey <- function(sTNames){
    prefSort <- c("mod","seq","rea","spe","ref")
    res <- sTNames[match(prefSort, sTNames)]
    paste(res[!is.na(res)], collapse="")
}

.makeTableKey <- function(x,cnames){
    sTNames <- substr(.getSimpleTableNames(x, cnames),1,3)
    .encodeSortedTableKey(sTNames)
}

# real joins for likely combos
.tableJoinSelector <- function(tName){
    #
    modseq <- paste("(SELECT * FROM modification ",
                     "JOIN seqnames ON (modification._sn_id = seqnames._sn_id) ) ")

    modrea <- paste("(SELECT * FROM modification ",
                    "LEFT JOIN modification2reaction ON (modification._mod_id = modification2reaction._mod_id) ",
                    "LEFT JOIN reaction ON (modification2reaction._rx_id = reaction._rx_id) )")

    modspe <- paste("(SELECT * FROM modification ",
                    "LEFT JOIN modification2specifier ON (modification._mod_id = modification2specifier._mod_id) ",
                    "LEFT JOIN specifier ON (modification2specifier._spec_id = specifier._spec_id) )")

    modref <- paste("(SELECT * FROM modification ",
                    "LEFT JOIN modification2reference ON (modification._mod_id = modification2reference._mod_id) ",
                    "LEFT JOIN reference ON (modification2reference._ref_id = reference._ref_id) )")
    #
    modseqrea <- paste("(SELECT * FROM modification ",
                    "JOIN seqnames ON (modification._sn_id = seqnames._sn_id) ",
                    "LEFT JOIN modification2reaction ON (modification._mod_id = modification2reaction._mod_id) ",
                    "LEFT JOIN reaction ON (modification2reaction._rx_id = reaction._rx_id) )")

    modseqspe <- paste("(SELECT * FROM modification ",
                    "JOIN seqnames ON (modification._sn_id = seqnames._sn_id) ",
                    "LEFT JOIN modification2specifier ON (modification._mod_id = modification2specifier._mod_id) ",
                    "LEFT JOIN specifier ON (modification2specifier._spec_id = specifier._spec_id) )")

    modseqref <- paste("(SELECT * FROM modification ",
                    "JOIN seqnames ON (modification._sn_id = seqnames._sn_id) ",
                    "LEFT JOIN modification2reference ON (modification._mod_id = modification2reference._mod_id) ",
                    "LEFT JOIN reference ON (modification2reference._ref_id = reference._ref_id) )")
    #
    modreaspe <- paste("(SELECT * FROM modification ",
                       "LEFT JOIN modification2reaction ON (modification._mod_id = modification2reaction._mod_id) ",
                       "LEFT JOIN reaction ON (modification2reaction._rx_id = reaction._rx_id)  ",
                       "LEFT JOIN modification2specifier ON (modification._mod_id = modification2specifier._mod_id) ",
                       "LEFT JOIN specifier ON (modification2specifier._spec_id = specifier._spec_id) ) ")

    modrearef <- paste("(SELECT * FROM modification ",
                       "LEFT JOIN modification2reaction ON (modification._mod_id = modification2reaction._mod_id) ",
                       "LEFT JOIN reaction ON (modification2reaction._rx_id = reaction._rx_id)  ",
                       "LEFT JOIN modification2reference ON (modification._mod_id = modification2reference._mod_id) ",
                       "LEFT JOIN reference ON (modification2reference._ref_id = reference._ref_id) ) ")
    #
    modsperef <- paste("(SELECT * FROM modification ",
                       "LEFT JOIN modification2specifier ON (modification._mod_id = modification2specifier._mod_id) ",
                       "LEFT JOIN specifier ON (modification2specifier._spec_id = specifier._spec_id) ",
                       "LEFT JOIN modification2reference ON (modification._mod_id = modification2reference._mod_id) ",
                       "LEFT JOIN reference ON (modification2reference._ref_id = reference._ref_id) ) ")
    #
    modreasperef <- paste("(SELECT * FROM modification ",
                          "JOIN seqnames ON (modification._sn_id = seqnames._sn_id) ",
                          "LEFT JOIN modification2reaction ON (modification._mod_id = modification2reaction._mod_id) ",
                          "LEFT JOIN reaction ON (modification2reaction._rx_id = reaction._rx_id) ",
                          "LEFT JOIN modification2specifier ON (modification._mod_id = modification2specifier._mod_id) ",
                          "LEFT JOIN specifier ON (modification2specifier._spec_id = specifier._spec_id) ",
                          "LEFT JOIN modification2reference ON (modification._mod_id = modification2reference._mod_id) ",
                          "LEFT JOIN reference ON (modification2reference._ref_id = reference._ref_id) ) ")
    #
    modseqreaspe <- paste("(SELECT * FROM modification ",
                          "JOIN seqnames ON (modification._sn_id = seqnames._sn_id) ",
                          "LEFT JOIN modification2reaction ON (modification._mod_id = modification2reaction._mod_id) ",
                          "LEFT JOIN reaction ON (modification2reaction._rx_id = reaction._rx_id) ",
                          "LEFT JOIN modification2specifier ON (modification._mod_id = modification2specifier._mod_id) ",
                          "LEFT JOIN specifier ON (modification2specifier._spec_id = specifier._spec_id) ) ")

    modseqrearef <- paste("(SELECT * FROM modification ",
                          "JOIN seqnames ON (modification._sn_id = seqnames._sn_id) ",
                          "LEFT JOIN modification2reaction ON (modification._mod_id = modification2reaction._mod_id) ",
                          "LEFT JOIN reaction ON (modification2reaction._rx_id = reaction._rx_id) ",
                          "LEFT JOIN modification2reference ON (modification._mod_id = modification2reference._mod_id) ",
                          "LEFT JOIN reference ON (modification2reference._ref_id = reference._ref_id) ) ")

    modseqsperef <- paste("(SELECT * FROM modification ",
                          "JOIN seqnames ON (modification._sn_id = seqnames._sn_id) ",
                          "LEFT JOIN modification2specifier ON (modification._mod_id = modification2specifier._mod_id) ",
                          "LEFT JOIN specifier ON (modification2specifier._spec_id = specifier._spec_id) ",
                          "LEFT JOIN modification2reference ON (modification._mod_id = modification2reference._mod_id) ",
                          "LEFT JOIN reference ON (modification2reference._ref_id = reference._ref_id) ) ")

    modreasperef <- paste("(SELECT * FROM modification ",
                          "LEFT JOIN modification2reaction ON (modification._mod_id = modification2reaction._mod_id) ",
                          "LEFT JOIN reaction ON (modification2reaction._rx_id = reaction._rx_id) ",
                          "LEFT JOIN modification2specifier ON (modification._mod_id = modification2specifier._mod_id) ",
                          "LEFT JOIN specifier ON (modification2specifier._spec_id = specifier._spec_id) ",
                          "LEFT JOIN modification2reference ON (modification._mod_id = modification2reference._mod_id) ",
                          "LEFT JOIN reference ON (modification2reference._ref_id = reference._ref_id) ) ")
    #
    modseqreasperef <- paste("(SELECT * FROM modification ",
                             "JOIN seqnames ON (modification._sn_id = seqnames._sn_id) ",
                             "LEFT JOIN modification2reaction ON (modification._mod_id = modification2reaction._mod_id) ",
                             "LEFT JOIN reaction ON (modification2reaction._rx_id = reaction._rx_id) ",
                             "LEFT JOIN modification2specifier ON (modification._mod_id = modification2specifier._mod_id) ",
                             "LEFT JOIN specifier ON (modification2specifier._spec_id = specifier._spec_id) ",
                             "LEFT JOIN modification2reference ON (modification._mod_id = modification2reference._mod_id) ",
                             "LEFT JOIN reference ON (modification2reference._ref_id = reference._ref_id) ) ")

    sql <- switch(EXPR = tName,
                  "mod" = "modification",
                  "seq" = "seqnames",
                  "rea" = "reaction",
                  "spe" = "specifier",
                  "ref" = "reference",
                  "modseq" = modseq,
                  "modrea" = modrea,
                  "modspe" = modspe,
                  "modref" = modref,
                  "seqrea" = modseqrea,
                  "seqspe" = modseqspe,
                  "seqref" = modseqref,
                  "reaspe" = modreaspe,
                  "rearef" = modrearef,
                  "speref" = modsperef,
                  "modseqrea" = modseqrea,
                  "modseqspe" = modseqspe,
                  "modseqref" = modseqref,
                  "modreaspe" = modreaspe,
                  "modrearef" = modrearef,
                  "modsperef" = modsperef,
                  "seqreaspe" = modseqreaspe,
                  "seqrearef" = modseqrearef,
                  "seqsperef" = modseqsperef,
                  "reasperef" = modreasperef,
                  "modseqreaspe" = modseqreaspe,
                  "modseqrearef" = modseqrearef,
                  "modreasperef" = modreasperef,
                  "modseqreasperef" = modseqreasperef,
                  stop(paste("No query for this combination of tables.",
                             "Please add ",tName," to the interpolator")))
    sql
}

# helper function to generate SQL statements -----------------------------------

#
.makeJoinSQL <- function(x, cnames){
    tKey <- .makeTableKey(x,cnames)
    .tableJoinSelector(tKey)
}

.makeSelectList <- function(x, cnames, abbrev = TRUE){
    tNames <- .getTableNames(x, cnames)
    ## Here is where we only grab the 1st one...
    tNames <- lapply(tNames,function(x){x[1]})
    ## then continue on...
    tabAbbrevs <- substr(unlist(tNames),1,1)
    names(tabAbbrevs) <- rep(names(tNames),elementNROWS(tNames))
    if(abbrev==TRUE){
        paste(paste0(tabAbbrevs, ".", names(tabAbbrevs)), collapse=", ")
    } else {
        paste(names(tabAbbrevs), collapse=", ")
    }
}

.makeKeyList <- function(x, keys, keytype, abbrev = TRUE){
    colType <- .makeSelectList(x, keytype, abbrev)
    keys <- paste(paste0("'", keys, "'"), collapse=",")
    paste(colType, "IN (", keys,")")
}


# select -----------------------------------------------------------------------

.select <- function(x, keys, columns, keytype, ...){
    extraArgs <- list(...)
    if(missing(keys) || !is.character(keys))
        stop("'keys' must be a character vector")
    if(missing(columns) || !is.character(columns))
        stop("'columns' must be a character vector")
    # Some argument checking
    if('skipValidKeysTest' %in% names(extraArgs)){
        skipValidKeysTest < -extraArgs[["skipValidKeysTest"]]
    } else {
        skipValidKeysTest <- FALSE
    }
    testSelectArgs(x, keys = keys, cols = columns, keytype = keytype,
                   skipValidKeysTest = skipValidKeysTest)

    # 1st we check the keytype to see if it is valid:
    if(is.na(keys(x, keytype)[1]) & length(keys(x, keytype)) == 1){
        stop(paste("There do not appear to be any keys",
                   "for the keytype you have specified."))
    }

    # generate column names
    cnames <- unique(c(keytype, columns))
    # create thejoin SQL
    majorJoin <- .makeJoinSQL(x, cnames)

    if(length(keys) <= 1000){  ##if more than about this, prolly faster to get all
        sql <- paste("SELECT DISTINCT",
                     .makeSelectList(x, cnames, abbrev = FALSE),
                     "FROM",
                     majorJoin,
                     "WHERE",
                     .makeKeyList(x, keys, keytype, abbrev = FALSE))
    } else {
        sql <- paste("SELECT DISTINCT",
                     .makeSelectList(x, cnames, abbrev = FALSE),
                     "FROM",
                     majorJoin)
    }

    res <- AnnotationDbi:::dbQuery(dbconn(x), sql)

    if(length(keys) > 1000){ ##Then drop the extras now(in event there are some)
        ktColId <- .reverseColAbbreviations(x, keytype)
        res <-  res[res[[ktColId]] %in% keys,,drop = FALSE]
    }

    ## Then drop any columns that were not explicitely requested but that may have
    ## been appended to make a joind (like TXID)
    res <- res[,.reverseColAbbreviations(x,cnames),drop = FALSE]

    ## Then sort rows and columns and drop the filtered rows etc. using resort_base
    ## from AnnotationDbi
    joinType <- .reverseColAbbreviations(x, keytype)
    if(dim(res)[1]>0){
        res <- resort_base(res, keys, joinType,
                           .reverseColAbbreviations(x, cnames))
    }

    ## Then put the user preferred headers onto the table
    fcNames <- .makeColAbbreviations(x)
    colnames(res) <- fcNames[match(colnames(res), names(fcNames))]
    res
}

#' @rdname select
#' @export
setMethod("select", "EpiTxDb",
          function(x, keys, columns, keytype, ...) {
              .select(x, keys, columns, keytype, ...)
          }
)

# columns ----------------------------------------------------------------------

.columns <- function(x){
    res <- .makeColAbbreviations(x)
    unname(res)
}

#' @rdname select
#' @export
setMethod("columns", "EpiTxDb",
          function(x) .columns(x)
)

# keys -------------------------------------------------------------------------

.keys <- function(x, keytype){
    testForValidKeytype(x, keytype)
    res <- switch(
        EXPR = keytype,
        "MODID" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT _mod_id FROM modification", 1L),
        "MODTYPE" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT mod_type FROM modification", 1L),
        "MODNAME" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT mod_name FROM modification", 1L),
        "MODSTRAND" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT mod_strand FROM modification", 1L),
        "SNID" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT _sn_id FROM seqnames", 1L),
        "SNNAME" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT sn_name FROM seqnames", 1L),
        "RXGENENAME" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT rx_genename FROM reaction", 1L),
        "RXENSEMBL" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT rx_ensembl FROM reaction", 1L),
        "RXENSEMBLTRANS" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT rx_ensembltrans FROM reaction", 1L),
        "RXENTREZID" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT rx_entrezid FROM reaction", 1L),
        "SPECTYPE" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT spec_type FROM specifier", 1L),
        "SPECGENENAME" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT spec_genename FROM specifier", 1L),
        "SPECENSEMBL" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT spec_entrezid FROM specifier", 1L),
        "SPECENTREZID" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT spec_ensembl FROM specifier", 1L),
        "REFTYPE" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT ref_type FROM reference", 1L),
        "REF" = AnnotationDbi:::dbQuery(
            dbconn(x),
            "SELECT DISTINCT ref FROM reference", 1L),
        stop(paste(keytype, "is not a supported keytype.",
                   " Please use the keytypes",
                   "method to identify viable keytypes")))
    as.character(res)
}

.keysDispatch <- function(x, keytype, ...){
    if (missing(keytype)) keytype <- "MODID"
    AnnotationDbi:::smartKeys(x = x, keytype = keytype, ..., FUN = .keys)
}

#' @rdname select
#' @export
setMethod("keys", "EpiTxDb", .keysDispatch)

# keytypes ---------------------------------------------------------------------

#' @rdname select
#' @export
setMethod("keytypes", "EpiTxDb",
          function(x) return(c("MODID","MODTYPE","MODNAME","MODSTRAND","SNID",
                               "SNNAME","RXGENENAME","RXENSEMBL",
                               "RXENSEMBLTRANS","RXENTREZID", "SPECTYPE",
                               "SPECGENENAME","SPECENSEMBL","SPECENSEMBLTRANS",
                               "SPECENTREZID","REFTYPE","REF"))
)

context("EpiTxDb select")
test_that("EpiTxDb select:",{
    etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
                             package="EpiTxDb")
    etdb <- loadDb(etdb_file)
    #
    actual <- EpiTxDb:::.getTableColMapping(etdb)
    expect_type(actual,"list")
    expect_type(actual[[1L]],"character")
    expect_named(actual, c("modification","modification2reaction",
                           "modification2reference","modification2specifier",
                           "reaction","reference","seqnames","specifier"))
    #
    actual <- EpiTxDb:::.makeColAbbreviations(etdb)
    expect_type(actual,"character")
    expect_named(actual)
    #
    expect_error(EpiTxDb:::.reverseColAbbreviations(etdb),
                 'argument "cnames" is missing')
    expect_equal(EpiTxDb:::.reverseColAbbreviations(etdb, "_mod_id"),
                 NA_character_)
    actual <- EpiTxDb:::.reverseColAbbreviations(etdb, c("MODID","MODNAME"))
    expect_equal(actual,c("_mod_id","mod_name"))
    #
    #.getTableNames
    expect_error(EpiTxDb:::.getTableNames(etdb),
                 'argument "cnames" is missing')
    actual <- EpiTxDb:::.getTableNames(etdb, "_mod_id")
    expect_type(actual,"list")
    expect_true(all(is.na(actual[[1L]])))
    actual <- EpiTxDb:::.getTableNames(etdb, c("MODID","MODNAME"))
    expect_type(actual,"list")
    expect_named(actual,c("_mod_id","mod_name"))
    expect_equal(actual[[2L]],"modification")
    #
    expect_error(EpiTxDb:::.getSimpleTableNames(etdb),
                 'argument "cnames" is missing')
    expect_equal(EpiTxDb:::.getSimpleTableNames(etdb, "_mod_id"),NA_character_)
    actual <- EpiTxDb:::.getSimpleTableNames(etdb,
                                             c("MODID","MODNAME","RXGENENAME"))
    expect_type(actual, "character")
    expect_equal(actual[c(1,5)], c("modification","reaction"))
    #
    expect_error(EpiTxDb:::.encodeSortedTableKey(),
                 'argument "sTNames" is missing')
    expect_equal(EpiTxDb:::.encodeSortedTableKey(c("seq","mod")),"modseq")
    expect_equal(EpiTxDb:::.encodeSortedTableKey(c("spec","seq","mod")),
                 "modseq")
    expect_equal(EpiTxDb:::.encodeSortedTableKey(c("spe","seq","mod")),
                 "modseqspe")
    #
    expect_error(EpiTxDb:::.makeTableKey(etdb),
                 'argument "cnames" is missing')
    actual <- EpiTxDb:::.makeTableKey(etdb,c("MODID","MODNAME"))
    expect_equal(actual,"mod")
    actual <- EpiTxDb:::.makeTableKey(etdb,c("MODID","MODNAME","RXGENENAME"))
    expect_equal(actual,"modrea")
    actual <- EpiTxDb:::.makeTableKey(etdb,c("SPECGENENAME","MODID","MODNAME"))
    expect_equal(actual,"modspe")
    #
    expect_error(EpiTxDb:::.tableJoinSelector(),
                 'argument "tName" is missing')
    expect_error(EpiTxDb:::.tableJoinSelector("test"),
                 'No query for this combination of tables')
    actual <- EpiTxDb:::.tableJoinSelector("modrea")
    expect_type(actual,"character")
    #
    expect_error(EpiTxDb:::.makeJoinSQL(etdb),
                 'argument "cnames" is missing')
    expect_equal(EpiTxDb:::.makeJoinSQL(etdb,c("MODID","MODNAME","RXGENENAME")),
                 actual)
    #
    expect_error(EpiTxDb:::.makeSelectList(etdb),
                 'argument "cnames" is missing')
    actual <- EpiTxDb:::.makeSelectList(etdb, c("MODID","MODNAME","RXGENENAME"))
    expect_type(actual,"character")
    expect_equal(actual,"m._mod_id, m.mod_name, r.rx_genename")
    actual <- EpiTxDb:::.makeSelectList(etdb, c("MODID","MODNAME","RXGENENAME"),
                                        abbrev = FALSE)
    expect_type(actual,"character")
    expect_equal(actual,"_mod_id, mod_name, rx_genename")
    #
    expect_error(EpiTxDb:::.makeKeyList(etdb),
                 'argument "keytype" is missing')
    expect_error(EpiTxDb:::.makeKeyList(etdb, "1"),
                 'argument "keytype" is missing')
    actual <- EpiTxDb:::.makeKeyList(etdb, "1", "MODID")
    expect_type(actual,"character")
    expect_equal(actual, "m._mod_id IN ( '1' )")
    #
    expect_error(EpiTxDb:::.select(etdb, "1", "RX_GENENAME", "MODID"),
                 'Invalid columns: RX_GENENAME')
    actual <- columns(etdb)
    expect_type(actual, "character")
    expect_equal(sort(unname(EpiTxDb:::.makeColAbbreviations(etdb))),actual)
    #
    expect_message(actual <- EpiTxDb:::.select(etdb, "1", "RXGENENAME", "MODID"),
                   "'select\\(\\)' returned 1:1 mapping")
    expect_s3_class(actual, "data.frame")
    expect_equal(colnames(actual),c("MODID","RXGENENAME"))
    expect_equal(expect_message(select(etdb, "1", "RXGENENAME", "MODID"),
                                "'select\\(\\)' returned 1:1 mapping"),actual)
    #
    actual <- keys(etdb, "RXGENENAME")
    expect_type(actual,"character")
    expect_equal(actual,c("dyskerin pseudouridine synthase 1","fibrillarin"))
    actual <- keys(etdb)
    expect_type(actual,"character")
    expect_equal(actual,as.character(seq(1,235)))
    #
    actual <- keytypes(etdb)
    expect_type(actual,"character")
    dbDisconnect(etdb$conn)
})

# =========================================================================
# EpiTxDb schema
# -------------------------------------------------------------------------
#
# Nothing in this file is exported.
#
# 9 tables:
#   - modification
#   - seqnames
#   - reaction
#   - specifier
#   - reference
#   - modification2reaction
#   - modification2specifier
#   - modification2reference
#   - metadata (not described here)

DB_TYPE_NAME <- "Db type"
DB_TYPE_VALUE <- "EpiTxDb"
DB_SCHEMA_VERSION <- "1.0"

# Return the *effective* schema version.
EpiTxDb_schema_version <- function(epitxdb)
{
    conn <- if (is(epitxdb, "EpiTxDb")) dbconn(epitxdb) else epitxdb
    version <- AnnotationDbi:::.getMetaValue(conn, "DBSCHEMAVERSION")
    numeric_version(version)
}


# Table columns ----------------------------------------------------------------

# 'modification' table

EPITXDB_MOD_COLDEFS <- c(
    `_mod_id` = "INTEGER PRIMARY KEY",
    mod_type = "TEXT NOT NULL",
    mod_name = "TEXT NULL",
    mod_start = "INTEGER NOT NULL",
    mod_end = "INTEGER NOT NULL",
    mod_strand = "TEXT NOT NULL",
    `_sn_id` = "INTEGER NOT NULL"
)
EPITXDB_MOD_COLUMNS <- names(EPITXDB_MOD_COLDEFS)

# 'seqnames' table

EPITXDB_SN_COLDEFS <- c(
    `_sn_id` = "INTEGER NOT NULL",
    sn_name = "TEXT NOT NULL"
)
EPITXDB_SN_COLUMNs <- names(EPITXDB_SN_COLDEFS)

# 'reaction' table

EPITXDB_RX_COLDEFS <- c(
    `_rx_id` = "INTEGER PRIMARY KEY",
    rx_genename = "TEXT NULL",
    rx_rank = "INTEGER NULL",
    rx_ensembl = "TEXT NULL",
    rx_ensembltrans = "TEXT NULL",
    rx_entrezid = "TEXT NULL"
)
EPITXDB_RX_COLUMNS <- names(EPITXDB_RX_COLDEFS)

# 'specifier' table

EPITXDB_SPEC_COLDEFS <- c(
    `_spec_id` = "INTEGER NOT NULL",
    spec_type = "TEXT NOT NULL",
    spec_genename = "TEXT NOT NULL",
    spec_ensembl = "TEXT NULL",
    spec_ensembltrans = "TEXT NULL",
    spec_entrezid = "TEXT NULL"
)

EPITXDB_SPEC_COLUMNS <- names(EPITXDB_SPEC_COLDEFS)

# 'reference' table

EPITXDB_REF_COLDEFS <- c(
    `_ref_id` = "INTEGER NOT NULL",
    ref_type = "TEXT NULL",
    ref = "TEXT NULL"
)

EPITXDB_REF_COLUMNS <- names(EPITXDB_REF_COLDEFS)

# modification2* tables

EPITXDB_MOD2RX_COLDEFS <- c(
    `_mod_id` = "INTEGER NOT NULL",
    `_rx_id` = "INTEGER NOT NULL"
)

EPITXDB_MOD2RX_COLUMNS <- names(EPITXDB_MOD2RX_COLDEFS)

EPITXDB_MOD2SPEC_COLDEFS <- c(
    `_mod_id` = "INTEGER NOT NULL",
    `_spec_id` = "INTEGER NOT NULL"
)

EPITXDB_MOD2SPEC_COLUMNS <- names(EPITXDB_MOD2SPEC_COLDEFS)

EPITXDB_MOD2REF_COLDEFS <- c(
    `_mod_id` = "INTEGER NOT NULL",
    `_ref_id` = "INTEGER NOT NULL"
)

EPITXDB_MOD2REF_COLUMNS <- names(EPITXDB_MOD2REF_COLDEFS)

#

EPITXDB_COLUMNS <- list(
    seqnames = EPITXDB_SN_COLUMNs,
    modification = EPITXDB_MOD_COLUMNS,
    reaction = EPITXDB_RX_COLUMNS,
    specifier = EPITXDB_SPEC_COLUMNS,
    reference = EPITXDB_REF_COLUMNS,
    modification2reaction = EPITXDB_MOD2RX_COLUMNS,
    modification2specifier = EPITXDB_MOD2SPEC_COLUMNS,
    modification2reference = EPITXDB_MOD2REF_COLUMNS
)


# Build CREATE TABLE statements ------------------------------------------------

.build_SQL_CREATE_TABLE <- function(table, coldefs, constraints = NULL)
{
    SQL <- "CREATE TABLE %s (%s\n)"
    coldefs <- c(paste(names(coldefs), coldefs), constraints)
    coldefs <- paste("\n  ", coldefs, collapse = ",")
    sprintf(SQL, table, coldefs)
}

build_SQL_CREATE_modification_table <- function()
{
    unique_key <- "UNIQUE (_mod_id, _sn_id)"
    foreign_key <- "FOREIGN KEY (_sn_id) REFERENCES seqnames"
    constraints <- c(unique_key, foreign_key)
    .build_SQL_CREATE_TABLE("modification", EPITXDB_MOD_COLDEFS, constraints)
}

build_SQL_CREATE_seqnames_table <- function()
{
    unique_key <- "UNIQUE (_sn_id)"
    constraints <- c(unique_key)
    .build_SQL_CREATE_TABLE("seqnames", EPITXDB_SN_COLDEFS, constraints)
}

build_SQL_CREATE_reaction_table <- function()
{
    unique_key <- "UNIQUE (_rx_id)"
    constraints <- c(unique_key)
    .build_SQL_CREATE_TABLE("reaction", EPITXDB_RX_COLDEFS, constraints)
}

build_SQL_CREATE_specifier_table <- function()
{
    unique_key <- "UNIQUE (_spec_id)"
    constraints <- c(unique_key)
    .build_SQL_CREATE_TABLE("specifier", EPITXDB_SPEC_COLDEFS, constraints)
}

build_SQL_CREATE_reference_table <- function()
{
    unique_key <- "UNIQUE (_ref_id)"
    constraints <- c(unique_key)
    .build_SQL_CREATE_TABLE("reference", EPITXDB_REF_COLDEFS, constraints)
}

build_SQL_CREATE_modification2reaction_table <- function(){
    unique_key <- "UNIQUE (_mod_id, _rx_id)"
    foreign_key <- paste("FOREIGN KEY (_mod_id) REFERENCES modification",
                         "FOREIGN KEY (_rx_id) REFERENCES reaction")
    constraints <- c(unique_key, foreign_key)
    .build_SQL_CREATE_TABLE("modification2reaction", EPITXDB_MOD2RX_COLDEFS,
                            constraints)
}

build_SQL_CREATE_modification2specifier_table <- function(){
    unique_key <- "UNIQUE (_mod_id, _spec_id)"
    foreign_key <- paste("FOREIGN KEY (_mod_id) REFERENCES modification",
                         "FOREIGN KEY (_spec_id) REFERENCES specifier")
    constraints <- c(unique_key, foreign_key)
    .build_SQL_CREATE_TABLE("modification2specifier", EPITXDB_MOD2SPEC_COLDEFS,
                            constraints)
}

build_SQL_CREATE_modification2reference_table <- function(){
    unique_key <- "UNIQUE (_mod_id, _ref_id)"
    foreign_key <- paste("FOREIGN KEY (_mod_id) REFERENCES modification",
                         "FOREIGN KEY (_ref_id) REFERENCES reference")
    constraints <- c(unique_key, foreign_key)
    .build_SQL_CREATE_TABLE("modification2reference", EPITXDB_MOD2REF_COLDEFS,
                            constraints)
}

# helper functions -------------------------------------------------------------

EPITXDB_tables <- function() names(EPITXDB_COLUMNS)

EPITXDB_table_columns <- function(table, schema_version = NA){
    columns <- EPITXDB_COLUMNS[[table]]
    columns
}

EPITXDB_column2table <- function(columns, from_table = NA, schema_version = NA){
    if (length(columns) == 0L)
        return(character(0))
    table_columns <- lapply(EPITXDB_tables(), EPITXDB_table_columns,
                            schema_version = schema_version)
    names(table_columns) <- EPITXDB_tables()
    table_columns <- unlist(IRanges::CharacterList(table_columns))
    if(any(!(columns %in% table_columns))){
        if (is.na(schema_version)) {
            in_schema <- ""
        } else {
            in_schema <- c(" in db schema ", as.character(schema_version))
        }
        stop(columns[!(columns %in% table_columns)], ": no such column",
             in_schema)
    }
    tables <- names(table_columns[match(columns,table_columns)])
    names(tables) <- columns
    if (!is.na(from_table)) {
        table_columns <- EPITXDB_table_columns(from_table,
                                               schema_version = schema_version)
        tables[columns %in% table_columns] <- from_table
    }
    tables
}

EPITXDB_table2joinColumns <- function(tables, schema_version = NA){
    if (length(tables) == 0L)
        return(character(0))
    f <- tables %in% EPITXDB_tables()
    if(any(!f)){
        if (is.na(schema_version)) {
            in_schema <- ""
        } else {
            in_schema <- c(" in db schema ", as.character(schema_version))
        }
        stop(tables[!f], ": no such table", in_schema)
    }
    columns <- vapply(lapply(tables, EPITXDB_table_columns,
                             schema_version = schema_version),
                      "[[",character(1),1L)
    names(columns) <- tables
    columns
}

context("EpiTxDb class")
test_that("EpiTxDb:",{
    expect_error(EpiTxDb:::.format_modifications(),
                 'argument "modifications" is missing')
    expect_error(EpiTxDb:::.format_seqnames(),
                 'argument "seqnames" is missing')
    expect_error(EpiTxDb:::.format_reactions(),
                 'argument "reactions" is missing')
    expect_error(EpiTxDb:::.format_specifiers(),
                 'argument "specifiers" is missing')
    expect_error(EpiTxDb:::.format_references(),
                 'argument "references" is missing')
    #
    actual <- EpiTxDb:::.format_modifications(NULL)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("mod_id","mod_type","mod_name","mod_start",
                                    "mod_end","sn_id"))
    df <- data.frame(mod_id = 1, mod_type = 1, mod_name = 1, mod_start = 1,
                     mod_end = 1, sn_id = 1)
    actual <- EpiTxDb:::.format_modifications(df)
    expect_type(actual[,1],"double")
    expect_type(actual[,3],"double")
    actual <- EpiTxDb:::.format_modifications(df, set.col.class = TRUE)
    expect_type(actual[,1],"integer")
    expect_type(actual[,3],"character")
    #
    actual <- EpiTxDb:::.format_seqnames(NULL)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("sn_id","sn_name"))
    df <- data.frame(sn_id = 1, sn_name = 1)
    actual <- EpiTxDb:::.format_seqnames(df)
    expect_type(actual[,1],"double")
    expect_type(actual[,2],"double")
    actual <- EpiTxDb:::.format_seqnames(df, set.col.class = TRUE)
    expect_type(actual[,1],"integer")
    expect_type(actual[,2],"character")
    #
    actual <- EpiTxDb:::.format_reactions(NULL)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("rx_id","rx_genename","rx_rank",
                                    "rx_ensembl","rx_ensembltrans",
                                    "rx_entrezid"))
    df <- data.frame(rx_id = 1, rx_genename = 1, rx_rank = 1, rx_ensembl = 1,
                     rx_ensembltrans = 1, rx_entrezid = 1)
    actual <- EpiTxDb:::.format_reactions(df)
    expect_type(actual[,1],"double")
    expect_type(actual[,3],"double")
    actual <- EpiTxDb:::.format_reactions(df, set.col.class = TRUE)
    expect_type(actual[,1],"integer")
    expect_type(actual[,2],"character")
    #
    actual <- EpiTxDb:::.format_specifiers(NULL)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("spec_id","spec_type","spec_genename",
                                    "spec_ensembl","spec_ensembltrans",
                                    "spec_entrezid"))
    df <- data.frame(spec_id = 1, spec_type = 1, spec_genename = 1,
                     spec_ensembl = 1, spec_ensembltrans = 1, spec_entrezid = 1)
    actual <- EpiTxDb:::.format_specifiers(df)
    expect_type(actual[,1],"double")
    expect_type(actual[,3],"double")
    actual <- EpiTxDb:::.format_specifiers(df, set.col.class = TRUE)
    expect_type(actual[,1],"integer")
    expect_true(is.factor(actual[,2]))
    #
    actual <- EpiTxDb:::.format_references(NULL)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("ref_id","ref_type","ref"))
    df <- data.frame(ref_id = 1, ref_type = 1, ref = 1)
    actual <- EpiTxDb:::.format_references(df)
    expect_type(actual[,1],"double")
    expect_type(actual[,3],"double")
    actual <- EpiTxDb:::.format_references(df, set.col.class = TRUE)
    expect_type(actual[,1],"integer")
    expect_true(is.factor(actual[,2]))
    expect_type(actual[,3],"character")
    ############################################################################
    etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
                             package="EpiTxDb")
    etdb <- loadDb(etdb_file)
    expect_s4_class(etdb,"EpiTxDb")
    # .get_EpiTxDb_seqinfo
    actual <- EpiTxDb:::.get_EpiTxDb_seqinfo(etdb)
    expect_s4_class(actual,"Seqinfo")
    expect_equal(seqinfo(etdb),actual)
    expect_type(seqlevels(etdb),"character")
    # as.list
    actual <- as.list(etdb)
    expect_type(actual,"list")
    expect_named(actual,c("modifications","seqnames","reactions",
                          "specifiers","references"))
    # compareEpiTxDbs
    expect_true(EpiTxDb:::compareEpiTxDbs(etdb,etdb))
    dbDisconnect(etdb$conn)
})

context("EpiTxDb SQL")
test_that("EpiTxDb SQL:",{
    expect_type(EpiTxDb:::build_SQL_CREATE_modification_table(),"character")
    expect_type(EpiTxDb:::build_SQL_CREATE_seqnames_table(),"character")
    expect_type(EpiTxDb:::build_SQL_CREATE_reaction_table(),"character")
    expect_type(EpiTxDb:::build_SQL_CREATE_specifier_table(),"character")
    expect_type(EpiTxDb:::build_SQL_CREATE_reference_table(),"character")
    expect_type(EpiTxDb:::build_SQL_CREATE_modification2reaction_table(),
                "character")
    expect_type(EpiTxDb:::build_SQL_CREATE_modification2specifier_table(),
                "character")
    expect_type(EpiTxDb:::build_SQL_CREATE_modification2reference_table(),
                "character")
    #
    expect_error(EpiTxDb:::EPITXDB_column2table(),
                 'argument "columns" is missing')
    expect_null(EpiTxDb:::EPITXDB_table_columns("test"))
    expect_error(EpiTxDb:::EPITXDB_column2table("mod_id"),
                 "mod_id: no such column")
    actual <- EpiTxDb:::EPITXDB_column2table("_mod_id")
    expect_equal(actual,c(`_mod_id` = "modification"))
    actual <- EpiTxDb:::EPITXDB_column2table("_sn_id")
    expect_equal(actual,c(`_sn_id` = "seqnames"))
    actual <- EpiTxDb:::EPITXDB_column2table("_rx_id")
    expect_equal(actual,c(`_rx_id` = "reaction"))
    actual <- EpiTxDb:::EPITXDB_column2table("_spec_id")
    expect_equal(actual,c(`_spec_id` = "specifier"))
    actual <- EpiTxDb:::EPITXDB_column2table("_ref_id")
    expect_equal(actual,c(`_ref_id` = "reference"))
    #
    expect_error(EpiTxDb:::EPITXDB_table2joinColumns(),
                 'argument "tables" is missing')
    expect_error(EpiTxDb:::EPITXDB_table2joinColumns("test"),
                 'test: no such table')
    actual <- EpiTxDb:::EPITXDB_table2joinColumns("modification")
    expect_equal(actual,c(modification = "_mod_id"))
    actual <- EpiTxDb:::EPITXDB_table2joinColumns("seqnames")
    expect_equal(actual,c(seqnames = "_sn_id"))
    actual <- EpiTxDb:::EPITXDB_table2joinColumns("reaction")
    expect_equal(actual,c(reaction = "_rx_id"))
    actual <- EpiTxDb:::EPITXDB_table2joinColumns("specifier")
    expect_equal(actual,c(specifier = "_spec_id"))
    actual <- EpiTxDb:::EPITXDB_table2joinColumns("reference")
    expect_equal(actual,c(reference = "_ref_id"))
    #
    expect_error(EpiTxDb:::.EPITXDB_add_table_bundle(),
                 'argument "tables" is missing')
    expect_equal(EpiTxDb:::.EPITXDB_add_table_bundle("modification"),
                 "modification")
    expect_equal(EpiTxDb:::.EPITXDB_add_table_bundle(c("modification",
                                                       "seqnames")),
                 c("modification", "seqnames"))
    expect_equal(EpiTxDb:::.EPITXDB_add_table_bundle(c("modification",
                                                       "reaction")),
                 c("modification", "modification2reaction", "reaction"))
    expect_equal(EpiTxDb:::.EPITXDB_add_table_bundle(c("modification",
                                                       "specifier")),
                 c("modification", "modification2specifier", "specifier"))
    expect_equal(EpiTxDb:::.EPITXDB_add_table_bundle(c("modification",
                                                       "reference")),
                 c("modification", "modification2reference", "reference"))
    #
    expect_error(EpiTxDb:::.EPITXDB_join_tables(),
                 'argument "tables" is missing')
    expect_error(EpiTxDb:::.EPITXDB_join_tables(c("specifier", "reference")),
                 'length\\(tables\\) == length\\(joinColumn\\)')
    expect_equal(EpiTxDb:::.EPITXDB_join_tables(c("modification", "reference"),
                                                c("_mod_id","_ref_id")),
                 c("modification",
                   "modification._mod_id=modification2reference._mod_id",
                   "modification2reference",
                   "modification2reference._ref_id=reference._ref_id",
                   "reference"))
    #
    etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
                             package="EpiTxDb")
    etdb <- loadDb(etdb_file)
    expect_output(show(etdb))
    expect_true(validObject(etdb))
    expect_type(organism(etdb),"character")
    expect_length(organism(etdb),1L)
    actual <- EpiTxDb:::EpiTxDb_SELECT_from_modification(etdb)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("_mod_id","mod_type","mod_name","mod_start",
                                    "mod_end","mod_strand","_sn_id"))
    actual <- EpiTxDb:::EpiTxDb_SELECT_from_seqnames(etdb)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("_sn_id","sn_name"))
    actual <- EpiTxDb:::EpiTxDb_SELECT_from_reaction(etdb)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("_rx_id","rx_genename","rx_rank",
                                    "rx_ensembl","rx_ensembltrans",
                                    "rx_entrezid"))
    actual <- EpiTxDb:::EpiTxDb_SELECT_from_specifier(etdb)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("_spec_id","spec_type","spec_genename",
                                    "spec_ensembl","spec_ensembltrans",
                                    "spec_entrezid"))
    actual <- EpiTxDb:::EpiTxDb_SELECT_from_reference(etdb)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("_ref_id","ref_type","ref"))
    #
    expect_error(EpiTxDb:::EpiTxDb_SELECT_from_LEFT_JOIN(),
                 'argument "epitxdb" is missing')
    expect_error(EpiTxDb:::EpiTxDb_SELECT_from_LEFT_JOIN(etdb),
                 'argument "columns" is missing')
    table <- "modification"
    cols <- c("_mod_id","mod_name","spec_genename")
    actual <- EpiTxDb:::EpiTxDb_SELECT_from_LEFT_JOIN(etdb, table, cols)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),cols)
    table <- "specifier"
    cols <- c("rx_genename","spec_genename")
    actual <- EpiTxDb:::EpiTxDb_SELECT_from_LEFT_JOIN(etdb, table, cols)
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),cols)
    dbDisconnect(etdb$conn)
})

## ----style, echo = FALSE, results = 'asis'------------------------------------
BiocStyle::markdown(css.files = c('custom.css'))

## ----eval=FALSE---------------------------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install(c("EpiTxDb","EpiTxDb.Hs.hg38"))

## ----results="hide", include=TRUE, message=FALSE, warning=FALSE---------------
library(EpiTxDb)
library(EpiTxDb.Hs.hg38)

## -----------------------------------------------------------------------------
etdb <- EpiTxDb.Hs.hg38.snoRNAdb()
etdb

## -----------------------------------------------------------------------------
keytypes(etdb)
columns(etdb)
head(keys(etdb, "MODID"))
select(etdb, keys = "1",
       columns = c("MODNAME","MODTYPE","MODSTART","MODSTRAND","SNNAME",
                   "RXGENENAME","SPECTYPE","SPECGENENAME"),
       keytype = "MODID")

## -----------------------------------------------------------------------------
species(etdb)
organism(etdb)
seqlevels(etdb)

## -----------------------------------------------------------------------------
modifications(etdb, columns = c("mod_id","mod_type","mod_name",
                                "rx_genename","spec_genename",
                                "ref_type","ref"),
              filter = list(mod_id = 1:3))

## -----------------------------------------------------------------------------
# split by sequence name, usually a transcipt identifier
modificationsBy(etdb, by = "seqnames")
# split modification type
modificationsBy(etdb, by = "modtype")

## ----echo = FALSE-------------------------------------------------------------
suppressPackageStartupMessages({
  library(TxDb.Hsapiens.UCSC.hg38.knownGene)
  library(BSgenome.Hsapiens.UCSC.hg38)
})

## ----eval = FALSE-------------------------------------------------------------
# library(TxDb.Hsapiens.UCSC.hg38.knownGene)
# library(BSgenome.Hsapiens.UCSC.hg38)

## -----------------------------------------------------------------------------
txdb <- TxDb.Hsapiens.UCSC.hg38.knownGene
seqlevels(txdb) <- "chr1"
bs <- BSgenome.Hsapiens.UCSC.hg38

etdb <- EpiTxDb.Hs.hg38.RMBase()

tx <- exonsBy(txdb)
mod <- modifications(etdb, filter = list(sn_name = "chr1"))
length(mod)

## -----------------------------------------------------------------------------
mod_tx <- shiftGenomicToTranscript(mod, tx)
length(mod_tx)

## -----------------------------------------------------------------------------
mod_tx <- split(mod_tx,seqnames(mod_tx))
names <- Reduce(intersect,list(names(mod_tx),names(tx)))

# Getting the corresponding 5'-UTR and 3'-UTR annotations
fp <- fiveUTRsByTranscript(txdb)
tp <- threeUTRsByTranscript(txdb)
tx <- tx[names]
mod_tx <- mod_tx[names]
fp_m <- match(names,names(fp))
fp_m <- fp_m[!is.na(fp_m)]
tp_m <- match(names,names(tp))
tp_m <- tp_m[!is.na(tp_m)]
fp <- fp[fp_m]
tp <- tp[tp_m]

# Getting lengths of transcripts, 5'-UTR and 3'-UTR
tx_lengths <- sum(width(tx))
fp_lengths <- rep(0L,length(tx))
names(fp_lengths) <- names
fp_lengths[names(fp)] <- sum(width(fp))
tp_lengths <- rep(0L,length(tx))
names(tp_lengths) <- names
tp_lengths[names(tp)] <- sum(width(tp))

# Rescale modifications
# CDS start is at position 1L and cds end at position 1000L
from <- IRanges(fp_lengths+1L, tx_lengths - tp_lengths)
to <- IRanges(1L,1000L)
mod_rescale <- rescale(mod_tx, to, from)

# Construct result data.frame
rel_pos <- data.frame(mod_type = unlist(mcols(mod_rescale,level="within")[,"mod_type"]),
                      rel_pos = unlist(start(mod_rescale)))
rel_pos <- rel_pos[rel_pos$rel_pos < 1500 & rel_pos$rel_pos > -500,]

## -----------------------------------------------------------------------------
library(ggplot2)
ggplot(rel_pos[rel_pos$mod_type %in% c("m6A","m1A","Y"),],
       aes(x = rel_pos, colour = mod_type)) + 
  geom_density()

## -----------------------------------------------------------------------------
sessionInfo()

#' @include EpiTxDb.R
#' @include AllGenerics.R
#' @include EpiTxDb-schema.R
#' @include EpiTxDb-SELECT-helpers.R
NULL

#' @name EpiTxDb-class
#' @aliases EpiTxDb
#'
#' @title EpiTxDb objects
#'
#' @description
#' The \code{EpiTxDb} class is a
#' \code{\link[AnnotationDbi:AnnotationDb-class]{AnnotationDb}} type container
#' for storing Epitranscriptomic information.
#'
#' The information are typically stored on a per transcript and not as genomic
#' coordinates, but the \code{EpiTxDb} class is agnostic to this. In case of
#' genomic coordinates \code{transcriptsBy} will return modifications per
#' chromosome.
#'
#' @param x,object a \code{EpiTxDb} object
#'
#' @seealso
#'   \itemize{
#'     \item \code{\link[=makeEpiTxDbFromGRanges]{makeEpiTxDbFromGRanges}} for
#'       creating a \code{EpiTxDb} object from a
#'       \code{\link[GenomicRanges:GRanges-class]{GRanges}} object and it's
#'        metadata columns
#'     \item \code{\link[=makeEpiTxDbFromRMBase]{makeEpiTxDbFromRMBase}} for
#'       creating a \code{EpiTxDb} object from RMBase online resources
#'     \item \code{\link[=makeEpiTxDbFromtRNAdb]{makeEpiTxDbFromtRNAdb}} for
#'       creating a \code{EpiTxDb} object from tRNAdb online resources
#'     \item \code{\link[=makeEpiTxDb]{makeEpiTxDb}} for creating a
#'       \code{EpiTxDb} object from \code{data.frames}
#'     \item \code{\link[=modifications]{modifications}},
#'       \code{\link[=modifications]{modificationsBy}} for getting
#'       epitranscriptomic modification locations
#'     \item \code{\link[=select]{select}} for using the default interface of
#'       \code{\link[AnnotationDbi:AnnotationDb-class]{AnnotationDb}} objects.
#'     \item \code{\link[=shiftGenomicToTranscript]{shiftGenomicToTranscript}}
#'       and \code{\link[=shiftGenomicToTranscript]{shiftTranscriptToGenomic}}
#'       for transfering genomic to transcript coordinates and back again.
#' }
#'
#' @return For
#'   \itemize{
#'     \item \code{organism()} and \code{seqlevels()}: a \code{character}
#'       vector
#'     \item \code{seqinfo()}: a
#'       \code{\link[Seqinfo:Seqinfo-class]{Seqinfo}} object
#'     \item \code{as.list()} a \code{list}
#'   }
#'
#' @export
#'
#' @examples
#' etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
#'                         package="EpiTxDb")
#' etdb <- loadDb(etdb_file)
#' etdb
#'
#' # general methods
#' seqinfo(etdb) #
#' seqlevels(etdb) # easy access to all transcript names
.EpiTxDb <- setRefClass("EpiTxDb", contains = "AnnotationDb",
                        fields = list( ),
                        methods = list(
                            initialize = function(...) {
                                callSuper(...)
                                .self
                            }))


# Low-level data loaders -------------------------------------------------------

makeZeroRowDataFrame <- txdbmaker:::makeZeroRowDataFrame
setDataFrameColClass <- txdbmaker:::setDataFrameColClass

.format_modifications <- function(modifications, set.col.class = FALSE){
    COL2CLASS <- c(
        mod_id = "integer",
        mod_type = "character",
        mod_name = "character",
        mod_start = "integer",
        mod_end = "integer",
        sn_id = "integer"
    )
    if (is.null(modifications)) {
        modifications <- makeZeroRowDataFrame(COL2CLASS)
    } else {
        if (!is.data.frame(modifications))
            stop("'modifications' must be a data frame")
        if (!identical(names(modifications), names(COL2CLASS)))
            modifications <- modifications[names(COL2CLASS)]
        if (set.col.class)
            modifications <- setDataFrameColClass(modifications, COL2CLASS)
    }
    modifications
}

load_modifications <- function(epitxdb, set.col.class = FALSE){
    modifications <- EpiTxDb_SELECT_from_modification(epitxdb)
    colnames(modifications) <- sub("^_", "", colnames(modifications))
    .format_modifications(modifications, set.col.class = set.col.class)
}

.format_seqnames <- function(seqnames, set.col.class = FALSE){
    COL2CLASS <- c(
        sn_id = "integer",
        sn_name = "character"
    )
    if (is.null(seqnames)) {
        seqnames <- makeZeroRowDataFrame(COL2CLASS)
    } else {
        if (!is.data.frame(seqnames))
            stop("'seqnames' must be a data frame")
        if (!identical(names(seqnames), names(COL2CLASS)))
            seqnames <- seqnames[names(COL2CLASS)]
        if (set.col.class)
            seqnames <- setDataFrameColClass(seqnames, COL2CLASS)
    }
    seqnames
}

load_seqnames <- function(epitxdb, set.col.class = FALSE){
    seqnames <- EpiTxDb_SELECT_from_seqnames(epitxdb)
    colnames(seqnames) <- sub("^_", "", colnames(seqnames))
    .format_seqnames(seqnames, set.col.class = set.col.class)
}

.format_reactions <- function(reactions, set.col.class = FALSE){
    COL2CLASS <- c(
        rx_id = "integer",
        rx_genename = "character",
        rx_rank = "integer",
        rx_ensembl = "character",
        rx_ensembltrans = "character",
        rx_entrezid = "character"
    )
    if (is.null(reactions)) {
        reactions <- makeZeroRowDataFrame(COL2CLASS)
    } else {
        if (!is.data.frame(reactions))
            stop("'reactions' must be a data frame")
        if (!identical(names(reactions), names(COL2CLASS)))
            reactions <- reactions[names(COL2CLASS)]
        if (set.col.class)
            reactions <- setDataFrameColClass(reactions, COL2CLASS)
    }
    reactions
}

load_reactions <- function(epitxdb, set.col.class = FALSE){
    reactions <- EpiTxDb_SELECT_from_reaction(epitxdb)
    colnames(reactions) <- sub("^_", "", colnames(reactions))
    .format_reactions(reactions,  set.col.class = set.col.class)
}

.format_specifiers <- function(specifiers, set.col.class = FALSE){
    COL2CLASS <- c(
        spec_id="integer",
        spec_type="factor",
        spec_genename="character",
        spec_ensembl="character",
        spec_ensembltrans="character",
        spec_entrezid="character"
    )
    if (is.null(specifiers)) {
        specifiers <- makeZeroRowDataFrame(COL2CLASS)
    } else {
        if (!is.data.frame(specifiers))
            stop("'specifiers' must be a data frame")
        if (!identical(names(specifiers), names(COL2CLASS)))
            specifiers <- specifiers[names(COL2CLASS)]
        if (set.col.class)
            specifiers <- setDataFrameColClass(specifiers, COL2CLASS)
    }
    specifiers
}

load_specifiers <- function(epitxdb, set.col.class = FALSE) {
    specifiers <- EpiTxDb_SELECT_from_specifier(epitxdb)
    colnames(specifiers) <- sub("^_", "", colnames(specifiers))
    .format_specifiers(specifiers, set.col.class = set.col.class)
}

.format_references <- function(references, set.col.class = FALSE){
    COL2CLASS <- c(
        ref_id="integer",
        ref_type="factor",
        ref="character"
    )
    if (is.null(references)) {
       references <- makeZeroRowDataFrame(COL2CLASS)
    } else {
        if (!is.data.frame(references))
            stop("'references' must be a data frame")
        if (!identical(names(references), names(COL2CLASS)))
            references <- references[names(COL2CLASS)]
        if (set.col.class)
            references <- setDataFrameColClass(references, COL2CLASS)
    }
    references
}

load_references <- function(epitxdb, set.col.class = FALSE) {
    references <- EpiTxDb_SELECT_from_reference(epitxdb)
    colnames(references) <- sub("^_", "", colnames(references))
    .format_references(references, set.col.class = set.col.class)
}


# validity EpiTxDb -------------------------------------------------------------

.valid.modification.table <- function(conn){
    schema_version <- EpiTxDb_schema_version(conn)
    colnames <- EPITXDB_table_columns("modification",
                                      schema_version=schema_version)
    msg <- AnnotationDbi:::.valid.table.colnames(conn, "modification", colnames)
    if (!is.null(msg))
      return(msg)
    NULL
}

.valid.seqnames.table <- function(conn){
    schema_version <- EpiTxDb_schema_version(conn)
    colnames <- EPITXDB_table_columns("seqnames",
                                      schema_version=schema_version)
    msg <- AnnotationDbi:::.valid.table.colnames(conn, "seqnames", colnames)
    if (!is.null(msg))
      return(msg)
    NULL
}

.valid.reaction.table <- function(conn){
    schema_version <- EpiTxDb_schema_version(conn)
    colnames <- EPITXDB_table_columns("reaction",
                                      schema_version=schema_version)
    msg <- AnnotationDbi:::.valid.table.colnames(conn, "reaction", colnames)
    if (!is.null(msg))
      return(msg)
    NULL
}

.valid.specifier.table <- function(conn){
    schema_version <- EpiTxDb_schema_version(conn)
    colnames <- EPITXDB_table_columns("specifier",
                                      schema_version=schema_version)
    msg <- AnnotationDbi:::.valid.table.colnames(conn, "specifier", colnames)
    if (!is.null(msg))
      return(msg)
    NULL
}

.valid.reference.table <- function(conn){
    schema_version <- EpiTxDb_schema_version(conn)
    colnames <- EPITXDB_table_columns("reference",
                                      schema_version=schema_version)
    msg <- AnnotationDbi:::.valid.table.colnames(conn, "reference", colnames)
    if (!is.null(msg))
      return(msg)
    NULL
}

.valid.EpiTxDb <- function(object){
    conn <- dbconn(object)
    c(AnnotationDbi:::.valid.metadata.table(conn, DB_TYPE_NAME, DB_TYPE_VALUE),
        .valid.modification.table(conn),
        .valid.seqnames.table(conn),
        .valid.reaction.table(conn),
        .valid.specifier.table(conn),
        .valid.reference.table(conn))
}

setValidity("EpiTxDb", .valid.EpiTxDb)

# constructor ------------------------------------------------------------------

EpiTxDb <- function(conn) .EpiTxDb$new(conn = conn)

# methods ----------------------------------------------------------------------

#' @rdname EpiTxDb-class
#' @importFrom BiocGenerics organism
#' @export
setMethod("organism", "EpiTxDb",
    function(object)
    {
        metadata <- metadata(object)
        names <- tolower(metadata[ , "name"])
        metadata <- metadata[ , "value"]
        names(metadata) <- names
        unname(metadata["organism"])
    }
)

# seqinfo ----------------------------------------------------------------------

.get_seqnames <- function(seqnames_db){
    if(!all(lengths(seqnames_db$`_sn_id`) == 1L) ||
       !all(lengths(seqnames_db$sn_name) == 1L)){
        stop(".")
    }
    sn_name <- seqnames_db$sn_name
    sn_id <- seqnames_db$`_sn_id`
    if(is(sn_name,"List")){
        sn_name <- unlist(sn_name)
    }
    if(is(sn_id,"List")){
        sn_id <- unlist(sn_id)
    }
    name_not_empty <- sn_name != ""
    if(all(name_not_empty)){
        return(sn_name)
    }
    seqnames <- sn_id
    if(length(intersect(seqnames, sn_name)) != 0L){
        return(seqnames)
    }
    seqnames[name_not_empty] <- sn_name[name_not_empty]
    seqnames
}

.get_EpiTxDb_seqinfo <- function(x){
    sql <- paste0("SELECT DISTINCT _sn_id, sn_name FROM ",
                  "seqnames")
    seqnames_db <- queryAnnotationDb(x, sql)
    ans <- Seqinfo(seqnames = .get_seqnames(seqnames_db))
    sql <- "SELECT value FROM metadata WHERE name='Genome'"
    genome <- unlist(queryAnnotationDb(x, sql))
    names(genome) <- NULL
    if (length(genome) != 0L)
      genome(ans) <- genome
    ans
}

#' @rdname EpiTxDb-class
#' @importFrom Seqinfo seqinfo
#' @export
setMethod("seqinfo", "EpiTxDb", .get_EpiTxDb_seqinfo)

.get_EpiTxDb_seqlevels <- function(x){
    seqnames(.get_EpiTxDb_seqinfo(x))
}

#' @rdname EpiTxDb-class
#' @importFrom Seqinfo seqlevels
#' @export
setMethod("seqlevels", "EpiTxDb", .get_EpiTxDb_seqlevels)

# as.list and comparison -------------------------------------------------------

.format_txdb_dump <- function(modifications = NULL, seqnames = NULL,
                              reactions = NULL, specifiers = NULL,
                              references = NULL){
    modifications <- .format_modifications(modifications, set.col.class = TRUE)
    seqnames <- .format_seqnames(seqnames, set.col.class = TRUE)
    reactions <- .format_reactions(reactions, set.col.class = TRUE)
    specifiers <- .format_specifiers(specifiers, set.col.class = TRUE)
    references <- .format_references(references, set.col.class = TRUE)
    list(modifications = modifications, seqnames = seqnames,
         reactions = reactions, specifiers = specifiers,
         references = references)
}

#' @rdname EpiTxDb-class
#' @export
setMethod("as.list", "EpiTxDb",
    function(x)
    {
        modifications <- load_modifications(x)
        seqnames <- load_seqnames(x)
        reactions <- load_reactions(x)
        specifiers <- load_specifiers(x)
        references <- load_references(x)
        .format_txdb_dump(modifications, seqnames, reactions, specifiers,
                          references)
    }
)

compareEpiTxDbs <- function(epitxdb1, epitxdb2)
{
    if (!is(epitxdb1, "EpiTxDb") || !is(epitxdb2, "EpiTxDb"))
        stop("'epitxdb1' and 'epitxdb2' must be EpiTxDb objects")
    epitxdb1_dump <- as.list(epitxdb1)
    epitxdb2_dump <- as.list(epitxdb2)
    identical(epitxdb1_dump, epitxdb2_dump)
}

# coercion ---------------------------------------------------------------------


# show -------------------------------------------------------------------------

setMethod("show", "EpiTxDb",
          function(object)
          {
              cat(class(object), "object:\n")
              metadata <- metadata(object)
              cat(apply(metadata,1L,
                        function(row){
                          paste0("# ", row["name"], ": ", row["value"],
                                 "\n", sep ="")
                        }),
                  sep ="")
          }
)

#' @rdname positionSequence
setGeneric("positionSequence",
           function(x, order = FALSE, decreasing = FALSE)
               standardGeneric("positionSequence"))

#' @rdname rescale
setGeneric("rescale",
           function(x, to = 1L, from = 1L)
               standardGeneric("rescale"))

#' @rdname shiftGenomicToTranscript
setGeneric("shiftTranscriptToGenomic",
           function(subject, tx)
               standardGeneric("shiftTranscriptToGenomic"))

#' @rdname shiftGenomicToTranscript
setGeneric("shiftGenomicToTranscript",
           function(subject, tx)
               standardGeneric("shiftGenomicToTranscript"))


#' @rdname modifications
setGeneric("modifications",
           function(x, columns  = c("mod_id","mod_type","mod_name"),
                    filter = NULL, use.names = FALSE, ...)
               standardGeneric("modifications"))

#' @rdname modifications
setGeneric("modificationsBy",
           function(x, by = c("seqnames","mod_type","reaction","specifier",
                              "specifier_type"),
                    ...)
               standardGeneric("modificationsBy"))

#' @rdname modifications
setGeneric("modifiedSeqsByTranscript",
           function(x, sequences, ...)
               standardGeneric("modifiedSeqsByTranscript"))

#' @include EpiTxDb-class.R
NULL

#' @name makeEpiTxDb
#'
#' @title Creating a \code{EpiTxDb} from user supplied annotations as
#' \code{data.frame}s
#'
#' @description
#' \code{makeEpiTxDb} is a low-level constructor for creating a
#' \code{\link[=EpiTxDb-class]{EpiTxDb}} object from user supplied annotations.
#'
#' This functions typically will not be used by regular users.
#'
#' @param modifications A \code{data.frame} containg the following columns:
#'   \itemize{
#'     \item \code{mod_id}: a unique \code{integer} value per modification.
#'     \item \code{mod_type}: the modification type as a \code{character} or
#'     \code{factor} value. Must be a value from
#'     \code{shortName(ModRNAString())}.
#'     \item \code{mod_name}: a \code{character} or \code{factor} name for the
#'     specific modification
#'     \item \code{mod_start}: the start position for the modification as
#'     \code{integer} value. Usually \code{mod_start = mod_end}
#'     \item \code{mod_end}: the end position for the modification as
#'     \code{integer} value. Usually \code{mod_start = mod_end}
#'     \item \code{mod_strand}: the strand information for the modificaion as a
#'     \code{character} or \code{factor}.
#'     \item \code{sn_id}: a \code{integer} value per unique sequence
#'     \item \code{sn_name}: a \code{character} or \code{factor} as sequence
#'     name, e.g a chromosome or a transcript identifier like \code{chr1}.
#'   }
#'   The first six are mandatory, whereas one of the last two has to be set.
#'   \code{sn_id} will be generated from \code{sn_name}, if \code{sn_id} is not
#'   set.
#'
#' @param reactions An optional \code{data.frame} containg the following
#'   columns:
#'   \itemize{
#'     \item \code{mod_id}: a \code{integer} value per modification and the
#'     link to the \code{modification} \code{data.frame}.
#'     \item \code{rx_genename}: a \code{character} or \code{factor} referencing
#'     a genename for the enzyme incorporating the modification.
#'     \item \code{rx_rank}: a \code{integer} for sorting enzyme reactions, if
#'     multiple enzymes are involved in the modification's
#'     incorporation/maintenance.
#'     \item \code{rx_ensembl}: a \code{character} or \code{factor} with an
#'     ensembl identifier for the genename of the enzyme.
#'     \item \code{rx_ensembltrans}: a \code{character} or \code{factor} with an
#'     ensembl identifier for the transcript being translated into the enzyme.
#'     \item \code{rx_entrezid}: a \code{character} or \code{factor} with an
#'     entrezid for the genename of the enzyme.
#'   }
#'   (default: \code{reactions = NULL})
#'
#' @param specifiers An optional \code{data.frame} containg the following
#'   columns:
#'     \itemize{
#'     \item \code{mod_id}: a \code{integer} value per modification and the
#'     link to the \code{modification} \code{data.frame}.
#'     \item \code{spec_type}: a \code{character} or \code{factor}
#'     referencing a type of specifier, e.g. \code{snoRNA}. Not checked for
#'     validity.
#'     \item \code{spec_genename}: a \code{character} or \code{factor}
#'     referencing a genename for the specifier directing an enzyme to the
#'     specific location for the modification to be incorporated.
#'     \item \code{spec_ensembl}: a \code{character} or \code{factor} with an
#'     ensembl identifier for the genename of the specifier.
#'     \item \code{spec_ensembltrans}: a \code{character} or \code{factor} with an
#'     ensembl identifier for the transcript being translated into the specifier.
#'     \item \code{spec_entrezid}: a \code{character} or \code{factor} with an
#'     entrezid for the genename of the specifier.
#'   }
#'   (default: \code{specifiers = NULL})
#'
#' @param references An optional \code{data.frame} containg the following
#'   columns:
#'   \itemize{
#'     \item \code{mod_id}: a \code{integer} value per modification and the
#'     link to the \code{modification} \code{data.frame}.
#'     \item \code{ref_type}: a \code{character} or \code{factor} with a
#'     reference type, e.g. \code{PMID}. Is not checked for validity.
#'     \item \code{ref}: a \code{character} or \code{factor} with a reference
#'     value, e.g. a specific pubmed id or an journal article. Is not checked
#'     for validity.
#'   }
#'   (default: \code{references = NULL})
#'
#' @param metadata An optional \code{data.frame} containg the following columns:
#'   \itemize{
#'     \item \code{name}: a \code{character} value used as name
#'     \item \code{value}: a \code{character} value
#'   }
#'   This dataframe will be returned
#'   by \code{metadata()} (default: \code{metadata = NULL})
#'
#' @param reassign.ids \code{TRUE} or \code{FALSE} Controls how internal
#'   \code{mod_id}s should be assigned. If \code{reassign.ids} is \code{FALSE}
#'   (the default) and if the ids are supplied, then they are used as the
#'   internal ids, otherwise the internal ids are assigned in a way that is
#'   compatible with the order defined by ordering the modifications first by
#'   chromosome, then by strand, then by start, and finally by end.
#'
#' @seealso
#' \itemize{
#'   \item \code{\link[=makeEpiTxDbFromGRanges]{makeEpiTxDbFromGRanges}} for
#'     creating a \code{EpiTxDb} object from a
#'   \code{\link[GenomicRanges:GRanges-class]{GRanges}} object and it's metadata
#'     columns
#'   \item \code{\link[=makeEpiTxDbFromRMBase]{makeEpiTxDbFromRMBase}} for
#'     creating a \code{EpiTxDb} object from RMBase online resources
#'   \item \code{\link[=makeEpiTxDbFromtRNAdb]{makeEpiTxDbFromtRNAdb}} for
#'     creating a \code{EpiTxDb} object from tRNAdb online resources
#'   \item \code{\link[Modstrings:alphabet]{shortName}} and
#'   \code{\link[Modstrings:ModRNAString]{ModRNAString}} for information on
#'   \code{ModRNAString} objects.
#' }
#'
#' @return a \code{EpiTxDb} object.
#'
#' @export
#'
#' @examples
#' mod <- data.frame("mod_id" = 1L,
#'                   "mod_type" = "m1A",
#'                   "mod_name" = "m1A_1",
#'                   "mod_start" = 1L,
#'                   "mod_end" = 1L,
#'                   "mod_strand" = "+",
#'                   "sn_id" = 1L,
#'                   "sn_name" = "test")
#' rx <- data.frame(mod_id = 1L,
#'                  rx_genename = "test",
#'                  rx_rank = 1L,
#'                  rx_ensembl = "test",
#'                  rx_ensembltrans = "test",
#'                  rx_entrezid = "test")
#' spec <- data.frame(mod_id = 1L,
#'                    spec_type = "test",
#'                    spec_genename = "test",
#'                    spec_ensembl = "test",
#'                    spec_ensembltrans = "test",
#'                    spec_entrezid = "test")
#' ref <- data.frame(mod_id = 1L,
#'                   ref_type = "test",
#'                   ref = "test")
#' etdb <- makeEpiTxDb(mod,rx,spec,ref)
NULL

# helper functions -------------------------------------------------------------

insert_data_into_table <- function(conn, table, data){
    stopifnot(is.list(data))
    placeholders <- paste(rep.int("?", length(data)), collapse = ",")
    SQL <- sprintf("INSERT INTO %s VALUES (%s)", table, placeholders)
    ## dbExecute() emits annoying warnings if 'params' is a named list or if
    ## some of its list elements are factors.
    params <- unname(as.list(data))
    params <- lapply(params,
                     function(x) if (is.factor(x)) as.character(x) else x)
    dbExecute(conn, SQL, params = params)
}

duplicatedIntegerQuads <- S4Vectors:::duplicatedIntegerQuads
orderIntegerQuads <- S4Vectors:::orderIntegerQuads
matchIntegerQuads <- S4Vectors:::matchIntegerQuads

makeFeatureIds <- function(name = NULL, type, start, end,
                           same.id.for.dups = FALSE){
    a <- type
    b <- name
    c <- start
    d <- end
    if (!same.id.for.dups) {
        oo <- orderIntegerQuads(a, b, c, d)
        ans <- integer(length(oo))
        ans[oo] <- seq_len(length(oo))
        return(ans)
    }
    ## There should be a better way to do this...
    is_not_dup <- !duplicatedIntegerQuads(a, b, c ,d)
    ua <- a[is_not_dup]
    ub <- b[is_not_dup]
    uc <- c[is_not_dup]
    ud <- d[is_not_dup]
    oo <- orderIntegerQuads(ua, ub, uc, ud)
    ua <- ua[oo]
    ub <- ub[oo]
    uc <- uc[oo]
    ud <- ud[oo]
    matchIntegerQuads(a, b, c, d, ua, ub, uc, ud)
}

# check helper functions -------------------------------------------------------


.is_character_or_factor <- txdbmaker:::.is_character_or_factor
.check_foreign_key <- txdbmaker:::.check_foreign_key
translateIds <- txdbmaker:::translateIds
check_colnames <- txdbmaker:::check_colnames
has_col <- txdbmaker:::has_col
dbEasyQuery <- txdbmaker:::dbEasyQuery

.makeEpiTxDb_normarg_modifications <- function(modifications){
    .REQUIRED_COLS <- c("mod_id","mod_type","mod_start","mod_end","mod_strand",
                        "sn_id")
    .OPTIONAL_COLS <- c("mod_name", "sn_name")
    # make sure 'sn_id is set'
    if(!has_col(modifications, "sn_id") &&
       !has_col(modifications, "sn_name")){
        stop("'modifications' must contain a column 'sn_name', if no column ",
             "'sn_id' is given.")
    }
    if(!has_col(modifications, "sn_id") &&
       has_col(modifications, "sn_name")){
        modifications$sn_id <- as.integer(factor(modifications$sn_name,
                                                 unique(modifications$sn_name)))
    }
    check_colnames(modifications, .REQUIRED_COLS, .OPTIONAL_COLS,
                   "modifications")
    ## Check 'mod_id'.
    if (!is.integer(modifications$mod_id) || any(is.na(modifications$mod_id)))
        stop("'modifications$mod_id' must be an integer vector ",
             "with no NAs")
    if (any(duplicated(modifications$mod_id)))
        stop("'modifications$mod_id' contains duplicated values")
    ## Check 'mod_start'.
    if (!is.integer(modifications$mod_start) ||
        any(is.na(modifications$mod_start)))
        stop("'modifications$mod_start' must be an integer vector ",
             "with no NAs")
    ## Check 'mod_end'.
    if (!is.integer(modifications$mod_end) ||
        any(is.na(modifications$mod_end)))
        stop("'modifications$mod_end' must be an integer vector ",
             "with no NAs")
    ## Check 'mod_start <= mod_end'.
    if (any(modifications$mod_start > modifications$mod_end))
        stop("modification starts must be <= modification ends")
    ## Check 'mod_name'.
    if (!.is_character_or_factor(modifications$mod_name))
        stop("'modifications$mod_name' must be a character vector ",
             "(or factor)")
    ## Check 'mod_type'.
    if (!.is_character_or_factor(modifications$mod_type))
        stop("'modifications$mod_type' must be a character vector ",
             "(or factor)")
    if(!all(as.character(modifications$mod_type) %in% shortName(ModRNAString()))){
        stop("'modifications$mod_type' must be a valid modification shortName ",
             "(shortName(ModRNAString()))")
    }
    ## Check 'mod_strand'.
    if (!.is_character_or_factor(modifications$mod_strand) ||
        any(!(modifications$mod_strand %in% c("+","-","*"))))
        stop("'modifications$mod_strand' must be a character vector ",
             "(or factor) and contain only '+', '-' or '*'")
    ## Check 'sn_id'.
    if (!is.integer(modifications$sn_id) ||
        any(is.na(modifications$sn_id)))
        stop("'modifications$sn_id' must be a integer vector with no ",
             "NAs")
    ## Check 'sn_name'.
    if (has_col(modifications, "sn_name")){
        if(!.is_character_or_factor(modifications$sn_name) ||
           any(is.na(modifications$sn_name))){
            stop("'modifications$sn_name' must be a character vector ",
                 "(or factor) with no NAs")
        }
        if(length(unique(modifications$sn_id)) !=
           length(unique(modifications$sn_name))){
            stop("'modifications$sn_name' must match ",
                 "'modifications$sn_id' in meaning.")
        }
        f1 <- factor(modifications$sn_id,
                     unique(modifications$sn_id))
        p1 <- PartitioningByWidth(split(modifications$sn_id,f1))
        f2 <- factor(modifications$sn_name,
                     unique(modifications$sn_name))
        p2 <- PartitioningByWidth(split(modifications$sn_name,f2))
        if(!all(all(p1 == p2))){
            stop("'modifications$sn_name' must match ",
                 "'modifications$sn_id' in meaning.")
        }
    } else {
        modifications$sn_name <- character(1)
    }
    modifications
}

.makeEpiTxDb_normarg_reactions <- function(reactions, modifications_mod_id){
    if (is.null(reactions)) {
        reactions <- data.frame(mod_id = modifications_mod_id[FALSE],
                                rx_genename = character(0),
                                rx_rank = integer(0),
                                rx_ensembl = character(0),
                                rx_ensembltrans = character(0),
                                rx_entrezid = character(0),
                                check.names = FALSE, stringsAsFactors = FALSE)
        return(reactions)
    }
    .REQUIRED_COLS <- c("mod_id", "rx_genename", "rx_rank")
    .OPTIONAL_COLS <- c("rx_ensembl", "rx_ensembltrans", "rx_entrezid")
    check_colnames(reactions, .REQUIRED_COLS, .OPTIONAL_COLS, "reactions")
    ## Check 'mod_id'.
    .check_foreign_key(reactions$mod_id, "integer", "reactions$mod_id",
                       modifications_mod_id, "integer", "modifications$mod_id")
    ## Check 'rx_rank'.
    if (!is.integer(reactions$rx_rank)
        || any(is.na(reactions$rx_rank)))
        stop("'reactions$rx_rank' must be an integer vector ",
             "with no NAs")
    if (any(reactions$rx_rank <= 0L))
        stop("'reactions$rx_rank' contains non-positive values")
    ## Check uniqueness of (mod_id, rx_rank) pairs.
    if (any(S4Vectors::duplicatedIntegerPairs(reactions$mod_id,
                                              reactions$rx_rank)))
        stop("'reactions' must contain unique (mod_id, rx_rank) pairs")
    ## Check 'rx_genename'.
    if (has_col(reactions, "rx_genename")
        && !.is_character_or_factor(reactions$rx_genename))
        stop("'reactions$rx_genename' must be a character vector ",
             "(or factor)")
    ## Check 'rx_ensembl'.
    if (has_col(reactions, "rx_ensembl")){
        if(!.is_character_or_factor(reactions$rx_ensembl))
            stop("'reactions$rx_ensembl' must be a character vector ",
                 "(or factor)")
    } else {
        reactions$rx_ensembl <- character(1)
    }
    ## Check 'rx_ensembltrans'.
    if (has_col(reactions, "rx_ensembltrans")){
        if(!.is_character_or_factor(reactions$rx_ensembltrans))
            stop("'reactions$rx_ensembltrans' must be a character vector ",
                 "(or factor)")
    } else {
        reactions$rx_ensembltrans <- character(1)
    }
    ## Check 'rx_entrezid'.
    if (has_col(reactions, "rx_entrezid")){
        if(!.is_character_or_factor(reactions$rx_entrezid))
            stop("'reactions$rx_entrezid' must be a character vector ",
                 "(or factor)")
    } else {
        reactions$rx_entrezid <- character(1)
    }
    reactions
}

.makeEpiTxDb_normarg_specifiers <- function(specifier, modifications_mod_id){
    if (is.null(specifier)) {
        specifier <- data.frame(mod_id = modifications_mod_id[FALSE],
                                spec_type = character(0),
                                spec_genename = character(0),
                                spec_ensembl = character(0),
                                spec_ensembltrans = character(0),
                                spec_entrezid = character(0),
                                check.names = FALSE, stringsAsFactors = FALSE)
        return(specifier)
    }
    .REQUIRED_COLS <- c("mod_id", "spec_type", "spec_genename")
    .OPTIONAL_COLS <- c("spec_ensembl","spec_ensembltrans","spec_entrezid")
    check_colnames(specifier, .REQUIRED_COLS, .OPTIONAL_COLS, "specifier")
    ## Check 'mod_id'.
    .check_foreign_key(specifier$mod_id, "integer", "specifier$mod_id",
                       modifications_mod_id, "integer", "modifications$mod_id")
    ## Check 'spec_type'.
    if (has_col(specifier, "spec_type")
        && !.is_character_or_factor(specifier$spec_type))
        stop("'specifier$spec_type' must be a character vector ",
             "(or factor)")
    ## Check 'spec_genename'.
    if (has_col(specifier, "spec_genename")
        && !.is_character_or_factor(specifier$spec_genename))
        stop("'specifier$spec_genename' must be a character vector ",
             "(or factor)")
    ## Check 'spec_ensembl'.
    if (has_col(specifier, "spec_ensembl")){
        if(!.is_character_or_factor(specifier$spec_ensembl))
            stop("'specifier$spec_ensembl' must be a character vector ",
                 "(or factor)")
    } else {
        specifier$spec_ensembl <- character(1)
    }
    ## Check 'spec_ensembltrans'.
    if (has_col(specifier, "spec_ensembltrans")){
        if(!.is_character_or_factor(specifier$spec_ensembltrans))
            stop("'specifier$spec_ensembltrans' must be a character vector ",
                 "(or factor)")
    } else {
        specifier$spec_ensembltrans <- character(1)
    }
    ## Check 'spec_entrezid'.
    if (has_col(specifier, "spec_entrezid")){
        if(!.is_character_or_factor(specifier$spec_entrezid))
            stop("'specifier$spec_entrezid' must be a character vector ",
                 "(or factor)")
    } else {
        specifier$spec_entrezid <- character(1)
    }
    specifier
}

.makeEpiTxDb_normarg_references <- function(references, modifications_mod_id){
    if (is.null(references)) {
        references <- data.frame(mod_id = modifications_mod_id[FALSE],
                                 ref_type = character(0),
                                 ref = character(0),
                                 check.names = FALSE, stringsAsFactors = FALSE)
        return(references)
    }
    .REQUIRED_COLS <- c("mod_id")
    .OPTIONAL_COLS <- c("ref_type", "ref")
    check_colnames(references, .REQUIRED_COLS, .OPTIONAL_COLS, "reference")
    ## Check 'mod_id'.
    .check_foreign_key(references$mod_id, "integer", "references$mod_id",
                       modifications_mod_id, "integer", "modifications$mod_id")
    ## Check 'ref_type'.
    if (has_col(references, "ref_type")){
        if(!.is_character_or_factor(references$ref_type)){
            stop("'references$ref_type' must be a character vector ",
                 "(or factor)")
        }
    } else {
        references$ref_type <- character(1)
    }
    ## Check 'ref'.
    if (has_col(references, "ref")){
        if(!.is_character_or_factor(references$ref)){
            stop("'references$ref' must be a character vector ",
                 "(or factor)")
        }
    } else {
        references$ref <- character(1)
    }
    references
}

.makeEpiTxDb_normarg_metadata <- function(metadata)
{
    if (!is.null(metadata)) {
        if (!is.data.frame(metadata))
            stop("'metadata' must be NULL or a data.frame")
        if (!setequal(colnames(metadata), c("name", "value")))
            stop("'metadata' columns must be \"name\" and \"value\"")
    }
    metadata
}

# id helper functions ----------------------------------------------------------

.make_modifications_internal_mod_id <- function(modifications, reassign.ids)
{
    if (!reassign.ids)
        return(modifications$mod_id)
    makeFeatureIds(name = factor(modifications$mod_name,
                                 unique(modifications$mod_name)),
                   type = factor(modifications$mod_type,
                                 unique(modifications$mod_type)),
                   start = factor(modifications$mod_start,
                                  unique(modifications$mod_start)),
                   end = factor(modifications$mod_end,
                                unique(modifications$mod_end)))
}

.make_ids <- function(df, cols){
    df[is.na(df)] <- ""
    id <- factor(do.call(paste,as.list(df[,colnames(df) %in% cols])))
    as.integer(id)
}

.make_reactions_ids <- function(reactions){
    cols <- c("rx_genename","rx_rank","rx_ensembl",
              "rx_ensembltrans","rx_entrezid")
    reactions$rx_id <- .make_ids(reactions, cols)
    reactions
}

.make_specifiers_ids <- function(specifiers){
    cols <- c("spec_type","spec_genename","spec_ensembl",
              "spec_ensembltrans","spec_entrezid")
    specifiers$spec_id <- .make_ids(specifiers, cols)
    specifiers
}

.make_references_ids <- function(references){
    cols <- c("ref_type","ref")
    references$ref_id <- .make_ids(references, cols)
    references
}

.shrink_df <- function(df, col){
    df[!duplicated(df[,colnames(df) %in% col]),]
}

# write table helper functions -------------------------------------------------

.write_seqnames_table <- function(conn,
                                  sn_id,
                                  sn_name){
    data <- data.frame(sn_id = sn_id,
                       sn_name = sn_name,
                       check.names = FALSE, stringsAsFactors = FALSE)
    data <- unique(data)

    ## Create the table.
    SQL <- build_SQL_CREATE_seqnames_table()
    dbExecute(conn, SQL)

    ## Fill the table.
    insert_data_into_table(conn, "seqnames", data)
}

.write_modification_table <- function(conn,
                                      mod_internal_id,
                                      mod_type,
                                      mod_name,
                                      mod_start,
                                      mod_end,
                                      mod_strand,
                                      sn_id){
    if (is.null(mod_name))
        mod_name <- rep.int(NA_character_, length(mod_internal_id))
    data <- data.frame(mod_internal_id = mod_internal_id,
                       mod_type = mod_type,
                       mod_name = mod_name,
                       mod_start = mod_start,
                       mod_end = mod_end,
                       mod_strand = mod_strand,
                       sn_id = sn_id,
                       check.names = FALSE, stringsAsFactors = FALSE)
    data <- unique(data)

    ## Create the table.
    SQL <- build_SQL_CREATE_modification_table()
    dbExecute(conn, SQL)

    ## Fill the table.
    insert_data_into_table(conn, "modification", data)
}

.write_reaction_table <- function(conn,
                                  rx_id,
                                  rx_genename,
                                  rx_rank,
                                  rx_ensembl,
                                  rx_ensembltrans,
                                  rx_entrezid){
    data <- data.frame(rx_id = rx_id,
                       rx_genename = rx_genename,
                       rx_rank = rx_rank,
                       rx_ensembl = rx_ensembl,
                       rx_ensembltrans = rx_ensembltrans,
                       rx_entrezid = rx_entrezid,
                       check.names = FALSE, stringsAsFactors = FALSE)

    ## Create the 'reaction' table and related indices.
    SQL <- build_SQL_CREATE_reaction_table()
    dbExecute(conn, SQL)

    ## Fill the 'reaction' table.
    insert_data_into_table(conn, "reaction", data)
}

.write_specifier_table <- function(conn,
                                   spec_id,
                                   spec_type,
                                   spec_genename,
                                   spec_ensembl,
                                   spec_ensembltrans,
                                   spec_entrezid){
    data <- data.frame(spec_id = spec_id,
                       spec_type = spec_type,
                       spec_genename = spec_genename,
                       spec_ensembl = spec_ensembl,
                       spec_ensembltrans = spec_ensembltrans,
                       spec_entrezid = spec_entrezid,
                       check.names = FALSE, stringsAsFactors = FALSE)

    ## Create the 'specifier' table and related indices.
    SQL <- build_SQL_CREATE_specifier_table()
    dbExecute(conn, SQL)

    ## Fill the 'specifier' table.
    insert_data_into_table(conn, "specifier", data)
}

.write_reference_table <- function(conn,
                                   ref_id,
                                   ref_type,
                                   ref){
    data <- data.frame(ref_id = ref_id,
                       ref_type = ref_type,
                       ref = ref,
                       check.names = FALSE, stringsAsFactors = FALSE)

    ## Create the 'reference' table and related indices.
    SQL <- build_SQL_CREATE_reference_table()
    dbExecute(conn, SQL)

    ## Fill the 'reference' table.
    insert_data_into_table(conn, "reference", data)
}

.write_modification2reaction_table <- function(conn,
                                                mod_id,
                                                rx_id){
    data <- data.frame(mod_id = mod_id,
                       rx_id = rx_id,
                       check.names = FALSE, stringsAsFactors = FALSE)

    ## Create the 'modification2reaction' table and related indices.
    SQL <- build_SQL_CREATE_modification2reaction_table()
    dbExecute(conn, SQL)

    ## Fill the 'modification2reaction' table.
    insert_data_into_table(conn, "modification2reaction", data)
}

.write_modification2specifier_table <- function(conn,
                                                mod_id,
                                                spec_id){
    data <- data.frame(mod_id = mod_id,
                       spec_id = spec_id,
                       check.names = FALSE, stringsAsFactors = FALSE)

    ## Create the 'modification2specifier' table and related indices.
    SQL <- build_SQL_CREATE_modification2specifier_table()
    dbExecute(conn, SQL)

    ## Fill the 'modification2specifier' table.
    insert_data_into_table(conn, "modification2specifier", data)
}

.write_modification2reference_table <- function(conn,
                                                mod_id,
                                                ref_id){
    data <- data.frame(mod_id = mod_id,
                       ref_id = ref_id,
                       check.names = FALSE, stringsAsFactors = FALSE)

    ## Create the 'modification2reference' table and related indices.
    SQL <- build_SQL_CREATE_modification2reference_table()
    dbExecute(conn, SQL)

    ## Fill the 'modification2reference' table.
    insert_data_into_table(conn, "modification2reference", data)
}

#' @importFrom utils packageDescription
.write_metadata_table <- function(conn, metadata){
    nb_modifications <- dbEasyQuery(conn,
                                    "SELECT COUNT(*) FROM modification")[[1L]]
    thispkg_version <- utils::packageDescription("EpiTxDb")$Version
    rsqlite_version <- utils::packageDescription("RSQLite")$Version
    mat1 <- matrix(c(
        DB_TYPE_NAME, DB_TYPE_VALUE,
        "Supporting package", "EpiTxDb"),
        ncol = 2, byrow = TRUE
    )
    mat2 <- matrix(c(
        "Nb of modifications", nb_modifications,
        "Db created by",     "EpiTxDb package from Bioconductor",
        "Creation time",     svn.time(),
        "EpiTxDb version at creation time", thispkg_version,
        "RSQLite version at creation time", rsqlite_version,
        "DBSCHEMAVERSION",   DB_SCHEMA_VERSION),
        ncol = 2, byrow = TRUE
    )
    colnames(mat1) <- colnames(mat2) <- c("name", "value")
    metadata <- rbind(data.frame(name = mat1[ , "name"], value = mat1[ , "value"],
                                 check.names = FALSE, stringsAsFactors = FALSE),
                      metadata,
                      data.frame(name = mat2[ , "name"], value = mat2[ , "value"],
                                 check.names = FALSE, stringsAsFactors = FALSE))
    dbWriteTable(conn, "metadata", metadata, row.names = FALSE)
}


# makeEpiTxDb------------------------------------------------------------------

#' @rdname makeEpiTxDb
#' @export
makeEpiTxDb <- function(modifications, reactions = NULL, specifiers = NULL,
                        references = NULL, metadata = NULL,
                        reassign.ids = FALSE){
    message("Creating EpiTxDb object ... ", appendLF = FALSE)
    if (!isTRUEorFALSE(reassign.ids)){
        stop("'reassign.ids' must be TRUE or FALSE")
    }
    #
    modifications <- .makeEpiTxDb_normarg_modifications(modifications)
    reactions <- .makeEpiTxDb_normarg_reactions(reactions, modifications$mod_id)
    specifiers <- .makeEpiTxDb_normarg_specifiers(specifiers, modifications$mod_id)
    references <- .makeEpiTxDb_normarg_references(references, modifications$mod_id)
    metadata <- .makeEpiTxDb_normarg_metadata(metadata)
    #
    internal_mod_id <- .make_modifications_internal_mod_id(modifications,
                                                      reassign.ids)
    mod2rx_mod_id <- translateIds(modifications$mod_id,
                                  internal_mod_id,
                                  reactions$mod_id)
    mod2spec_mod_id <- translateIds(modifications$mod_id,
                                    internal_mod_id,
                                  specifiers$mod_id)
    mod2ref_mod_id <- translateIds(modifications$mod_id,
                                   internal_mod_id,
                                  references$mod_id)
    #
    reactions <- .make_reactions_ids(reactions)
    specifiers <- .make_specifiers_ids(specifiers)
    references <- .make_references_ids(references)
    mod2rx_rx_id <- reactions$rx_id
    mod2spec_spec_id <- specifiers$spec_id
    mod2ref_ref_id <- references$ref_id
    reactions <- .shrink_df(reactions, "rx_id")
    specifiers <- .shrink_df(specifiers, "spec_id")
    references <- .shrink_df(references, "ref_id")
    ## Create the db in a temp file.
    conn <- dbConnect(SQLite(), dbname="")
    .write_seqnames_table(conn,
                          modifications$sn_id,
                          modifications$sn_name)
    .write_modification_table(conn,
                              internal_mod_id,
                              modifications$mod_type,
                              modifications$mod_name,
                              modifications$mod_start,
                              modifications$mod_end,
                              modifications$mod_strand,
                              modifications$sn_id)
    .write_reaction_table(conn,
                          reactions$rx_id,
                          reactions$rx_genename,
                          reactions$rx_rank,
                          reactions$rx_ensembl,
                          reactions$rx_ensembltrans,
                          reactions$rx_entrezid)
    .write_specifier_table(conn,
                           specifiers$spec_id,
                           specifiers$spec_type,
                           specifiers$spec_genename,
                           specifiers$spec_ensembl,
                           specifiers$spec_ensembltrans,
                           specifiers$spec_entrezid)
    .write_reference_table(conn,
                           references$ref_id,
                           references$ref_type,
                           references$ref)
    # write link tables
    .write_modification2reaction_table(conn,
                                       mod2rx_mod_id,
                                       mod2rx_rx_id)

    .write_modification2specifier_table(conn,
                                        mod2spec_mod_id,
                                        mod2spec_spec_id)

    .write_modification2reference_table(conn,
                                        mod2ref_mod_id,
                                        mod2ref_ref_id)
    .write_metadata_table(conn, metadata)
    ans <- EpiTxDb(conn)
    message("done")
    ans
}

context("makeEpiTxDb")
test_that("makeEpiTxDb:",{
    mod <- data.frame("mod_id" = 1L,
                      "mod_type" = "m1A",
                      "mod_name" = "m1A_1",
                      "mod_start" = 1L,
                      "mod_end" = 1L,
                      "mod_strand" = "+",
                      "sn_id" = 1L,
                      "sn_name" = "test")
    expect_equal(EpiTxDb:::.makeEpiTxDb_normarg_modifications(mod),mod)
    df <- mod
    df[,"mod_id"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_modifications(df),
                 "'modifications\\$mod_id' must be an integer")
    df <- mod
    df[,"mod_type"] <- "Z"
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_modifications(df),
                 "'modifications\\$mod_type' must be a valid")
    df <- mod
    df[,"mod_name"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_modifications(df),
                 "'modifications\\$mod_name' must be a character")
    df <- mod
    df[,"mod_start"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_modifications(df),
                 "'modifications\\$mod_start' must be an integer")
    df <- mod
    df[,"mod_end"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_modifications(df),
                 "'modifications\\$mod_end' must be an integer")
    df <- mod
    df[,"mod_start"] <- 3L
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_modifications(df),
                 "modification starts must be <= modification ends")
    df <- mod
    df[,"mod_strand"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_modifications(df),
                 "'modifications\\$mod_strand' must be a character")
    df <- mod
    df[,"sn_id"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_modifications(df),
                 "'modifications\\$sn_id' must be a integer")
    df <- mod
    df[,"sn_name"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_modifications(df),
                 "'modifications\\$sn_name' must be a character")
    #
    rx <- data.frame(mod_id = 1L,
                     rx_genename = "test",
                     rx_rank = 1L,
                     rx_ensembl = "test",
                     rx_ensembltrans = "test",
                     rx_entrezid = "test")
    expect_equal(EpiTxDb:::.makeEpiTxDb_normarg_reactions(rx,1L),rx)
    df <- rx
    df[,"mod_id"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_reactions(df,1L),
                 "'reactions\\$mod_id' must be of type integer")
    df <- rx
    df[,"rx_genename"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_reactions(df,1L),
                 "'reactions\\$rx_genename' must be a character")
    df <- rx
    df[,"rx_rank"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_reactions(df,1L),
                 "'reactions\\$rx_rank' must be an integer vector")
    df <- rx
    df[,"rx_ensembl"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_reactions(df,1L),
                 "'reactions\\$rx_ensembl' must be a character")
    df <- rx
    df[,"rx_ensembltrans"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_reactions(df,1L),
                 "'reactions\\$rx_ensembltrans' must be a character")
    df <- rx
    df[,"rx_entrezid"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_reactions(df,1L),
                 "'reactions\\$rx_entrezid' must be a character")
    #
    spec <- data.frame(mod_id = 1L,
                       spec_type = "test",
                       spec_genename = "test",
                       spec_ensembl = "test",
                       spec_ensembltrans = "test",
                       spec_entrezid = "test")
    expect_equal(EpiTxDb:::.makeEpiTxDb_normarg_specifiers(spec,1L),spec)
    df <- spec
    df[,"mod_id"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_specifiers(df,1L),
                 "'specifier\\$mod_id' must be of type integer")
    df <- spec
    df[,"spec_type"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_specifiers(df,1L),
                 "'specifier\\$spec_type' must be a character")
    df <- spec
    df[,"spec_genename"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_specifiers(df,1L),
                 "'specifier\\$spec_genename' must be a character")
    df <- spec
    df[,"spec_ensembl"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_specifiers(df,1L),
                 "'specifier\\$spec_ensembl' must be a character")
    df <- spec
    df[,"spec_ensembltrans"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_specifiers(df,1L),
                 "'specifier\\$spec_ensembltrans' must be a character")
    df <- spec
    df[,"spec_entrezid"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_specifiers(df,1L),
                 "'specifier\\$spec_entrezid' must be a character")
    #
    ref <- data.frame(mod_id = 1L,
                      ref_type = "test",
                      ref = "test")
    expect_equal(EpiTxDb:::.makeEpiTxDb_normarg_references(ref,1L),ref)
    df <- ref
    df[,"mod_id"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_references(df,1L),
                 "'references\\$mod_id' must be of type integer")
    df <- ref
    df[,"ref_type"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_references(df,1L),
                 "'references\\$ref_type' must be a character")
    df <- ref
    df[,"ref"] <- 1
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_references(df,1L),
                 "'references\\$ref' must be a character")
    #
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_metadata(),
                 'argument "metadata" is missing')
    expect_equal(EpiTxDb:::.makeEpiTxDb_normarg_metadata(NULL),NULL)
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_metadata("1"),
                 "'metadata' must be NULL or a data.frame")
    expect_error(EpiTxDb:::.makeEpiTxDb_normarg_metadata(data.frame(a = 1,
                                                                    b = 2)),
                 "'metadata' columns must be \"name\" and \"value\"")
    actual <- EpiTxDb:::.makeEpiTxDb_normarg_metadata(data.frame(name = 1,
                                                                 value = 2))
    expect_s3_class(actual,"data.frame")
    expect_equal(colnames(actual),c("name","value"))
    #
    mod2 <- data.frame("mod_id" = 2L,
                       "mod_type" = "m1A",
                       "mod_name" = "m1A_1",
                       "mod_start" = 1L,
                       "mod_end" = 1L,
                       "mod_strand" = "+",
                       "sn_id" = 1L,
                       "sn_name" = "test")
    actual <- EpiTxDb:::.make_modifications_internal_mod_id(mod, FALSE)
    expect_equal(actual, 1L)
    actual <- EpiTxDb:::.make_modifications_internal_mod_id(mod2, FALSE)
    expect_equal(actual, 2L)
    actual <- EpiTxDb:::.make_modifications_internal_mod_id(mod, TRUE)
    expect_equal(actual, 1L)
    actual <- EpiTxDb:::.make_modifications_internal_mod_id(mod2, TRUE)
    expect_equal(actual, 1L)
    #
    rx2 <- rx
    rx2$rx_id <- 2L
    spec2 <- spec
    spec2$spec_id <- 2L
    ref2 <- ref
    ref2$ref_id <- 2L
    expect_equal(EpiTxDb:::.make_reactions_ids(rx2)$rx_id,1L)
    expect_equal(EpiTxDb:::.make_specifiers_ids(spec2)$spec_id,1L)
    expect_equal(EpiTxDb:::.make_references_ids(ref2)$ref_id,1L)
    expect_equal(EpiTxDb:::.shrink_df(rbind(rx2,rx2),"rx_id"),rx2)
    #
    etdb <- makeEpiTxDb(mod,rx,spec,ref)
    expect_s4_class(etdb,"EpiTxDb")
    expect_s3_class(metadata(etdb),"data.frame")
    dbDisconnect(etdb$conn)
})

# Low-level helpers imported from GenomicFeatures ------------------------------

#' @importFrom GenomicFeatures transcripts
.as_qualified <- GenomicFeatures:::.as_qualified
.tables_in_joins <- GenomicFeatures:::.tables_in_joins
.build_SQL_SELECT <- GenomicFeatures:::.build_SQL_SELECT
queryAnnotationDb <- txdbmaker:::queryAnnotationDb

# .EPITXDB_join_tables() -------------------------------------------------------

.EPITXDB_add_table_bundle <- function(tables){
  if (length(tables) == 1L)
    return(tables)
  skip_tables <- c("modification",  "seqnames")
  skip_tables <- tables %in% skip_tables
  if(all(skip_tables)){
    return(tables)
  }
  tables <- c(tables[skip_tables],
              pc(paste0("modification2",tables[!skip_tables]),
                 tables[!skip_tables]))
  unlist(tables)
}

.EPITXDB_join_tables <- function(tables, joinColumn = "_mod_id")
{
    # browser()
    tables <- unique(tables)
    if (length(tables) == 1L)
        return(tables)
    joinColumn <- unique(joinColumn)
    tables <- .EPITXDB_add_table_bundle(tables)
    joinColumn <- unique(joinColumn)
    stopifnot(length(tables) == length(joinColumn) + (length(tables) - 1L) %/% 2)
    ## Order tables & remove duplicates.
    join_order <- c("modification", "seqnames",
                    "modification2reaction", "reaction",
                    "modification2specifier", "specifier",
                    "modification2reference", "reference")
    joinColumn_order <- c("_sn_id","_mod_id","_rx_id","_spec_id","_ref_id")
    tables <- intersect(join_order, tables)
    joinColumn <- intersect(joinColumn_order, joinColumn)
    jc_offset_tx <- sum("_sn_id" %in% joinColumn)
    jc_offset_sel <- c("2"=1,"3"=2,"4"=3,"5"=3,"6"=4,"7"=4,"8"=5)
    joins <- character(2L * length(tables) - 1L)
    ON_idx <- 2L * seq_len(length(tables) - 1L)
    ON <- vapply(2:length(tables), function(i) {
        Rtable <- tables[[i]]
        if(i%%2 == jc_offset_tx){
            USING <- joinColumn[[1L+jc_offset_tx]]
            Ltable <- tables[[1L]]
        } else {
            USING <- joinColumn[[jc_offset_sel[[as.character(i)]]]]
            Ltable <- tables[[i-1L]]
        }
        Lcolumn <- .as_qualified(Ltable, USING)
        Rcolumn <- .as_qualified(Rtable, USING)
        paste(Lcolumn, Rcolumn, sep="=")
    },character(1))
    joins[ON_idx] <- ON
    joins[c(1L, ON_idx + 1L)] <- tables
    joins
}

EpiTxDb_SELECT_from_LEFT_JOIN <- function(epitxdb, table, columns,
                                          filter = list(),
                                          orderby = character(0),
                                          joinColumn = character(0))
{
    schema_version <- EpiTxDb_schema_version(epitxdb)
    tables <- EPITXDB_column2table(columns, from_table = table,
                                   schema_version = schema_version)
    if(!("modification" %in% tables)){
        tables <- c("modification",tables)
    }
    where_columns <- names(filter)
    where_tables <- EPITXDB_column2table(where_columns, from_table = table,
                                         schema_version = schema_version)
    joinColumn <- EPITXDB_table2joinColumns(tables = c(table, tables, where_tables),
                                            schema_version = schema_version)
    joins <- .EPITXDB_join_tables(c(table, tables, where_tables), joinColumn)
    orderby_tables <- EPITXDB_column2table(orderby, from_table = table,
                                           schema_version = schema_version)
    stopifnot(all(orderby_tables %in% .tables_in_joins(joins)))
    use_joins <- length(joins) > 1L
    if (use_joins) {
        columns <- .as_qualified(tables[names(tables) != ""], columns)
        names(filter) <- .as_qualified(where_tables, where_columns)
        orderby <- .as_qualified(orderby_tables, orderby)
    }
    ## .build_SQL_SELECT() uses LEFT joins.
    SQL <- .build_SQL_SELECT(columns, joins, distinct = use_joins,
                             filter = filter, orderby = orderby)
    queryAnnotationDb(epitxdb, SQL)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

EpiTxDb_SELECT_from_modification <- function(epitxdb, filter = list(),
                                             orderby = c("_mod_id", "_sn_id"))
{
    schema_version <- EpiTxDb_schema_version(epitxdb)
    columns <- EPITXDB_table_columns("modification",
                                     schema_version = schema_version)
    EpiTxDb_SELECT_from_LEFT_JOIN(epitxdb, "modification", columns,
                                  filter = filter, orderby = orderby)
}
EpiTxDb_SELECT_from_seqnames <- function(epitxdb, filter = list(),
                                         orderby = "_sn_id")
{
    schema_version <- EpiTxDb_schema_version(epitxdb)
    columns <- EPITXDB_table_columns("seqnames",
                                     schema_version = schema_version)
    EpiTxDb_SELECT_from_LEFT_JOIN(epitxdb, "seqnames", columns,
                                  filter = filter, orderby = orderby)
}

EpiTxDb_SELECT_from_reaction <- function(epitxdb, filter = list(),
                                         orderby = c("_rx_id", "rx_rank"))
{
    schema_version <- EpiTxDb_schema_version(epitxdb)
    columns <- EPITXDB_table_columns("reaction",
                                     schema_version = schema_version)
    EpiTxDb_SELECT_from_LEFT_JOIN(epitxdb, "reaction", columns,
                                  filter = filter, orderby = orderby)
}

EpiTxDb_SELECT_from_specifier <- function(epitxdb, filter = list(),
                                          orderby = c("_spec_id"))
{
    schema_version <- EpiTxDb_schema_version(epitxdb)
    columns <- EPITXDB_table_columns("specifier",
                                     schema_version = schema_version)
    EpiTxDb_SELECT_from_LEFT_JOIN(epitxdb, "specifier", columns,
                                  filter = filter, orderby = orderby)
}

EpiTxDb_SELECT_from_reference <- function(epitxdb, filter = list(),
                                          orderby = c("_ref_id"))
{
    schema_version <- EpiTxDb_schema_version(epitxdb)
    columns <- EPITXDB_table_columns("reference",
                                     schema_version = schema_version)
    EpiTxDb_SELECT_from_LEFT_JOIN(epitxdb, "reference", columns,
                                  filter = filter, orderby = orderby)
}

#' @include EpiTxDb-class.R
NULL

.format_feature_columns <- function(epitxdb, columns, by){
    abbr <- .makeColAbbreviations(epitxdb)
    user_columns <- abbr[abbr %in% columns]
    names(user_columns) <- gsub("^_","",names(user_columns))
    names(user_columns) <- gsub("^mod_type","mod",names(user_columns))
    user_columns
}

.split_into_GRL <- function(epitxdb, gr, columns, by){
    gr_mcols <- mcols(gr)
    colnames(gr_mcols)[match(columns,colnames(gr_mcols))] <- names(columns)
    mcols(gr) <- gr_mcols
    FUN <- function(x){
        x <- unique(x)
        x <- vapply(x,paste,character(1),collapse = "-")
        x[x == ""] <- "unknown"
        x
    }
    f <- switch(by,
                "seqnames" = seqnames(gr),
                "modtype" = gr_mcols$mod,
                "reaction" = FUN(gr_mcols$rx_genename),
                "specifier" = FUN(gr_mcols$spec_genename),
                "specifiertype" = FUN(gr_mcols$spec_type),
                stop("unsupported 'by' value."))
    grl <- S4Vectors::split(gr, f)
    .assignMetadataList(grl, epitxdb)
}


.extract_features_by <- function(epitxdb, by = "seqnames"){
    ans <- switch(by,
                  "seqnames" = modifications(epitxdb,
                                             c("MODID", "MODTYPE","MODNAME")),
                  "modtype" = modifications(epitxdb,
                                             c("MODTYPE", "MODID", "MODNAME")),
                  "reaction" = modifications(epitxdb,
                                             c("RXENSEMBL", "RXENSEMBLTRANS",
                                               "RXENTREZID", "RXGENENAME")),
                  "specifier" = modifications(epitxdb,
                                              c("SPECENSEMBL","SPECENTREZID",
                                                "SPECGENENAME", "SPECTYPE")),
                  "specifiertype" = modifications(epitxdb,
                                                   c("SPECENSEMBL","SPECENTREZID",
                                                     "SPECGENENAME", "SPECTYPE")),
                  stop("unsupported 'by' value."))
    mcolumns <- .format_feature_columns(epitxdb, colnames(mcols(ans)), by)
    .split_into_GRL(epitxdb, ans, mcolumns, by)
}

#' @rdname modifications
#' @export
setMethod("modificationsBy", "EpiTxDb",
    function(x, by = c("seqnames","modtype","reaction","specifier",
                       "specifiertype")){
      by <- match.arg(by)
      .extract_features_by(x, by = by)
    }
)


#' @rdname modifications
#' @export
setMethod("modifiedSeqsByTranscript", c("EpiTxDb","DNAStringSet"),
    function(x, sequences){
      grl <- modificationsBy(x, "seqnames")
      combineIntoModstrings(sequences, grl)
    }
)

#' @include EpiTxDb-class.R
#' @include makeEpiTxDb.R
NULL

#' @name makeEpiTxDbFromtRNAdb
#'
#' @title Create a \code{EpiTxDb} object from tRNAdb resources
#'
#' @description
#' \code{makeEpiTxDbFromtRNAdb} will make use of the tRNAdb online
#' resources and extract the modification information from the RNA database.
#'
#' If a named \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} is
#' provided as \code{sequences}, the result from the tRNAdb will be matched
#' against the sequences. Valid matches will be used as transcript identifiers
#' and returned after a check of modification compatibility with the provided
#' sequence. By this process multiple copies of transcripts can be associated
#' with a single modification.
#'
#' \code{makeEpiTxDbFromtRNAdb} uses the functions provided by the
#' \code{\link[tRNAdbImport:tRNAdbImport]{tRNAdbImport}} package.
#' \code{\link[tRNAdbImport:import.tRNAdb]{import.tRNAdb}} will be used with
#' \code{database = "RNA"} and the three different values for \code{origin}.
#'
#' @param organism A \code{character} value for an organism available on the
#'   tRNAdb website.
#' @param sequences A named \code{DNAStringSet} or \code{RNAStringSet}, which
#'   will be used to associate a tRNAdb result with a specific transcript.
#' @param dbURL The URL to the tRNA db website.
#' @param metadata See \code{\link[=makeEpiTxDb]{makeEpiTxDb}}
#'
#' @references
#' Juehling F, Moerl M, Hartmann RK, Sprinzl M, Stadler PF, Puetz J. 2009.
#' "tRNAdb 2009: compilation of tRNA sequences and tRNA genes." Nucleic Acids
#' Research, Volume 37 (suppl_1): D159–162. doi:10.1093/nar/gkn772.
#'
#' @return a \code{EpiTxDb} object.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # getting just the annotation data
#' etdb <- makeEpiTxDbFromtRNAdb("Saccharomyces cerevisiae")
#'
#' # For associating the result with transcripts, provide and additional
#' # named DNAStringSet object. Matching will be done against each sequence
#' # allowing 5 mismatches and indels. The final result will be checked for
#' # validity regarding the identity of the modifications
#' etdb <- makeEpiTxDbFromtRNAdb("Saccharomyces cerevisiae",
#'                               some_transcript_sequences)
#' }
NULL

# makeEpiTxDbFromtRNAdb --------------------------------------------------------

.add_reference_information_tRNAdb <- function(mod, gr){
    # collect references
    f_ref <- mcols(gr)$tRNAdb_reference != ""
    f_pmid <- mcols(gr)$tRNAdb_pmid != ""

    ref_type <- list(IRanges::CharacterList(as.list(rep("tRNAdb_ID",
                                                        length(gr)))),
                     IRanges::CharacterList(as.list(rep("tRNAdb_REF",
                                                        length(gr)))),
                     IRanges::CharacterList(as.list(rep("PMID",
                                                        length(gr)))))
    ref_type[[2]][!f_ref] <- ""
    ref_type[[3]][!f_pmid] <- ""
    ref_type <- do.call(pc,ref_type)
    ref_type <- ref_type[ref_type != ""]

    references <- list(mcols(gr)$tRNAdb_ID,
                       mcols(gr)$tRNAdb_reference,
                       mcols(gr)$tRNAdb_pmid)
    references[[2]][!f_ref] <- ""
    references[[3]][!f_pmid] <- ""
    references <- do.call(pc,references)
    references <- references[references != ""]

    m <- as.integer(S4Vectors::match(seqnames(mod), seqnames(gr)))
    mcols(mod)$ref_type <- ref_type[m]
    mcols(mod)$ref <- references[m]
    mod
}

#' @importFrom Biostrings subseq vwhichPDict
.map_modifications_to_sequences <- function(mod, gr, sequences){
    seq <- as(gr$tRNA_seq,"RNAStringSet")
    width <- nchar(seq)
    width[gr$tRNA_CCA.end] <- width[gr$tRNA_CCA.end] - 3L
    seq <- Biostrings::subseq(seq, 1L, width)
    # remove sequences which are definitly to long
    max_length <- max(lengths(seq))
    min_length <- min(lengths(seq))
    sequences[lengths(sequences) > (max_length + 50)] <-
        do.call(class(sequences),
                list(paste0(rep("A",(max_length + 50)),collapse = "")))
    sequences[lengths(sequences) < min_length] <-
        do.call(class(sequences),
                list(paste0(rep("A",(max_length + 50)),collapse = "")))
    hits <- Biostrings::vwhichPDict(sequences, seq,
                                    with.indels = TRUE, max.mismatch = 5L)
    hits <- Hits(unlist(hits),as.integer(unlist(Map(rep.int,seq_along(hits),lengths(hits)))),
                 length(sequences), length(seq))
    hits_mod <- findMatches(seqnames(mod),names(seq))
    # transfer names of sequences to seqnames for modifications and expand the
    # result with multiple matches
    sn_name <- names(sequences)[queryHits(hits)]
    sn_name <- IRanges::CharacterList(split(sn_name,subjectHits(hits)))
    sn_name <- vapply(sn_name,paste,character(1),collapse=",")
    hits_mod <- hits_mod[subjectHits(hits_mod) %in% unique(subjectHits(hits))]
    mcols(mod)[queryHits(hits_mod),"sn_name"] <- sn_name[as.character(subjectHits(hits_mod))]
    sn_name <- IRanges::CharacterList(strsplit(as.character(mcols(mod)$sn_name),","))
    # expand results
    mod <- mod[unlist(Map(rep,seq_along(sn_name),lengths(sn_name)))]
    # assemble new result
    mcols <- mcols(mod)[,colnames(mcols(mod)) != "sn_name"]
    mod <- GenomicRanges::GRanges(seqnames = unlist(sn_name),
                                  ranges = IRanges::ranges(mod),
                                  strand = strand(mod),
                                  mcols)
    mcols(mod)$mod_id <- seq_along(mod)
    mod
}

.add_tRNAdb_metadata_columns <- function(gr){
    # assemble metadata columns
    mcols <- mcols(gr)
    colnames(mcols) <- gsub("^mod$","mod_type",colnames(mcols))
    mcols$mod_name <- paste0(mcols$mod_type,"_",start(gr),"_",seqnames(gr))
    mcols$mod_id <- seq_len(nrow(mcols))
    mcols(gr) <- mcols
    gr
}

#' @rdname makeEpiTxDbFromtRNAdb
#' @importFrom tRNAdbImport import.tRNAdb
#' @export
gettRNAdbDataAsGRanges <- function(organism, sequences = NULL,
                                   dbURL = tRNAdbImport::TRNA_DB_URL){

    if(!is.character(organism) || length(organism) > 1L ||
       !(organism %in% listAvailableOrganismsFromtRNAdb())){
        stop("'organism' must be a single character value and match an entry ",
             "from listAvailableOrganismsFromtRNAdb()")
    }

    # get tRNAdb information
    message("Loading data from tRNAdb ...")
    gr <- lapply(c("allothers","plastid","mitochondrial"),
                 function(origin){
                     try(tRNAdbImport::import.tRNAdb(organism = organism,
                                                     database = "RNA",
                                                     origin = origin,
                                                     dbURL = dbURL),
                         silent = TRUE)
                 })
    gr <- gr[!vapply(gr,is,logical(1),"try-error")]
    if(length(gr) == 0L){
        stop("No results returned from tRNAdb. Check the output of ",
             "tRNAdbImport::import.tRNAdb with database set to 'RNA' from ",
             "all available origins.",
             call. = FALSE)
    }
    gr <- suppressWarnings(do.call(c,gr))
    message("Assembling data ...")
    mod <- separate(gr$tRNA_seq)
    mod <- .add_reference_information_tRNAdb(mod, gr)
    #
    if(!is.null(sequences)){
        message("Trying to associate tRNAdb entries with sequences ...")
        if(is.null(names(sequences)) ||
           anyDuplicated(names(sequences)) ||
           any(names(sequences) == "")){
            stop("names() of 'sequences' must be set, unique and not empty.",
                 call. = FALSE)
        }
        mod <- .map_modifications_to_sequences(mod, gr, sequences)
    }
    # assemble metadata columns
    mod <- .add_tRNAdb_metadata_columns(mod)
    mod
}

#' @rdname makeEpiTxDbFromtRNAdb
#' @export
makeEpiTxDbFromtRNAdb <- function(organism, sequences = NULL, metadata = NULL,
                                  dbURL = tRNAdbImport::TRNA_DB_URL){
    gr <- gettRNAdbDataAsGRanges(organism, sequences = sequences, dbURL = dbURL)
    if(!is.null(sequences)){
        gr <- gr[!duplicated(paste0(as.character(gr),"-",gr$mod_type))]
        colnames(mcols(gr)) <- gsub("mod_type","mod",colnames(mcols(gr)))
        gr <- Modstrings::removeIncompatibleModifications(gr, sequences)
        colnames(mcols(gr)) <- gsub("^mod$","mod_type",colnames(mcols(gr)))
    }
    makeEpiTxDbFromGRanges(gr, metadata = metadata, reassign.ids = FALSE)
}


#' @rdname makeEpiTxDbFromtRNAdb
#' @importFrom httr modify_url POST content
#' @importFrom xml2 xml_find_all
#' @export
listAvailableOrganismsFromtRNAdb <- function(){
    res <- httr::POST(paste0(httr::modify_url(tRNAdbImport::TRNA_DB_URL),
                             "DataOutput/Organisms"))
    html <- httr::content(res)
    organsisms <- as.character(xml2::xml_find_all(html,'.//div[@id="inhalt"]//span[count(.//span)=0]//a//b/text()'))
    rna_seq <- as.character(xml2::xml_find_all(html,'.//div[@id="inhalt"]//span[count(.//span)=0]//a//small/text()'))
    rna_n <- vapply(regmatches(rna_seq,regexec("genes,[ ]{1}([0-9]+)[ ]{1}RNA",rna_seq)),"[",character(1),2)
    organsisms[rna_n != 0]
}

library(AnnotationHub)
library(BSgenome.Scerevisiae.UCSC.sacCer3)

context("makeEpiTxDbFromtRNAdb")
test_that("makeEpiTxDbFromtRNAdb:",{
    httptest::skip_if_disconnected(url = "http://trna.bioinf.uni-leipzig.de/")
    bs <- BSgenome.Scerevisiae.UCSC.sacCer3
    # get tx
    ah <- AnnotationHub()
    edb <- query(ah, c("EnsDb","Saccharomyces cerevisiae", "99"))[[1]]
    seqlevelsStyle(edb) <- "UCSC"
    tx <- exonsBy(edb,"tx")
    tx_id <- IRanges::CharacterList(Map(rep,names(tx),lengths(tx)))
    mcols(tx, level="within")[,"tx_id"] <- tx_id
    genome(tx) <- "sacCer3"
    tx <- tx[lengths(tx) != 0L]
    #
    seq <- getSeq(bs,tx)
    seq <- relist(unlist(unlist(seq)),
                  IRanges::PartitioningByWidth(sum(nchar(seq))))
    seq_rna <- as(seq,"RNAStringSet")
    expect_message(etdb <- makeEpiTxDbFromtRNAdb("Saccharomyces cerevisiae",
                                                 seq_rna))
    expect_s4_class(etdb,"EpiTxDb")
    expect_s3_class(metadata(etdb),"data.frame")
    #
    actual <- modifications(etdb)
    expect_s4_class(actual,"GRanges")
    expect_length(actual,2990L)
    dbDisconnect(etdb$conn)
})

library(GenomicRanges)

context("Ranges helpers")
test_that("Ranges helpers:",{
    gr <- GRanges("chr1:1-5:+")
    actual <- positionSequence(gr)
    expect_type(actual,"integer")
    expect_equal(actual,seq.int(1L,5L))
    #
    grl <- GRangesList("1" = gr,"2" = gr,"3" = gr)
    actual2 <- positionSequence(grl)
    expect_named(actual2,c("1","2","3"))
    expect_s4_class(actual2,"IntegerList")
    expect_equal(actual2[[1L]],actual)
    expect_equal(actual2[[2L]],actual)
    expect_equal(actual2[[3L]],actual)
    #
    strand(grl[[2]]) <- "-"
    actual2 <- positionSequence(grl, TRUE, TRUE)
    expect_equal(actual2[[1L]],actual)
    expect_equal(actual2[[2L]],rev(actual))
    expect_equal(actual2[[3L]],actual)
    # rescale
    x <- IRanges("5-10")
    actual <- rescale(x, 100, 10)
    expect_s4_class(actual,"IRanges")
    expect_equal(start(actual), 41L)
    expect_equal(end(actual), 100L)
    #
    actual <- rescale(x, "31-60", "5-14")
    expect_s4_class(actual,"IRanges")
    expect_equal(start(actual), 31L)
    expect_equal(end(actual), 48L)
    #
    x <- IRangesList(x,x)
    actual <- rescale(x, 100, 10)
    expect_s4_class(actual,"IRangesList")
    expect_equal(start(actual[[2L]]), 41L)
    expect_equal(end(actual[[2L]]), 100L)
    #
    actual <- rescale(x, "31-60", "5-14")
    expect_s4_class(actual,"IRangesList")
    expect_equal(start(actual[[2L]]), 31L)
    expect_equal(end(actual[[2L]]), 48L)
})

context("shiftGenomicToTranscript")
test_that("shiftGenomicToTranscript:",{
    subject1 <- GRanges("chr1", IRanges(3, 6),
                        strand = "+")
    subject2 <- GRanges("chr1", IRanges(c(17,23), width=3),
                        strand = c("+","-"))
    subject3 <- GRanges("chr2", IRanges(c(51, 54), c(53, 59)),
                        strand = "-")
    subject <- GRangesList(a=subject1, b=subject2, c=subject3)
    tx1 <- GRanges("chr1", IRanges(1, 40),
                   strand="+")
    tx2 <- GRanges("chr1", IRanges(10, 30),
                   strand="+")
    tx3 <- GRanges("chr2", IRanges(50, 60),
                   strand="-")
    tx <- GRangesList(a=tx1, b=tx2, c=tx3)
    expect_warning(actual <- shiftGenomicToTranscript(unlist(subject),tx))
    expect_s4_class(actual,"GRanges")
    expect_equal(colnames(mcols(actual)),c("seq_start","seq_end","seq_strand",
                                           "seq_name"))
    expect_equal(start(actual),c(3L,17L,8L,8L,2L))
    expect_equal(end(actual),c(6L,19L,10L,10L,7L))
    actual2 <- shiftTranscriptToGenomic(actual,tx)
    expect_s4_class(actual2,"GRanges")
    expect_equal(ranges(actual2),
                 unname(ranges(unlist(subject)[c(1L,2L,2L,4L,5L)])))
    #
    expect_warning(actual <- shiftGenomicToTranscript(subject,tx))
    expect_s4_class(actual,"GRangesList")
    expect_equal(unlist(start(actual), use.names = FALSE),
                 c(3L, 17L, 8L, 8L, 2L))
    expect_equal(unlist(end(actual), use.names = FALSE),
                 c(6L, 19L, 10L, 10L, 7L))
    actual2 <- shiftTranscriptToGenomic(actual,tx)
    expect_s4_class(actual2,"GRangesList")
    expect_equal(unlist(start(actual2), use.names = FALSE),
                 c(3L, 26L, 17L, 51L, 54L))
    expect_equal(unlist(end(actual2), use.names = FALSE),
                 c(6L, 28L, 19L, 53L, 59L))
})

context("modifications")
test_that("modifications:",{
    etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
                             package="EpiTxDb")
    etdb <- loadDb(etdb_file)
    actual <- modifications(etdb)
    expect_s4_class(actual,"GRanges")
    expect_equal(colnames(mcols(actual)),c("mod_id","mod_type","mod_name"))
    expect_error(modifications(etdb, columns = c("MOD")),
                 "MOD: no such column in db schema")
    actual <- modifications(etdb, columns = c("MODID","MODTYPE","MODNAME"))
    expect_s4_class(actual,"GRanges")
    expect_equal(colnames(mcols(actual)),c("MODID","MODTYPE","MODNAME"))
    #
    expect_error(modifications(etdb, filter = 1),
                 "NA: no such column in db schema")
    actual <- modifications(etdb, filter = c(mod_id = 1))
    expect_s4_class(actual,"GRanges")
    expect_equal(length(actual),1L)
    # actual <- modifications(etdb, filter = c(mod_id = 236))
    # expect_equal(length(actual),0L)
    #
    actual <- modificationsBy(etdb)
    expect_s4_class(actual,"GRangesList")
    expect_named(actual)
    expect_equal(colnames(mcols(unlist(actual))),
                 c("mod_id","mod","mod_name"))
    actual <- modificationsBy(etdb, by = "specifier")
    expect_s4_class(actual,"GRangesList")
    expect_named(actual)
    actual <- modificationsBy(etdb, by = "reaction")
    expect_s4_class(actual,"GRangesList")
    expect_named(actual)
    actual <- modificationsBy(etdb, by = "modtype")
    expect_s4_class(actual,"GRangesList")
    expect_named(actual)
    expect_named(actual,c("Am","Cm","Gm","Um","Y"))
    actual <- modificationsBy(etdb, by = "specifiertype")
    expect_s4_class(actual,"GRangesList")
    expect_named(actual)
    dbDisconnect(etdb$conn)
})

---
title: "EpiTxDb: creating an EpiTxDb object"
author: "Felix G.M. Ernst"
date: "`r Sys.Date()`"
package: EpiTxDb
output:
  BiocStyle::html_document:
    toc: true
    toc_float: true
    df_print: paged
vignette: >
  %\VignetteIndexEntry{EpiTxDb-creation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
---

```{r style, echo = FALSE, results = 'asis'}
BiocStyle::markdown(css.files = c('custom.css'))
```

# Introduction

To create an `EpiTxDb` object a number of different functions are available.
The most univeral functions are `makeEpiTxDb` and `makeEpiTxDbFromGRanges`.
`makeEpiTxDb` uses four `data.frame`s as input, whereas `makeEpiTxDbFromGRanges`
is a wrapper for information available as a `GRanges` object.

The other functions are `makeEpiTxDbFromRMBase` and `makeEpiTxDbFromtRNAdb`,
which are aimed to make data available from the RMBase v2.0 database
[[@Xuan.2017;@Sun.2015]](#References) or the tRNAdb
[[@Juehling.2009;@Sprinzl.2005]](#References). However, before creating your 
`EpiTxDb` objects, have a look at the already available resources for 
`H. sapiens`, `M. musculus` and `S. cerevisiae.`

Additional metadata can be provided as separate `data.frame` for all functions.
The `data.frame` must have two columns `name` and `value`.

```{r, echo = FALSE}
suppressPackageStartupMessages({
    library(GenomicRanges)
    library(EpiTxDb)
})
```
```{r, eval = FALSE}
library(GenomicRanges)
library(EpiTxDb)
```

# `makeEpiTxDb` and `makeEpiTxDbFromGRanges`

The creation of an etdb object is quite easy starting with a `GRanges` object.

```{r}
gr <- GRanges(seqnames = "test",
              ranges = IRanges::IRanges(1,1),
              strand = "+",
              DataFrame(mod_id = 1L,
                        mod_type = "Am",
                        mod_name = "Am_1"))
etdb <- makeEpiTxDbFromGRanges(gr, metadata = data.frame(name = "test",
                                                         value = "Yes"))
etdb
metadata(etdb)
```

Additional data can be provided via the metadata columns of the `GRanges`
object. For supported columns have a look at `?makeEpiTxDb` or 
`?makeEpiTxDbFromGRanges`.

# `makeEpiTxDbFromtRNAdb`

The information of the tRNAdb can be accessed via the `tRNAdbImport` package 
using the RNA database. As a result a `ModRNAStringSet` object is returned from
which the modifications can be extracted using `separate()`.

The only input require is a valid organism name returned by 
`listAvailableOrganismsFromtRNAdb()`.

```{r, eval=FALSE}
# Currently not run since the server is not available
etdb <- makeEpiTxDbFromtRNAdb("Saccharomyces cerevisiae")
etdb
```

For additional information have a look at `?makeEpiTxDbFromtRNAdb`. The result
returned from the tRNAdb is also available as `GRanges` object, if 
`gettRNAdbDataAsGRanges()` is used.

# `makeEpiTxDbFromRMBase`

Analogous to the example above `makeEpiTxDbFromRMBase()` will download the data
from the RMBase v2.0. Three inputs are required, `organism`, `genome` and 
`modtype`, which have to valid bia the functions 
`listAvailableOrganismsFromRMBase()`, `.listAvailableGenomesFromRMBase()` and 
`listAvailableModFromRMBase`.

```{r, eval=FALSE}
etdb <- makeEpiTxDbFromRMBase(organism = "yeast",
                              genome = "sacCer3",
                              modtype = "m1A")
```

Internally, the files are cached using the `BiocFileCache` package and passed
to `makeEpiTxDbFromRMBaseFiles()`, which can also be used with locally stored
files. The resuls for creating the `EpiTxDb` class are processed from these
files via the `getRMBaseDataAsGRanges()` function.

# Session info

```{r}
sessionInfo()
```

<a name="References"></a>

# References

---
title: "EpiTxDb: Storing and accessing epitranscriptomic information using the AnnotationDbi interface"
author: "Felix G.M. Ernst"
date: "`r Sys.Date()`"
package: EpiTxDb
output:
  BiocStyle::html_document:
    toc: true
    toc_float: true
    df_print: paged
vignette: >
  %\VignetteIndexEntry{EpiTxDb}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
---

```{r style, echo = FALSE, results = 'asis'}
BiocStyle::markdown(css.files = c('custom.css'))
```

# Installation

```{r, eval=FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install(c("EpiTxDb","EpiTxDb.Hs.hg38"))
```

# Introduction

The epitranscriptome includes all post-transcriptional modifications of the RNA
and describes and additional layer of information encoded on RNA. Like the term
epigenome it is not about a change in nucleotide sequences, but the addition of
functional elements through modifications.

With the development of high throughput detection strategies for specific RNA
modifications, such as miCLIP and Pseudo-Seq amongst other, a large number of
modified positions have been identified and were summarized via the RMBase 
project [[@Xuan.2017;@Sun.2015]](#References) project.

To make these information avaialble within the Bioconductor universe `EpiTxDb`
was developed, which facilitates the storage of epitranscriptomic information.
More specifically, it can keep track of modification identity, position, the
enzyme for introducing it on the RNA, a specifier which determines the position
on the RNA to be modified and the literature references each modification is
associated with.

# Getting started

```{r, results="hide", include=TRUE, message=FALSE, warning=FALSE}
library(EpiTxDb)
library(EpiTxDb.Hs.hg38)
```

The class `EpiTxDb` is the class for storing the epitranscriptomic data. It 
inherits the inner workings of `AnnotationDb` class from the `AnnotationDbi`
package.

As an example for the vignette the snoRNAdb data [[@Lestrade.2006]](#References)
from the `EpiTxDb.Hs.hg38` package will be used. The data is stored in the
`AnnotationHub` and is downloaded and cached upon the first request.

```{r}
etdb <- EpiTxDb.Hs.hg38.snoRNAdb()
etdb
```

As expected for an `AnnotationDb` class the general accessors are available.

```{r}
keytypes(etdb)
columns(etdb)
head(keys(etdb, "MODID"))
select(etdb, keys = "1",
       columns = c("MODNAME","MODTYPE","MODSTART","MODSTRAND","SNNAME",
                   "RXGENENAME","SPECTYPE","SPECGENENAME"),
       keytype = "MODID")
```

The columns with the prefix `RX` or `SPEC` reference the reaction enzyme and the
location specifier. This can be the same information, but for ribosomal 
modifications from the snoRNAdb it is of course fibrillarin and a snoRNA.

In addition the following accessor for metadata are available as well.

```{r}
species(etdb)
organism(etdb)
seqlevels(etdb)
```

# Accessing RNA modifications

The specialized accessors are `modifications()` and `modificationsBy()`. 
`modifications()` allows for filtering results, whereas `modificationsBy()`
returns all the modifications in batches separated by certain information.

```{r}
modifications(etdb, columns = c("mod_id","mod_type","mod_name",
                                "rx_genename","spec_genename",
                                "ref_type","ref"),
              filter = list(mod_id = 1:3))
```

```{r}
# split by sequence name, usually a transcipt identifier
modificationsBy(etdb, by = "seqnames")
# split modification type
modificationsBy(etdb, by = "modtype")
```

# Shifting coordinates from genomic to transcriptomic

Since epitranscriptomic modifications by their nature can have different meaning
for each of the individual transcript variants. This also introduces conflicts
for saving epitranscriptomics coordinates. In the example above the coordinates
are given per transcript, because of the source data.

However, not all sources report transcript coordinates. It might be of interest
to shift the coordinates to transcript coordinates and at the same time taking
care that with transcript variants multiple options exist for each of the
transcript maturation process: From one genomic coordinate, multiple
transcriptomic coordinates can be spawned.

Whether this is biologically relevant or whether biological evidence does exist
for each modification on each transcript cannot be guaranteed or differentiated
technically depending on the methods used. This might change with the arrival of
new techniques allowing for detection of modified nucleotides per individual
transcript variant.

```{r, echo = FALSE}
suppressPackageStartupMessages({
  library(TxDb.Hsapiens.UCSC.hg38.knownGene)
  library(BSgenome.Hsapiens.UCSC.hg38)
})
```
```{r, eval = FALSE}
library(TxDb.Hsapiens.UCSC.hg38.knownGene)
library(BSgenome.Hsapiens.UCSC.hg38)
```
```{r}
txdb <- TxDb.Hsapiens.UCSC.hg38.knownGene
seqlevels(txdb) <- "chr1"
bs <- BSgenome.Hsapiens.UCSC.hg38

etdb <- EpiTxDb.Hs.hg38.RMBase()

tx <- exonsBy(txdb)
mod <- modifications(etdb, filter = list(sn_name = "chr1"))
length(mod)
```

In the following example we will focus on shifting the coordinates to individual
mature transcripts. However, keep in mind, that premature transcript might be
of interest as well and this can be controlled via the `tx` arguments of
`shiftGenomicToTranscript()`

```{r}
mod_tx <- shiftGenomicToTranscript(mod, tx)
length(mod_tx)
```

Due to multiple matches for each transcript variant the number of modifications
has increased.

With the we can plot the relative positions of modifications by type on 
`chr1` transcripts.

```{r}
mod_tx <- split(mod_tx,seqnames(mod_tx))
names <- Reduce(intersect,list(names(mod_tx),names(tx)))

# Getting the corresponding 5'-UTR and 3'-UTR annotations
fp <- fiveUTRsByTranscript(txdb)
tp <- threeUTRsByTranscript(txdb)
tx <- tx[names]
mod_tx <- mod_tx[names]
fp_m <- match(names,names(fp))
fp_m <- fp_m[!is.na(fp_m)]
tp_m <- match(names,names(tp))
tp_m <- tp_m[!is.na(tp_m)]
fp <- fp[fp_m]
tp <- tp[tp_m]

# Getting lengths of transcripts, 5'-UTR and 3'-UTR
tx_lengths <- sum(width(tx))
fp_lengths <- rep(0L,length(tx))
names(fp_lengths) <- names
fp_lengths[names(fp)] <- sum(width(fp))
tp_lengths <- rep(0L,length(tx))
names(tp_lengths) <- names
tp_lengths[names(tp)] <- sum(width(tp))

# Rescale modifications
# CDS start is at position 1L and cds end at position 1000L
from <- IRanges(fp_lengths+1L, tx_lengths - tp_lengths)
to <- IRanges(1L,1000L)
mod_rescale <- rescale(mod_tx, to, from)

# Construct result data.frame
rel_pos <- data.frame(mod_type = unlist(mcols(mod_rescale,level="within")[,"mod_type"]),
                      rel_pos = unlist(start(mod_rescale)))
rel_pos <- rel_pos[rel_pos$rel_pos < 1500 & rel_pos$rel_pos > -500,]
```

```{r}
library(ggplot2)
ggplot(rel_pos[rel_pos$mod_type %in% c("m6A","m1A","Y"),],
       aes(x = rel_pos, colour = mod_type)) + 
  geom_density()
```

# Session info

```{r}
sessionInfo()
```

<a name="References"></a>

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EpiTxDb-class.R
\docType{class}
\name{EpiTxDb-class}
\alias{EpiTxDb-class}
\alias{.EpiTxDb}
\alias{EpiTxDb}
\alias{organism,EpiTxDb-method}
\alias{seqinfo,EpiTxDb-method}
\alias{seqlevels,EpiTxDb-method}
\alias{as.list,EpiTxDb-method}
\title{EpiTxDb objects}
\usage{
\S4method{organism}{EpiTxDb}(object)

\S4method{seqinfo}{EpiTxDb}(x)

\S4method{seqlevels}{EpiTxDb}(x)

\S4method{as.list}{EpiTxDb}(x)
}
\arguments{
\item{x, object}{a \code{EpiTxDb} object}
}
\value{
For
  \itemize{
    \item \code{organism()} and \code{seqlevels()}: a \code{character}
      vector
    \item \code{seqinfo()}: a
      \code{\link[Seqinfo:Seqinfo-class]{Seqinfo}} object
    \item \code{as.list()} a \code{list}
  }
}
\description{
The \code{EpiTxDb} class is a
\code{\link[AnnotationDbi:AnnotationDb-class]{AnnotationDb}} type container
for storing Epitranscriptomic information.

The information are typically stored on a per transcript and not as genomic
coordinates, but the \code{EpiTxDb} class is agnostic to this. In case of
genomic coordinates \code{transcriptsBy} will return modifications per
chromosome.
}

\examples{
etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
                        package="EpiTxDb")
etdb <- loadDb(etdb_file)
etdb

# general methods
seqinfo(etdb) #
seqlevels(etdb) # easy access to all transcript names
}
\seealso{
\itemize{
    \item \code{\link[=makeEpiTxDbFromGRanges]{makeEpiTxDbFromGRanges}} for
      creating a \code{EpiTxDb} object from a
      \code{\link[GenomicRanges:GRanges-class]{GRanges}} object and it's
       metadata columns
    \item \code{\link[=makeEpiTxDbFromRMBase]{makeEpiTxDbFromRMBase}} for
      creating a \code{EpiTxDb} object from RMBase online resources
    \item \code{\link[=makeEpiTxDbFromtRNAdb]{makeEpiTxDbFromtRNAdb}} for
      creating a \code{EpiTxDb} object from tRNAdb online resources
    \item \code{\link[=makeEpiTxDb]{makeEpiTxDb}} for creating a
      \code{EpiTxDb} object from \code{data.frames}
    \item \code{\link[=modifications]{modifications}},
      \code{\link[=modifications]{modificationsBy}} for getting
      epitranscriptomic modification locations
    \item \code{\link[=select]{select}} for using the default interface of
      \code{\link[AnnotationDbi:AnnotationDb-class]{AnnotationDb}} objects.
    \item \code{\link[=shiftGenomicToTranscript]{shiftGenomicToTranscript}}
      and \code{\link[=shiftGenomicToTranscript]{shiftTranscriptToGenomic}}
      for transfering genomic to transcript coordinates and back again.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/ranges-helpers.R
\name{rescale}
\alias{rescale}
\alias{rescale,IRanges-method}
\alias{rescale,IRangesList-method}
\alias{rescale,GRanges-method}
\alias{rescale,GRangesList-method}
\title{Rescaling \code{Ranges} object}
\usage{
rescale(x, to = 1L, from = 1L)

\S4method{rescale}{IRanges}(x, to = 1L, from = 1L)

\S4method{rescale}{IRangesList}(x, to = 1L, from = 1L)

\S4method{rescale}{GRanges}(x, to = 1L, from = 1L)

\S4method{rescale}{GRangesList}(x, to = 1L, from = 1L)
}
\arguments{
\item{x}{a \code{IRanges}, \code{GRanges}, \code{IRangesList} and
\code{GRangesList} object}

\item{to, from}{an \code{IRanges} object, a \code{character} vector coercible
to \code{IRanges} or a \code{integer} vector parallel to \code{x} or with
\code{length = 1L}.}
}
\value{
an object of the same type and dimensions as \code{x}
}
\description{
\code{rescale()} rescales \code{IRanges}, \code{GRanges}, \code{IRangesList}
and \code{GRangesList} by using minima and maxima derived from \code{to} and
\code{from}.
}
\examples{
x <- IRanges("5-10")
# widen the ranges
rescale(x, 100, 10)
# widen and shift
rescale(x, "31-60", "5-14")
}
\seealso{
\code{\link[IRanges:IRanges-constructor]{IRanges}} for details on
\code{character} vectors coercible to \code{IRanges}.
}
\author{
H. Pagès, F. Ernst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/ranges-helpers.R
\name{positionSequence}
\alias{positionSequence}
\alias{positionSequence,Ranges-method}
\alias{positionSequence,RangesList-method}
\alias{as.integer,Ranges-method}
\title{Generate integer sequences from position information of \code{Ranges}}
\usage{
positionSequence(x, order = FALSE, decreasing = FALSE)

\S4method{positionSequence}{Ranges}(x, order = FALSE, decreasing = FALSE)

\S4method{positionSequence}{RangesList}(x, order = FALSE, decreasing = FALSE)

\S4method{as.integer}{Ranges}(x)
}
\arguments{
\item{x}{a \code{Ranges} object, like a
\code{\link[GenomicRanges:GRanges-class]{GRanges}} or
\code{\link[IRanges:IRanges-class]{IRanges}}, or a \code{RangesList}
object, like a \code{\link[GenomicRanges:GRangesList-class]{GRangesList}}
or \code{\link[IRanges:IRangesList-class]{IRangesList}}}

\item{order}{\code{TRUE} or \code{FALSE}: Should the position be ordered?
(default: \code{order = FALSE})}

\item{decreasing}{\code{TRUE} or \code{FALSE}: If \code{order = TRUE} Should
the position be ordered in a decreasing order? (default: \code{order =
FALSE})}
}
\value{
a \code{integer} vector if x is a
  \code{\link[GenomicRanges:GRanges-class]{GRanges}} object and a
  \code{IntegerList} if x is a
  \code{\link[GenomicRanges:GRangesList-class]{GRangesList}}
}
\description{
\code{positionSequence} generates sequences of integer values
along the range information of \code{x}. This can be used for navigating
specific positions on a range information.
}
\examples{
library(GenomicRanges)
# Returns an integer vector
gr <- GRanges("chr1:1-5:+")
positionSequence(gr)
gr2 <- GRanges("chr1:1-5:-")
positionSequence(gr)
# returns an IntegerList
grl <- GRangesList("1" = gr,"2" = gr,"3" = gr2) # must be named
positionSequence(grl)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EpiTxDb.R
\docType{package}
\name{EpiTxDb-package}
\alias{EpiTxDb-package}
\title{EpiTxDb: Storing and accessing epitranscriptomic information using the AnnotationDbi interface}
\description{
EpiTxDb facilitates the storage of epitranscriptomic information. More specifically, it can keep track of modification identity, position, the enzyme for introducing it on the RNA, a specifier which determines the position on the RNA to be modified and the literature references each modification is associated with.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/FelixErnst/EpiTxDb}
  \item Report bugs at \url{https://github.com/FelixErnst/EpiTxDb/issues}
}

}
\author{
\strong{Maintainer}: Felix G.M. Ernst \email{felix.gm.ernst@outlook.com} (\href{https://orcid.org/0000-0001-5064-0928}{ORCID})

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EpiTxDb.R
\name{EpiTxDb-package#'}
\alias{EpiTxDb-package#'}
\title{\code{EpiTxDb} - Storing and accessing epitranscriptomic information
using the AnnotationDbi interface}
\description{
title
}
\references{
Jia-Jia Xuan, Wen-Ju Sun, Ke-Ren Zhou, Shun Liu, Peng-Hui Lin, Ling-Ling
Zheng, Liang-Hu Qu, Jian-Hua Yang (2017): "RMBase v2.0: Deciphering the Map of
RNA Modifications from Epitranscriptome Sequencing Data." Nucleic Acids
Research, Volume 46, Issue D1, 4 January 2018, Pages D327–D334.
doi: 10.1093/nar/gkx934

Jühling, Frank; Mörl, Mario; Hartmann, Roland K.; Sprinzl, Mathias; Stadler,
Peter F.; Pütz, Joern (2009): "TRNAdb 2009: Compilation of tRNA Sequences and
tRNA Genes." Nucleic Acids Research 37 (suppl_1): D159–D162.
doi: 10.1093/nar/gkn772

Sprinzl, Mathias; Vassilenko, Konstantin S. (2005): "Compilation of tRNA
Sequences and Sequences of tRNA Genes." Nucleic Acids Research 33 (suppl_1):
D139–D140. doi: 10.1093/nar/gki012
}
\author{
Felix G M Ernst [aut]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select-methods.R
\name{select}
\alias{select}
\alias{columns}
\alias{keys}
\alias{keytypes}
\alias{select,EpiTxDb-method}
\alias{columns,EpiTxDb-method}
\alias{keys,EpiTxDb-method}
\alias{keytypes,EpiTxDb-method}
\title{Using the "select" interface on \code{EpiTxDb} objects}
\usage{
\S4method{select}{EpiTxDb}(x, keys, columns, keytype, ...)

\S4method{columns}{EpiTxDb}(x)

\S4method{keys}{EpiTxDb}(x, keytype, ...)

\S4method{keytypes}{EpiTxDb}(x)
}
\arguments{
\item{x}{a \code{\link[=EpiTxDb-class]{EpiTxDb}} object}

\item{keys, columns, keytype, ...}{See
\code{\link[AnnotationDbi:AnnotationDb-class]{AnnotationDb}} for more
comprehensive description of the methods \code{select}, \code{keys},
\code{columns} and \code{keytypes} and their arguments.}
}
\value{
a \code{data.frame} object for \code{select()} and a \code{character}
  vecor for \code{keytypes()}, \code{keys()} and \code{columns()}.
}
\description{
As expected for a \code{AnnotationDb} object, the general accessors
  \code{select}, \code{keys}, \code{columns} and \code{keytypes} can be used
  to get information from a \code{\link[=EpiTxDb-class]{EpiTxDb}} object.
}
\examples{
etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
                         package="EpiTxDb")
etdb <- loadDb(etdb_file)
etdb
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeEpiTxDb.R
\name{makeEpiTxDb}
\alias{makeEpiTxDb}
\title{Creating a \code{EpiTxDb} from user supplied annotations as
\code{data.frame}s}
\usage{
makeEpiTxDb(
  modifications,
  reactions = NULL,
  specifiers = NULL,
  references = NULL,
  metadata = NULL,
  reassign.ids = FALSE
)
}
\arguments{
\item{modifications}{A \code{data.frame} containg the following columns:
\itemize{
  \item \code{mod_id}: a unique \code{integer} value per modification.
  \item \code{mod_type}: the modification type as a \code{character} or
  \code{factor} value. Must be a value from
  \code{shortName(ModRNAString())}.
  \item \code{mod_name}: a \code{character} or \code{factor} name for the
  specific modification
  \item \code{mod_start}: the start position for the modification as
  \code{integer} value. Usually \code{mod_start = mod_end}
  \item \code{mod_end}: the end position for the modification as
  \code{integer} value. Usually \code{mod_start = mod_end}
  \item \code{mod_strand}: the strand information for the modificaion as a
  \code{character} or \code{factor}.
  \item \code{sn_id}: a \code{integer} value per unique sequence
  \item \code{sn_name}: a \code{character} or \code{factor} as sequence
  name, e.g a chromosome or a transcript identifier like \code{chr1}.
}
The first six are mandatory, whereas one of the last two has to be set.
\code{sn_id} will be generated from \code{sn_name}, if \code{sn_id} is not
set.}

\item{reactions}{An optional \code{data.frame} containg the following
columns:
\itemize{
  \item \code{mod_id}: a \code{integer} value per modification and the
  link to the \code{modification} \code{data.frame}.
  \item \code{rx_genename}: a \code{character} or \code{factor} referencing
  a genename for the enzyme incorporating the modification.
  \item \code{rx_rank}: a \code{integer} for sorting enzyme reactions, if
  multiple enzymes are involved in the modification's
  incorporation/maintenance.
  \item \code{rx_ensembl}: a \code{character} or \code{factor} with an
  ensembl identifier for the genename of the enzyme.
  \item \code{rx_ensembltrans}: a \code{character} or \code{factor} with an
  ensembl identifier for the transcript being translated into the enzyme.
  \item \code{rx_entrezid}: a \code{character} or \code{factor} with an
  entrezid for the genename of the enzyme.
}
(default: \code{reactions = NULL})}

\item{specifiers}{An optional \code{data.frame} containg the following
columns:
  \itemize{
  \item \code{mod_id}: a \code{integer} value per modification and the
  link to the \code{modification} \code{data.frame}.
  \item \code{spec_type}: a \code{character} or \code{factor}
  referencing a type of specifier, e.g. \code{snoRNA}. Not checked for
  validity.
  \item \code{spec_genename}: a \code{character} or \code{factor}
  referencing a genename for the specifier directing an enzyme to the
  specific location for the modification to be incorporated.
  \item \code{spec_ensembl}: a \code{character} or \code{factor} with an
  ensembl identifier for the genename of the specifier.
  \item \code{spec_ensembltrans}: a \code{character} or \code{factor} with an
  ensembl identifier for the transcript being translated into the specifier.
  \item \code{spec_entrezid}: a \code{character} or \code{factor} with an
  entrezid for the genename of the specifier.
}
(default: \code{specifiers = NULL})}

\item{references}{An optional \code{data.frame} containg the following
columns:
\itemize{
  \item \code{mod_id}: a \code{integer} value per modification and the
  link to the \code{modification} \code{data.frame}.
  \item \code{ref_type}: a \code{character} or \code{factor} with a
  reference type, e.g. \code{PMID}. Is not checked for validity.
  \item \code{ref}: a \code{character} or \code{factor} with a reference
  value, e.g. a specific pubmed id or an journal article. Is not checked
  for validity.
}
(default: \code{references = NULL})}

\item{metadata}{An optional \code{data.frame} containg the following columns:
\itemize{
  \item \code{name}: a \code{character} value used as name
  \item \code{value}: a \code{character} value
}
This dataframe will be returned
by \code{metadata()} (default: \code{metadata = NULL})}

\item{reassign.ids}{\code{TRUE} or \code{FALSE} Controls how internal
\code{mod_id}s should be assigned. If \code{reassign.ids} is \code{FALSE}
(the default) and if the ids are supplied, then they are used as the
internal ids, otherwise the internal ids are assigned in a way that is
compatible with the order defined by ordering the modifications first by
chromosome, then by strand, then by start, and finally by end.}
}
\value{
a \code{EpiTxDb} object.
}
\description{
\code{makeEpiTxDb} is a low-level constructor for creating a
\code{\link[=EpiTxDb-class]{EpiTxDb}} object from user supplied annotations.

This functions typically will not be used by regular users.
}
\examples{
mod <- data.frame("mod_id" = 1L,
                  "mod_type" = "m1A",
                  "mod_name" = "m1A_1",
                  "mod_start" = 1L,
                  "mod_end" = 1L,
                  "mod_strand" = "+",
                  "sn_id" = 1L,
                  "sn_name" = "test")
rx <- data.frame(mod_id = 1L,
                 rx_genename = "test",
                 rx_rank = 1L,
                 rx_ensembl = "test",
                 rx_ensembltrans = "test",
                 rx_entrezid = "test")
spec <- data.frame(mod_id = 1L,
                   spec_type = "test",
                   spec_genename = "test",
                   spec_ensembl = "test",
                   spec_ensembltrans = "test",
                   spec_entrezid = "test")
ref <- data.frame(mod_id = 1L,
                  ref_type = "test",
                  ref = "test")
etdb <- makeEpiTxDb(mod,rx,spec,ref)
}
\seealso{
\itemize{
  \item \code{\link[=makeEpiTxDbFromGRanges]{makeEpiTxDbFromGRanges}} for
    creating a \code{EpiTxDb} object from a
  \code{\link[GenomicRanges:GRanges-class]{GRanges}} object and it's metadata
    columns
  \item \code{\link[=makeEpiTxDbFromRMBase]{makeEpiTxDbFromRMBase}} for
    creating a \code{EpiTxDb} object from RMBase online resources
  \item \code{\link[=makeEpiTxDbFromtRNAdb]{makeEpiTxDbFromtRNAdb}} for
    creating a \code{EpiTxDb} object from tRNAdb online resources
  \item \code{\link[Modstrings:alphabet]{shortName}} and
  \code{\link[Modstrings:ModRNAString]{ModRNAString}} for information on
  \code{ModRNAString} objects.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/modifications.R,
%   R/modificationsBy.R
\name{modifications}
\alias{modifications}
\alias{modificationsBy}
\alias{modifiedSeqsByTranscript}
\alias{modifications,EpiTxDb-method}
\alias{modificationsBy,EpiTxDb-method}
\alias{modifiedSeqsByTranscript,EpiTxDb,DNAStringSet-method}
\title{Getting modification data from a \code{EpiTxDb-object}}
\usage{
modifications(
  x,
  columns = c("mod_id", "mod_type", "mod_name"),
  filter = NULL,
  use.names = FALSE,
  ...
)

modificationsBy(
  x,
  by = c("seqnames", "mod_type", "reaction", "specifier", "specifier_type"),
  ...
)

modifiedSeqsByTranscript(x, sequences, ...)

\S4method{modifications}{EpiTxDb}(
  x,
  columns = c("mod_id", "mod_type", "mod_name"),
  filter = NULL,
  use.names = FALSE
)

\S4method{modificationsBy}{EpiTxDb}(
  x,
  by = c("seqnames", "modtype", "reaction", "specifier", "specifiertype")
)

\S4method{modifiedSeqsByTranscript}{EpiTxDb,DNAStringSet}(x, sequences)
}
\arguments{
\item{x}{a \code{\link[=EpiTxDb-class]{EpiTxDb}}}

\item{columns}{Columns to include in the result. If the vector is named,
those names are used for the corresponding column in the element metadata
of the returned object. (default: \code{columns =
c("mod_id","mod_type","mod_name")})}

\item{filter}{Either NULL or a named list of vectors to be used to restrict
the output. Valid names for this list are: "mod_id", "mod_type",
"mod_name", "sn_id", "sn_name", "rx_genename", "rx_ensembl",
"rx_ensembltrans", "rx_entrezid", "spec_genename", "spec_type",
"spec_ensembl", "spec_ensembltrans", "spec_entrezid" , "ref_type" and
"ref". (default: \code{filter = NULL})}

\item{use.names}{\code{TRUE} or \code{FALSE}. If \code{TRUE}, the
modification names are set as the names of the returned object. (default:
\code{use.names = FALSE})}

\item{...}{Not used.}

\item{by}{By which information type should the result be split into? A
\code{character} value from one of the following values:
\itemize{
  \item seqnames
  \item mod_type
  \item reaction
  \item specifier
  \item specifier_type
}}

\item{sequences}{A \code{RNAStringSet}, which can be used as input for
\code{\link[Modstrings:separate]{combineIntoModstrings()}}. See
\code{\link[Modstrings:separate]{?combineIntoModstrings}} for additional
details.}
}
\value{
a \code{\link[GenomicRanges:GRanges-class]{GRanges}} object for
  \code{modifications} and a
  \code{\link[GenomicRanges:GRangesList-class]{GRangesList}} for
  \code{modificationsBy}.
}
\description{
\code{modifications} and \code{modificationsBy} are functions to
extract modification annotation from a \code{\link[=EpiTxDb-class]{EpiTxDb}}
object.

\code{modifiedSeqsByTranscript} returns a
\code{\link[Modstrings:ModStringSet]{ModRNAStringSet}} from a \code{EpiTxDb}
object and compatible \code{RNAStringSet} object. This used the
\code{\link[Modstrings:separate]{combineIntoModstrings()}} function from the
\code{Modstrings} package.
}
\examples{
etdb_file <- system.file("extdata", "EpiTxDb.Hs.hg38.snoRNAdb.sqlite",
                        package="EpiTxDb")
etdb <- loadDb(etdb_file)
etdb
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/shiftGenomicToTranscript.R
\name{shiftTranscriptToGenomic}
\alias{shiftTranscriptToGenomic}
\alias{shiftGenomicToTranscript}
\alias{shiftTranscriptToGenomic,GRanges,GRangesList-method}
\alias{shiftTranscriptToGenomic,GRangesList,GRangesList-method}
\alias{shiftGenomicToTranscript,GRanges,GRangesList-method}
\alias{shiftGenomicToTranscript,GRangesList,GRangesList-method}
\title{Shift \code{GRanges} coordinates based on another \code{GRanges}
object}
\usage{
shiftTranscriptToGenomic(subject, tx)

shiftGenomicToTranscript(subject, tx)

\S4method{shiftTranscriptToGenomic}{GRanges,GRangesList}(subject, tx)

\S4method{shiftTranscriptToGenomic}{GRangesList,GRangesList}(subject, tx)

\S4method{shiftGenomicToTranscript}{GRanges,GRangesList}(subject, tx)

\S4method{shiftGenomicToTranscript}{GRangesList,GRangesList}(subject, tx)
}
\arguments{
\item{subject}{a \code{\link[GenomicRanges:GRanges-class]{GRanges}} or
\code{\link[GenomicRanges:GRangesList-class]{GRangesList}} object}

\item{tx}{a named \code{\link[GenomicRanges:GRangesList-class]{GRangesList}}
object.}
}
\value{
a \code{\link[GenomicRanges:GRanges-class]{GRanges}} or
  \code{\link[GenomicRanges:GRangesList-class]{GRangesList}} object depending
  on the type of \code{subject}
}
\description{
\code{shiftGenomicToTranscript} shifts positions of a
\code{\link[GenomicRanges:GRanges-class]{GRanges}}
object based on coordinates of another \code{GRanges} object. The most common
application is to shift genomic coordinates to transcript coordinates, which
is reflected in the name. \code{shiftTranscriptToGenomic} implements the
reverse operation.

Matches are determined by
\code{\link[GenomicRanges:findOverlaps-methods]{findOverlaps}} for
\code{shiftGenomicToTranscript} and by
\code{\link[S4Vectors:Vector-comparison]{findMatches}} for
\code{shiftTranscriptToGenomic} using the \code{seqnames} of the
\code{subject} and the \code{names} of \code{tx}.
}
\examples{
library(GenomicRanges)
# Construct some example data
subject1 <- GRanges("chr1", IRanges(3, 6),
                    strand = "+")
subject2 <- GRanges("chr1", IRanges(c(17,23), width=3),
                    strand = c("+","-"))
subject3 <- GRanges("chr2", IRanges(c(51, 54), c(53, 59)),
                    strand = "-")
subject <- GRangesList(a=subject1, b=subject2, c=subject3)
tx1 <- GRanges("chr1", IRanges(1, 40),
               strand="+")
tx2 <- GRanges("chr1", IRanges(10, 30),
               strand="+")
tx3 <- GRanges("chr2", IRanges(50, 60),
               strand="-")
tx <- GRangesList(a=tx1, b=tx2, c=tx3)

# shift to transcript coordinates. Since the third subject does not have
# a match in tx it is dropped with a warning
shifted_grl <- shiftGenomicToTranscript(subject,tx)

# ... and back
shifted_grl2 <- shiftTranscriptToGenomic(shifted_grl,tx)

# comparison of ranges work. However the seqlevels differ
ranges(shifted_grl2) == ranges(subject[list(1,c(1,1),c(1,2))])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeEpiTxDbFromRMBase.R
\docType{data}
\name{makeEpiTxDbFromRMBase}
\alias{makeEpiTxDbFromRMBase}
\alias{EPITXDB_RMBASE_URL}
\alias{downloadRMBaseFiles}
\alias{getRMBaseDataAsGRanges}
\alias{makeEpiTxDbFromRMBaseFiles}
\alias{listAvailableOrganismsFromRMBase}
\alias{listAvailableGenomesFromRMBase}
\alias{listAvailableModFromRMBase}
\title{Create a \code{EpiTxDb} object from RMBase v2.0 online resources}
\format{
An object of class \code{character} of length 1.
}
\usage{
EPITXDB_RMBASE_URL

downloadRMBaseFiles(organism, genome, modtype)

makeEpiTxDbFromRMBase(
  organism,
  genome,
  modtype,
  tx = NULL,
  sequences = NULL,
  metadata = NULL,
  reassign.ids = FALSE,
  verbose = FALSE
)

getRMBaseDataAsGRanges(files, verbose = FALSE)

makeEpiTxDbFromRMBaseFiles(
  files,
  tx = NULL,
  sequences = NULL,
  metadata = NULL,
  reassign.ids = FALSE,
  verbose = FALSE
)

listAvailableOrganismsFromRMBase()

listAvailableGenomesFromRMBase(organism)

listAvailableModFromRMBase(organism, genome)
}
\arguments{
\item{organism}{A \code{character} value, which must match an organism
descriptor on the RMBase download website.}

\item{genome}{A \code{character} value, which must match a genome
descriptor on the RMBase download website.}

\item{modtype}{A \code{character} value, which must match one or more
modification descriptors on the RMBase download website.}

\item{tx}{A \code{\link[GenomicRanges:GRangesList-class]{GRangesList}} object
which will be used to shift the genomic coordinates to transcript
coordinates. This is optional, but highly recommended. (default:
\code{tx = NULL}).}

\item{sequences}{A named \code{DNAStringSet} or \code{RNAStringSet}, which
will be used to check whether the defined modifications are compatible with
the original base. This uses
\code{\link[Modstrings:separate]{removeIncompatibleModifications()}}
function from the \code{Modstrings} package.}

\item{metadata, reassign.ids}{See \code{\link[=makeEpiTxDb]{makeEpiTxDb}}}

\item{verbose}{\code{TRUE} or \code{FALSE}: Should verbose message be prined?}

\item{files}{From \code{organism}, \code{genome} and \code{modtype} the
available files will be downloaded using the
\code{\link[BiocFileCache:BiocFileCache-class]{BiocFileCache}} interface
and passed on to \code{makeEpiTxDbFromRMBaseFiles}. However, individual
files can be provided as well.}
}
\value{
a \code{EpiTxDb} object.
}
\description{
\code{makeEpiTxDbFromRMBase} will make use of the RMBase v2.0 online
resources.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeEpiTxDbFromGRanges.R
\name{makeEpiTxDbFromGRanges}
\alias{makeEpiTxDbFromGRanges}
\title{Create a \code{EpiTxDb} object from a \code{GRanges} object}
\usage{
makeEpiTxDbFromGRanges(gr, metadata = NULL, reassign.ids = FALSE)
}
\arguments{
\item{gr}{A \code{\link[GenomicRanges:GRanges-class]{GRanges}} object, which
contains at least the mandatory columns.}

\item{metadata}{A 2-column \code{data.frame} containing meta information to
be included in the \code{EpiTxDb} object. This \code{data.frame} is just
passed to \code{\link[=makeEpiTxDb]{makeEpiTxDb}}. See
\code{\link[=makeEpiTxDb]{makeEpiTxDb}} for more information about the format
of metadata. (default: \code{metadata = NULL})}

\item{reassign.ids}{= FALSE}
}
\value{
a \code{EpiTxDb} object.
}
\description{
\code{makeEpiTxDbFromGRanges} extracts informations from a
\code{\link[GenomicRanges:GRanges-class]{GRanges}} object. The following
metadata columns can be used:
\itemize{
\item \code{mod_id}, \code{mod_type}, \code{mod_name} and \code{tx_ensembl}.
The first three are mandatory, whereas \code{tx_ensembl} is optional.
\item \code{rx_genename}, \code{rx_rank}, \code{rx_ensembl},
\code{rx_ensembltrans} and \code{rx_entrezid}
\item \code{spec_type}, \code{spec_genename}, \code{spec_ensembl},
\code{spec_ensembltrans} and \code{spec_entrezid}
\item \code{ref_type} and \code{ref}
}
... and passed on the \code{\link[=makeEpiTxDb]{makeEpiTxDb}}.
}
\examples{
library(GenomicRanges)
gr <- GRanges(seqnames = "test",
              ranges = IRanges::IRanges(1,1),
              strand = "+",
              DataFrame(mod_id = 1L,
                        mod_type = "Am",
                        mod_name = "Am_1"))
etdb <- makeEpiTxDbFromGRanges(gr)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeEpiTxDbFromtRNAdb.R
\name{makeEpiTxDbFromtRNAdb}
\alias{makeEpiTxDbFromtRNAdb}
\alias{gettRNAdbDataAsGRanges}
\alias{listAvailableOrganismsFromtRNAdb}
\title{Create a \code{EpiTxDb} object from tRNAdb resources}
\usage{
gettRNAdbDataAsGRanges(
  organism,
  sequences = NULL,
  dbURL = tRNAdbImport::TRNA_DB_URL
)

makeEpiTxDbFromtRNAdb(
  organism,
  sequences = NULL,
  metadata = NULL,
  dbURL = tRNAdbImport::TRNA_DB_URL
)

listAvailableOrganismsFromtRNAdb()
}
\arguments{
\item{organism}{A \code{character} value for an organism available on the
tRNAdb website.}

\item{sequences}{A named \code{DNAStringSet} or \code{RNAStringSet}, which
will be used to associate a tRNAdb result with a specific transcript.}

\item{dbURL}{The URL to the tRNA db website.}

\item{metadata}{See \code{\link[=makeEpiTxDb]{makeEpiTxDb}}}
}
\value{
a \code{EpiTxDb} object.
}
\description{
\code{makeEpiTxDbFromtRNAdb} will make use of the tRNAdb online
resources and extract the modification information from the RNA database.

If a named \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} is
provided as \code{sequences}, the result from the tRNAdb will be matched
against the sequences. Valid matches will be used as transcript identifiers
and returned after a check of modification compatibility with the provided
sequence. By this process multiple copies of transcripts can be associated
with a single modification.

\code{makeEpiTxDbFromtRNAdb} uses the functions provided by the
\code{\link[tRNAdbImport:tRNAdbImport]{tRNAdbImport}} package.
\code{\link[tRNAdbImport:import.tRNAdb]{import.tRNAdb}} will be used with
\code{database = "RNA"} and the three different values for \code{origin}.
}
\examples{
\dontrun{
# getting just the annotation data
etdb <- makeEpiTxDbFromtRNAdb("Saccharomyces cerevisiae")

# For associating the result with transcripts, provide and additional
# named DNAStringSet object. Matching will be done against each sequence
# allowing 5 mismatches and indels. The final result will be checked for
# validity regarding the identity of the modifications
etdb <- makeEpiTxDbFromtRNAdb("Saccharomyces cerevisiae",
                              some_transcript_sequences)
}
}
\references{
Juehling F, Moerl M, Hartmann RK, Sprinzl M, Stadler PF, Puetz J. 2009.
"tRNAdb 2009: compilation of tRNA sequences and tRNA genes." Nucleic Acids
Research, Volume 37 (suppl_1): D159–162. doi:10.1093/nar/gkn772.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EpiTxDb.R
\docType{data}
\name{EpiTxDb-data}
\alias{EpiTxDb-data}
\alias{rmbase_data}
\title{EpiTxDb internal data}
\format{
data.frame
}
\usage{
data(rmbase_data)
}
\description{
EpiTxDb internal data
}
\keyword{datasets}
